'''
**Application programming interface for metaphor modeling engine.**

This interface is able to manage standard Neural Networks model as well as Chemical Graphmachine models.

Preliminary release with minimum usage functionalities.

@author: Jean-Luc PLOIX
'''
