# -*- coding: ISO-8859-1 -*-
'''
Created on 15 déc. 2017

@author: jeanluc
'''
import unittest

from . import modelutils, monalconst, monalrecords, Property, specialmath
from .datareader import _excel, check_sigs, datasource, excelsource, fileInspect
from .driver import _dataset, _driver, _driverlib, _drivermultidyn, _networkfactory, _optimizer, _parallelJobs, _workshop
from .lcode import codeconst, codewriter, codewriter, codewriterdetailC
from .library import libmanager
from .model import _link, _model, _node, _modelcontainer, _modellib, _monalfunc, _network, _network, _pseudomatrix
from .util import cleanxml, const_st, leverage, lineartoolbox, mathtoolbox, modelworkshop, monalbase, monaltoolbox, setuptoolbox, sqlite_copy_db, toolbox, utils


class Test(unittest.TestCase):


    def testName(self):
        pass

class Test_modelutils(unittest.TestCase):


    def testName(self):
        pass

class Test_monalconst(unittest.TestCase):


    def testName(self):
        pass

class Test_monalrecords(unittest.TestCase):


    def testName(self):
        pass

class Test_Property(unittest.TestCase):


    def testName(self):
        pass

class Test_specialmath(unittest.TestCase):


    def testName(self):
        pass

class Test_datareader_excel(unittest.TestCase):


    def testName(self):
        pass

class Test_datareader_check_sigs(unittest.TestCase):


    def testName(self):
        pass

class Test_datareader_datasource(unittest.TestCase):


    def testName(self):
        pass

class Test_datareader_excelsource(unittest.TestCase):


    def testName(self):
        pass

class Test_datareader_fileinspectr(unittest.TestCase):


    def testName(self):
        pass

# class Test_monal(unittest.TestCase):
# 
# 
#     def testName(self):
#         pass

class Test_driver_dataset(unittest.TestCase):


    def testName(self):
        pass

class Test_driver_driver(unittest.TestCase):


    def testName(self):
        pass

class Test_driver_driverlib(unittest.TestCase):


    def testName(self):
        pass

class Test_driver_drivermultudyn(unittest.TestCase):


    def testName(self):
        pass

class Test_driver_networkfactory(unittest.TestCase):


    def testName(self):
        pass

class Test_driver_optimizer(unittest.TestCase):


    def testName(self):
        pass

class Test_driver_parallelJob(unittest.TestCase):


    def testName(self):
        pass

class Test_driver_workshop(unittest.TestCase):


    def testName(self):
        pass

class Test_lcode_codeconst(unittest.TestCase):


    def testName(self):
        pass

class Test_lcode_codewriter(unittest.TestCase):


    def testName(self):
        pass

class Test_lcode_codewriterC(unittest.TestCase):


    def testName(self):
        pass

class Test_lcode_codewriterdetailC(unittest.TestCase):


    def testName(self):
        pass

class Test_library_libmabager(unittest.TestCase):


    def testName(self):
        pass

class Test_model_link(unittest.TestCase):


    def testName(self):
        pass

class Test_model_linmodel(unittest.TestCase):


    def testName(self):
        pass

class Test_model_model(unittest.TestCase):


    def testName(self):
        pass

class Test_model_modelcontainer(unittest.TestCase):


    def testName(self):
        pass

class Test_model_modellib(unittest.TestCase):


    def testName(self):
        pass

class Test_model_monalfunc(unittest.TestCase):


    def testName(self):
        pass

class Test_model_network(unittest.TestCase):


    def testName(self):
        pass

class Test_model_node(unittest.TestCase):


    def testName(self):
        pass

class Test_model_pseudomatrix(unittest.TestCase):


    def testName(self):
        pass

class Test_util_cleanxml(unittest.TestCase):


    def testName(self):
        pass

class Test_util_const_st(unittest.TestCase):


    def testName(self):
        pass

class Test_util_leverage(unittest.TestCase):


    def testName(self):
        pass

class Test_util_lineartoolbox(unittest.TestCase):


    def testName(self):
        pass

class Test_util_mathtoolbox(unittest.TestCase):


    def testName(self):
        pass

class Test_util_modelworkshop(unittest.TestCase):


    def testName(self):
        pass

class Test_util_modelbase(unittest.TestCase):


    def testName(self):
        pass

class Test_util_monaltoolbox(unittest.TestCase):


    def testName(self):
        pass

class Test_util_setuptoolbox(unittest.TestCase):


    def testName(self):
        pass

class Test_util_sqlite_copy_db(unittest.TestCase):


    def testName(self):
        pass

if __name__ == "__main__":
    unittest.main()