/* Generated by Cython 0.28.5 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "name": "metaphor.monal.lcode.codeconst",
        "sources": [
            "metaphor/monal/lcode/codeconst.py"
        ]
    },
    "module_name": "metaphor.monal.lcode.codeconst"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_28_5"
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (0 && PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0; // PyThread_create_key reports success always
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif // TSS (Thread Specific Storage) API
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__metaphor__monal__lcode__codeconst
#define __PYX_HAVE_API__metaphor__monal__lcode__codeconst
/* Early includes */
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "metaphor/monal/lcode/codeconst.py",
};

/*--- Type declarations ---*/

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'metaphor.monal.lcode.codeconst' */
#define __Pyx_MODULE_NAME "metaphor.monal.lcode.codeconst"
extern int __pyx_module_is_main_metaphor__monal__lcode__codeconst;
int __pyx_module_is_main_metaphor__monal__lcode__codeconst = 0;

/* Implementation of 'metaphor.monal.lcode.codeconst' */
static PyObject *__pyx_builtin_range;
static const char __pyx_k_Eq[] = "Eq";
static const char __pyx_k_nl[] = "nl";
static const char __pyx_k_DEq[] = "DEq";
static const char __pyx_k_DLM[] = "DLM";
static const char __pyx_k_I_s[] = "{$I %s}";
static const char __pyx_k_MAX[] = "MAX";
static const char __pyx_k_MEq[] = "MEq";
static const char __pyx_k_MIN[] = "MIN";
static const char __pyx_k_PEq[] = "PEq";
static const char __pyx_k__10[] = ".";
static const char __pyx_k__11[] = "\n";
static const char __pyx_k__12[] = ";\n";
static const char __pyx_k__13[] = "";
static const char __pyx_k__14[] = " = ";
static const char __pyx_k__15[] = " *= ";
static const char __pyx_k__16[] = " += ";
static const char __pyx_k__17[] = " /= ";
static const char __pyx_k__18[] = "|";
static const char __pyx_k_ctC[] = "ctC";
static const char __pyx_k_nlf[] = "nlf";
static const char __pyx_k_Desc[] = "Desc";
static const char __pyx_k_Make[] = "Make";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_ncPi[] = "ncPi";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_P_std[] = "(?P<std>%.+%)";
static const char __pyx_k_ctXML[] = "ctXML";
static const char __pyx_k_ftDLM[] = "ftDLM";
static const char __pyx_k_ftMax[] = "ftMax";
static const char __pyx_k_max_s[] = "max: %s";
static const char __pyx_k_min_s[] = "min: %s";
static const char __pyx_k_nkDyn[] = "nkDyn";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_train[] = "train";
static const char __pyx_k_vtMax[] = "vtMax";
static const char __pyx_k_vtMin[] = "vtMin";
static const char __pyx_k_Date_s[] = "Date     : %s";
static const char __pyx_k_MATRIX[] = "MATRIX";
static const char __pyx_k_Time_s[] = "Time     : %s";
static const char __pyx_k_User_s[] = "User     : %s";
static const char __pyx_k_cs_All[] = "cs_All";
static const char __pyx_k_dlmuse[] = "dlmuse";
static const char __pyx_k_ftBase[] = "ftBase";
static const char __pyx_k_ftDesc[] = "ftDesc";
static const char __pyx_k_ftMake[] = "ftMake";
static const char __pyx_k_ncNone[] = "ncNone";
static const char __pyx_k_nkStat[] = "nkStat";
static const char __pyx_k_nttype[] = "nttype";
static const char __pyx_k_std_re[] = "std_re";
static const char __pyx_k_MakeDLM[] = "MakeDLM";
static const char __pyx_k_SetupPy[] = "SetupPy";
static const char __pyx_k_T_sNode[] = "T%sNode";
static const char __pyx_k_WEIGHTS[] = "WEIGHTS";
static const char __pyx_k_ctExcel[] = "ctExcel";
static const char __pyx_k_dlmutil[] = "dlmutil";
static const char __pyx_k_fttrain[] = "fttrain";
static const char __pyx_k_ncSigma[] = "ncSigma";
static const char __pyx_k_nncName[] = "nncName";
static const char __pyx_k_nncNode[] = "nncNode";
static const char __pyx_k_npConst[] = "npConst";
static const char __pyx_k_npInput[] = "npInput";
static const char __pyx_k_npState[] = "npState";
static const char __pyx_k_stAlert[] = "stAlert";
static const char __pyx_k_usebase[] = "usebase";
static const char __pyx_k_version[] = "__version__";
static const char __pyx_k_Gradient[] = "Gradient";
static const char __pyx_k_KT_AFTER[] = "KT_AFTER";
static const char __pyx_k_REAL_8_X[] = "REAL*8 X";
static const char __pyx_k_STDTRDEV[] = "STDTRDEV";
static const char __pyx_k_Setupcfg[] = "Setupcfg";
static const char __pyx_k_StAdvert[] = "StAdvert";
static const char __pyx_k_StdDev_s[] = "StdDev: %s";
static const char __pyx_k_T_sInOut[] = "T%sInOut";
static const char __pyx_k_codename[] = "codename";
static const char __pyx_k_cs_Model[] = "cs_Model";
static const char __pyx_k_ctCython[] = "ctCython";
static const char __pyx_k_ctPascal[] = "ctPascal";
static const char __pyx_k_ctPython[] = "ctPython";
static const char __pyx_k_ctVBasic[] = "ctVBasic";
static const char __pyx_k_ftdlmuse[] = "ftdlmuse";
static const char __pyx_k_ftusedll[] = "ftusedll";
static const char __pyx_k_lower_re[] = "lower_re";
static const char __pyx_k_mathplus[] = "mathplus";
static const char __pyx_k_matrix_d[] = "matrix[%d]";
static const char __pyx_k_matrix_s[] = "matrix: %s";
static const char __pyx_k_npHidden[] = "npHidden";
static const char __pyx_k_npOutput[] = "npOutput";
static const char __pyx_k_readfile[] = "readfile";
static const char __pyx_k_sDIM_d_d[] = "%sDIM%d= %d";
static const char __pyx_k_sNODES_d[] = "%sNODES = %d";
static const char __pyx_k_sORDER_d[] = "%sORDER = %d";
static const char __pyx_k_stNodedS[] = "stNodedS";
static const char __pyx_k_stXdSF77[] = "stXdSF77";
static const char __pyx_k_stXdSF90[] = "stXdSF90";
static const char __pyx_k_st_InDev[] = "st_InDev";
static const char __pyx_k_upper_re[] = "upper_re";
static const char __pyx_k_vtStdDev[] = "vtStdDev";
static const char __pyx_k_vtWeight[] = "vtWeight";
static const char __pyx_k_KT_BEFORE[] = "KT_BEFORE";
static const char __pyx_k_StMatIniS[] = "StMatIniS";
static const char __pyx_k_StMaxIniS[] = "StMaxIniS";
static const char __pyx_k_StMinIniS[] = "StMinIniS";
static const char __pyx_k_T_sMatrix[] = "T%sMatrix";
static const char __pyx_k_T_sMemsyn[] = "T%sMemsyn";
static const char __pyx_k_T_sOutput[] = "T%sOutput";
static const char __pyx_k_Transfert[] = "Transfert";
static const char __pyx_k_code_Body[] = "code_Body";
static const char __pyx_k_cs_Common[] = "cs_Common";
static const char __pyx_k_ctUnknown[] = "ctUnknown";
static const char __pyx_k_dlmdriver[] = "dlmdriver";
static const char __pyx_k_ftMakeDLM[] = "ftMakeDLM";
static const char __pyx_k_ftSetupPy[] = "ftSetupPy";
static const char __pyx_k_ftbaseuse[] = "ftbaseuse";
static const char __pyx_k_ftdlmutil[] = "ftdlmutil";
static const char __pyx_k_include_s[] = "#include \"%s\"";
static const char __pyx_k_jurgb_txt[] = "jurgb.txt";
static const char __pyx_k_ntdefines[] = "ntdefines";
static const char __pyx_k_nttoolbox[] = "nttoolbox";
static const char __pyx_k_sINPUTS_d[] = "%sINPUTS= %d";
static const char __pyx_k_sMEMSYN_d[] = "%sMEMSYN= %d";
static const char __pyx_k_setup_cfg[] = "setup.cfg";
static const char __pyx_k_stNodedS2[] = "stNodedS2";
static const char __pyx_k_stNodedS3[] = "stNodedS3";
static const char __pyx_k_stNodedSK[] = "stNodedSK";
static const char __pyx_k_stNodedSP[] = "stNodedSP";
static const char __pyx_k_trainbase[] = "trainbase";
static const char __pyx_k_weights_d[] = "weights[%d]";
static const char __pyx_k_weights_s[] = "weights: %s";
static const char __pyx_k_Computer_s[] = "Computer : %s";
static const char __pyx_k_PyCodeList[] = "PyCodeList";
static const char __pyx_k_StIntParam[] = "StIntParam";
static const char __pyx_k_StMatIniSP[] = "StMatIniSP";
static const char __pyx_k_StMaxIniSF[] = "StMaxIniSF";
static const char __pyx_k_StMaxIniSP[] = "StMaxIniSP";
static const char __pyx_k_StMinIniSF[] = "StMinIniSF";
static const char __pyx_k_StMinIniSP[] = "StMinIniSP";
static const char __pyx_k_T_sSynapse[] = "T%sSynapse";
static const char __pyx_k_commondict[] = "commondict";
static const char __pyx_k_ftGradient[] = "ftGradient";
static const char __pyx_k_ftmathplus[] = "ftmathplus";
static const char __pyx_k_ftreadfile[] = "ftreadfile";
static const char __pyx_k_ftsetupcfg[] = "ftsetupcfg";
static const char __pyx_k_fttraindll[] = "fttraindll";
static const char __pyx_k_ftwithCext[] = "ftwithCext";
static const char __pyx_k_modeltrain[] = "modeltrain";
static const char __pyx_k_npOutState[] = "npOutState";
static const char __pyx_k_sINPUTNAME[] = "%sINPUTNAME";
static const char __pyx_k_sOUTPUTS_d[] = "%sOUTPUTS = %d";
static const char __pyx_k_stMemSyndS[] = "stMemSyndS";
static const char __pyx_k_stNodedSP2[] = "stNodedSP2";
static const char __pyx_k_stNodedSP3[] = "stNodedSP3";
static const char __pyx_k_stNodedSPK[] = "stNodedSPK";
static const char __pyx_k_st_Juridic[] = "st_Juridic";
static const char __pyx_k_subpattern[] = "subpattern";
static const char __pyx_k_useversion[] = "__useversion__";
static const char __pyx_k_GetFileName[] = "_GetFileName";
static const char __pyx_k_GetUseCount[] = "_GetUseCount";
static const char __pyx_k_GetUserName[] = "_GetUserName";
static const char __pyx_k_P_lower_l_l[] = "(?P<lower>%l.+l%)";
static const char __pyx_k_P_upper_U_U[] = "(?P<upper>%U.+U%)";
static const char __pyx_k_PyCodeList0[] = "PyCodeList0";
static const char __pyx_k_PyCodeList2[] = "PyCodeList2";
static const char __pyx_k_PyCodeMulti[] = "PyCodeMulti";
static const char __pyx_k_StAdvertAdv[] = "StAdvertAdv";
static const char __pyx_k_StConstDimn[] = "StConstDimn";
static const char __pyx_k_StIniBackdS[] = "StIniBackdS";
static const char __pyx_k_StInputName[] = "StInputName";
static const char __pyx_k_StMatrixIni[] = "StMatrixIni";
static const char __pyx_k_StWeightIni[] = "StWeightIni";
static const char __pyx_k_code_Header[] = "code_Header";
static const char __pyx_k_ctFortran77[] = "ctFortran77";
static const char __pyx_k_ctFortran90[] = "ctFortran90";
static const char __pyx_k_ftTransfert[] = "ftTransfert";
static const char __pyx_k_ftbasetrain[] = "ftbasetrain";
static const char __pyx_k_ftdlmdriver[] = "ftdlmdriver";
static const char __pyx_k_ftntdefines[] = "ftntdefines";
static const char __pyx_k_ftnttoolbox[] = "ftnttoolbox";
static const char __pyx_k_sMATRIXES_d[] = "%sMATRIXES= %d";
static const char __pyx_k_sOUTPUTNAME[] = "%sOUTPUTNAME";
static const char __pyx_k_sSTDTRDEV_g[] = "%sSTDTRDEV = %g";
static const char __pyx_k_sSYNAPSES_d[] = "%sSYNAPSES= %d";
static const char __pyx_k_stMemSyndSP[] = "stMemSyndSP";
static const char __pyx_k_stNodedSF77[] = "stNodedSF77";
static const char __pyx_k_stNodedSF90[] = "stNodedSF90";
static const char __pyx_k_stTypeNodes[] = "stTypeNodes";
static const char __pyx_k_GetLicenseNb[] = "_GetLicenseNb";
static const char __pyx_k_GetStudent95[] = "_GetStudent95";
static const char __pyx_k_StConstOrder[] = "StConstOrder";
static const char __pyx_k_StIniBackdSP[] = "StIniBackdSP";
static const char __pyx_k_StInputNameF[] = "StInputNameF";
static const char __pyx_k_StInputNameP[] = "StInputNameP";
static const char __pyx_k_StMatrixIniF[] = "StMatrixIniF";
static const char __pyx_k_StOutputName[] = "StOutputName";
static const char __pyx_k_StStdDEvIniF[] = "StStdDEvIniF";
static const char __pyx_k_StStdDEvIniP[] = "StStdDEvIniP";
static const char __pyx_k_StStdDEvInis[] = "StStdDEvInis";
static const char __pyx_k_StWeightIniF[] = "StWeightIniF";
static const char __pyx_k_StWeightIniS[] = "StWeightIniS";
static const char __pyx_k_ftOnlyHeader[] = "ftOnlyHeader";
static const char __pyx_k_ftSetupMulti[] = "ftSetupMulti";
static const char __pyx_k_ftSpecialSub[] = "ftSpecialSub";
static const char __pyx_k_ftWithHeader[] = "ftWithHeader";
static const char __pyx_k_ftmodeltrain[] = "ftmodeltrain";
static const char __pyx_k_ftnetraltype[] = "ftnetraltype";
static const char __pyx_k_generictrain[] = "generictrain";
static const char __pyx_k_node_T_sNode[] = "node: T%sNode";
static const char __pyx_k_sCONSTANTS_d[] = "%sCONSTANTS = %d";
static const char __pyx_k_sDIMENSION_d[] = "%sDIMENSION = %d";
static const char __pyx_k_sSTUDENT95_g[] = "%sSTUDENT95 = %g";
static const char __pyx_k_stAdvertDate[] = "stAdvertDate";
static const char __pyx_k_stAdvertTime[] = "stAdvertTime";
static const char __pyx_k_stAdvertUser[] = "stAdvertUser";
static const char __pyx_k_stConstLinks[] = "stConstLinks";
static const char __pyx_k_stConstNodes[] = "stConstNodes";
static const char __pyx_k_stDefineDimn[] = "stDefineDimn";
static const char __pyx_k_stGradientdS[] = "stGradientdS";
static const char __pyx_k_stNodedS2F77[] = "stNodedS2F77";
static const char __pyx_k_stNodedS2F90[] = "stNodedS2F90";
static const char __pyx_k_stNodedS3F77[] = "stNodedS3F77";
static const char __pyx_k_stNodedS3F90[] = "stNodedS3F90";
static const char __pyx_k_stTypeMinMax[] = "stTypeMinMax";
static const char __pyx_k_stTypeOutput[] = "stTypeOutput";
static const char __pyx_k_stTypeRetroP[] = "stTypeRetroP";
static const char __pyx_k_StArrayInOutP[] = "StArrayInOutP";
static const char __pyx_k_StArrayNodesP[] = "StArrayNodesP";
static const char __pyx_k_StConstInputs[] = "StConstInputs";
static const char __pyx_k_StConstOrderF[] = "StConstOrderF";
static const char __pyx_k_StConstStdDev[] = "StConstStdDev";
static const char __pyx_k_StConstStud95[] = "StConstStud95";
static const char __pyx_k_StDefineOrder[] = "StDefineOrder";
static const char __pyx_k_StInputNameFU[] = "StInputNameFU";
static const char __pyx_k_StOutputNameF[] = "StOutputNameF";
static const char __pyx_k_StOutputNameP[] = "StOutputNameP";
static const char __pyx_k_StTransferDyn[] = "StTransferDyn";
static const char __pyx_k_StWeightIniSP[] = "StWeightIniSP";
static const char __pyx_k_dlmdriverhelp[] = "dlmdriverhelp";
static const char __pyx_k_node_T_sNode2[] = "node: T%sNode2";
static const char __pyx_k_node_T_sNode3[] = "node: T%sNode3";
static const char __pyx_k_sSTDTRDEV_d_g[] = "%sSTDTRDEV[%d] = %g";
static const char __pyx_k_stArrayRetroP[] = "stArrayRetroP";
static const char __pyx_k_stConstLinksF[] = "stConstLinksF";
static const char __pyx_k_stConstMatrix[] = "stConstMatrix";
static const char __pyx_k_stConstNodesF[] = "stConstNodesF";
static const char __pyx_k_stDefineLinks[] = "stDefineLinks";
static const char __pyx_k_stDefineNodes[] = "stDefineNodes";
static const char __pyx_k_stGradientdSP[] = "stGradientdSP";
static const char __pyx_k_stIncludeFile[] = "stIncludeFile";
static const char __pyx_k_stMemSyndSF77[] = "stMemSyndSF77";
static const char __pyx_k_stMemSyndSF90[] = "stMemSyndSF90";
static const char __pyx_k_stTypeSynapse[] = "stTypeSynapse";
static const char __pyx_k_PyCodeListFull[] = "PyCodeListFull";
static const char __pyx_k_StArrayNodes2P[] = "StArrayNodes2P";
static const char __pyx_k_StArrayNodes3P[] = "StArrayNodes3P";
static const char __pyx_k_StConstInputsF[] = "StConstInputsF";
static const char __pyx_k_StConstOutputs[] = "StConstOutputs";
static const char __pyx_k_StConstStdDev1[] = "StConstStdDev1";
static const char __pyx_k_StDefineInputs[] = "StDefineInputs";
static const char __pyx_k_StDefineMIndex[] = "StDefineMIndex";
static const char __pyx_k_StOutputNameFU[] = "StOutputNameFU";
static const char __pyx_k_StTransGradDyn[] = "StTransGradDyn";
static const char __pyx_k_StTransferDyn2[] = "StTransferDyn2";
static const char __pyx_k_StTransferDynF[] = "StTransferDynF";
static const char __pyx_k_StTransferStat[] = "StTransferStat";
static const char __pyx_k_StdefineStdDev[] = "StdefineStdDev";
static const char __pyx_k_StdefineStud95[] = "StdefineStud95";
static const char __pyx_k_ftgenerictrain[] = "ftgenerictrain";
static const char __pyx_k_node_T_sOutput[] = "node: T%sOutput";
static const char __pyx_k_stConstMatrixF[] = "stConstMatrixF";
static const char __pyx_k_stDefineRetros[] = "stDefineRetros";
static const char __pyx_k_stIncludeFileP[] = "stIncludeFileP";
static const char __pyx_k_stPotentialdSP[] = "stPotentialdSP";
static const char __pyx_k_stTypeMatrixes[] = "stTypeMatrixes";
static const char __pyx_k_DO_I_2_sOUTPUTS[] = "DO I = 2, %sOUTPUTS";
static const char __pyx_k_GetComputerName[] = "_GetComputerName";
static const char __pyx_k_PyCodeListShort[] = "PyCodeListShort";
static const char __pyx_k_PyCodeMultiFull[] = "PyCodeMultiFull";
static const char __pyx_k_StArrayOutputsP[] = "StArrayOutputsP";
static const char __pyx_k_StConstOutputsF[] = "StConstOutputsF";
static const char __pyx_k_StDataMemSynF77[] = "StDataMemSynF77";
static const char __pyx_k_StDefineOutputs[] = "StDefineOutputs";
static const char __pyx_k_StForAllOutputs[] = "StForAllOutputs";
static const char __pyx_k_StInputNameDecl[] = "StInputNameDecl";
static const char __pyx_k_StInputsDeclF77[] = "StInputsDeclF77";
static const char __pyx_k_StInputsDeclF90[] = "StInputsDeclF90";
static const char __pyx_k_StMatrixCompute[] = "StMatrixCompute";
static const char __pyx_k_StTransGradDynF[] = "StTransGradDynF";
static const char __pyx_k_StTransGradStat[] = "StTransGradStat";
static const char __pyx_k_StTransferDynF2[] = "StTransferDynF2";
static const char __pyx_k_StTransferDynPP[] = "StTransferDynPP";
static const char __pyx_k_StTransferStat2[] = "StTransferStat2";
static const char __pyx_k_StTransferStatF[] = "StTransferStatF";
static const char __pyx_k_define_sDIM_d_d[] = "#define %sDIM%d %d";
static const char __pyx_k_define_sNODES_d[] = "#define %sNODES %d";
static const char __pyx_k_define_sORDER_d[] = "#define %sORDER %d";
static const char __pyx_k_ftdlmdriverhelp[] = "ftdlmdriverhelp";
static const char __pyx_k_interfaceusedll[] = "interfaceusedll";
static const char __pyx_k_proprietaryname[] = "proprietaryname";
static const char __pyx_k_singletrainbase[] = "singletrainbase";
static const char __pyx_k_stAdvertMachine[] = "stAdvertMachine";
static const char __pyx_k_stConstSynapses[] = "stConstSynapses";
static const char __pyx_k_stStatesDeclF77[] = "stStatesDeclF77";
static const char __pyx_k_stStatesDeclF90[] = "stStatesDeclF90";
static const char __pyx_k_stXPotentialdSP[] = "stXPotentialdSP";
static const char __pyx_k_stddev_sOUTPUTS[] = "stddev[%sOUTPUTS]";
static const char __pyx_k_DO_I_1_sSYNAPSES[] = "DO I = 1, %sSYNAPSES";
static const char __pyx_k_DO_J_1_sSYNAPSES[] = "DO J = 1, %sSYNAPSES";
static const char __pyx_k_PADouble_ADouble[] = "PADouble = ^ADouble";
static const char __pyx_k_REAL_8_POTENTIAL[] = "REAL*8 POTENTIAL";
static const char __pyx_k_StArrayMatrixF77[] = "StArrayMatrixF77";
static const char __pyx_k_StArrayMatrixF90[] = "StArrayMatrixF90";
static const char __pyx_k_StArraySynapsesP[] = "StArraySynapsesP";
static const char __pyx_k_StConstConstants[] = "StConstConstants";
static const char __pyx_k_StConstDimension[] = "StConstDimension";
static const char __pyx_k_StDefineTransfer[] = "StDefineTransfer";
static const char __pyx_k_StForAllOutputsP[] = "StForAllOutputsP";
static const char __pyx_k_StForAllSynapses[] = "StForAllSynapses";
static const char __pyx_k_StIfNDefTransfer[] = "StIfNDefTransfer";
static const char __pyx_k_StInputNameDeclP[] = "StInputNameDeclP";
static const char __pyx_k_StInputNameProto[] = "StInputNameProto";
static const char __pyx_k_StMatrixComputeF[] = "StMatrixComputeF";
static const char __pyx_k_StMatrixComputeP[] = "StMatrixComputeP";
static const char __pyx_k_StOutPutsDeclF77[] = "StOutPutsDeclF77";
static const char __pyx_k_StOutPutsDeclF90[] = "StOutPutsDeclF90";
static const char __pyx_k_StOutputNameDecl[] = "StOutputNameDecl";
static const char __pyx_k_StTransGradDynPP[] = "StTransGradDynPP";
static const char __pyx_k_StTransGradStatF[] = "StTransGradStatF";
static const char __pyx_k_StTransferDynPP2[] = "StTransferDynPP2";
static const char __pyx_k_StTransferStatF2[] = "StTransferStatF2";
static const char __pyx_k_StTransferStatPP[] = "StTransferStatPP";
static const char __pyx_k_StWeightRedirIni[] = "StWeightRedirIni";
static const char __pyx_k_code_common_maxi[] = "code_common_maxi";
static const char __pyx_k_code_common_mini[] = "code_common_mini";
static const char __pyx_k_define_sINPUTS_d[] = "#define %sINPUTS %d";
static const char __pyx_k_define_sMEMSYN_d[] = "#define %sMEMSYN %d";
static const char __pyx_k_ftsingletraindll[] = "ftsingletraindll";
static const char __pyx_k_matrix_sMATRIXES[] = "matrix[%sMATRIXES]";
static const char __pyx_k_memsyn_T_sMemsyn[] = "memsyn: T%sMemsyn";
static const char __pyx_k_potential_Double[] = "potential: Double";
static const char __pyx_k_stArrayMatrixesP[] = "stArrayMatrixesP";
static const char __pyx_k_stConstSynapsesF[] = "stConstSynapsesF";
static const char __pyx_k_stDefineMatrixes[] = "stDefineMatrixes";
static const char __pyx_k_stDefineSynapses[] = "stDefineSynapses";
static const char __pyx_k_stInibackDeclF77[] = "stInibackDeclF77";
static const char __pyx_k_stInibackDeclF90[] = "stInibackDeclF90";
static const char __pyx_k_stPotentialdSF77[] = "stPotentialdSF77";
static const char __pyx_k_stPotentialdSF90[] = "stPotentialdSF90";
static const char __pyx_k_INTEGER_PARAMETER[] = "INTEGER, PARAMETER :: ";
static const char __pyx_k_PyCodeSingleTrain[] = "PyCodeSingleTrain";
static const char __pyx_k_PyInterfaceCMulti[] = "PyInterfaceCMulti";
static const char __pyx_k_StConstConstantsF[] = "StConstConstantsF";
static const char __pyx_k_StDataGrIniLevF77[] = "StDataGrIniLevF77";
static const char __pyx_k_StDefineConstants[] = "StDefineConstants";
static const char __pyx_k_StForAllSynapsesP[] = "StForAllSynapsesP";
static const char __pyx_k_StForAllSynapsesj[] = "StForAllSynapsesj";
static const char __pyx_k_StOutputNameDeclP[] = "StOutputNameDeclP";
static const char __pyx_k_StOutputNameProto[] = "StOutputNameProto";
static const char __pyx_k_StReturnInputName[] = "StReturnInputName";
static const char __pyx_k_StTransGradStatPP[] = "StTransGradStatPP";
static const char __pyx_k_StTransferStatPP2[] = "StTransferStatPP2";
static const char __pyx_k_TransfertLeverage[] = "TransfertLeverage";
static const char __pyx_k_define_sOUTPUTS_d[] = "#define %sOUTPUTS %d";
static const char __pyx_k_ftsinglebasetrain[] = "ftsinglebasetrain";
static const char __pyx_k_iniback_T_sOutput[] = "iniback: T%sOutput";
static const char __pyx_k_interfacetraindll[] = "interfacetraindll";
static const char __pyx_k_stDefineDimension[] = "stDefineDimension";
static const char __pyx_k_stGradientDeclF77[] = "stGradientDeclF77";
static const char __pyx_k_stGradientDeclF90[] = "stGradientDeclF90";
static const char __pyx_k_stInibackDeclF90S[] = "stInibackDeclF90S";
static const char __pyx_k_stLeverageDeclF77[] = "stLeverageDeclF77";
static const char __pyx_k_stLeverageDeclF90[] = "stLeverageDeclF90";
static const char __pyx_k_stXPotentialdSF77[] = "stXPotentialdSF77";
static const char __pyx_k_stXPotentialdSF90[] = "stXPotentialdSF90";
static const char __pyx_k_weights_sSYNAPSES[] = "weights[%sSYNAPSES]";
static const char __pyx_k_REAL_8_X_POTENTIAL[] = "REAL*8 X, POTENTIAL";
static const char __pyx_k_StArraySynapsesF77[] = "StArraySynapsesF77";
static const char __pyx_k_StArraySynapsesF90[] = "StArraySynapsesF90";
static const char __pyx_k_StDefineTransferGr[] = "StDefineTransferGr";
static const char __pyx_k_StForAllInitMemSyn[] = "StForAllInitMemSyn";
static const char __pyx_k_StForAllOutputsF77[] = "StForAllOutputsF77";
static const char __pyx_k_StForAllOutputsF90[] = "StForAllOutputsF90";
static const char __pyx_k_StForAllSynapsesjP[] = "StForAllSynapsesjP";
static const char __pyx_k_StIfNDefTransferGr[] = "StIfNDefTransferGr";
static const char __pyx_k_StInputNameDeclF77[] = "StInputNameDeclF77";
static const char __pyx_k_StInputNameDeclF90[] = "StInputNameDeclF90";
static const char __pyx_k_StReturnOutputName[] = "StReturnOutputName";
static const char __pyx_k_StTransGradDynCall[] = "StTransGradDynCall";
static const char __pyx_k_StTransferDynProto[] = "StTransferDynProto";
static const char __pyx_k_Under_construction[] = "Under construction";
static const char __pyx_k_X_potential_Double[] = "X, potential: Double";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_define_sMATRIXES_d[] = "#define %sMATRIXES %d";
static const char __pyx_k_define_sMEMNEURO_d[] = "#define %sMEMNEURO %d";
static const char __pyx_k_define_sSTDTRDEV_g[] = "#define %sSTDTRDEV %g";
static const char __pyx_k_define_sSYNAPSES_d[] = "#define %sSYNAPSES %d";
static const char __pyx_k_define_sTRANSFER_H[] = "#define %sTRANSFER_H";
static const char __pyx_k_double_node_sNODES[] = "double node[%sNODES]";
static const char __pyx_k_ifndef_sTRANSFER_H[] = "#ifndef %sTRANSFER_H";
static const char __pyx_k_stArrayOfDoubleDef[] = "stArrayOfDoubleDef";
static const char __pyx_k_stGradientDeclF90S[] = "stGradientDeclF90S";
static const char __pyx_k_stLeverageDeclF90S[] = "stLeverageDeclF90S";
static const char __pyx_k_DO_140_I_2_sOUTPUTS[] = "DO 140 I = 2, %sOUTPUTS";
static const char __pyx_k_StDefineTransferLev[] = "StDefineTransferLev";
static const char __pyx_k_StForAllInitIniBack[] = "StForAllInitIniBack";
static const char __pyx_k_StForAllInitMemSynP[] = "StForAllInitMemSynP";
static const char __pyx_k_StForAllSynapsesF77[] = "StForAllSynapsesF77";
static const char __pyx_k_StForAllSynapsesF90[] = "StForAllSynapsesF90";
static const char __pyx_k_StIfNDefTransferLev[] = "StIfNDefTransferLev";
static const char __pyx_k_StOutputNameDeclF77[] = "StOutputNameDeclF77";
static const char __pyx_k_StOutputNameDeclF90[] = "StOutputNameDeclF90";
static const char __pyx_k_StReturnInputNameP1[] = "StReturnInputNameP1";
static const char __pyx_k_StTransGradDynCallF[] = "StTransGradDynCallF";
static const char __pyx_k_StTransGradDynCallP[] = "StTransGradDynCallP";
static const char __pyx_k_StTransGradDynProto[] = "StTransGradDynProto";
static const char __pyx_k_StTransGradStatCall[] = "StTransGradStatCall";
static const char __pyx_k_StTransferDynProto2[] = "StTransferDynProto2";
static const char __pyx_k_StTransferStatProto[] = "StTransferStatProto";
static const char __pyx_k_char_sinputname_int[] = "char *%sinputname(int)";
static const char __pyx_k_define_sCONSTANTS_d[] = "#define %sCONSTANTS %d";
static const char __pyx_k_define_sDIMENSION_d[] = "#define %sDIMENSION %d";
static const char __pyx_k_define_sSTUDENT95_g[] = "#define %sSTUDENT95 %g";
static const char __pyx_k_ftTransfertLeverage[] = "ftTransfertLeverage";
static const char __pyx_k_gradient_T_sSynapse[] = "gradient: T%sSynapse";
static const char __pyx_k_stPArrayOfDoubleDef[] = "stPArrayOfDoubleDef";
static const char __pyx_k_DO_140_I_1_sSYNAPSES[] = "DO 140 I = 1, %sSYNAPSES";
static const char __pyx_k_DO_150_J_1_sSYNAPSES[] = "DO 150 J = 1, %sSYNAPSES";
static const char __pyx_k_StForAllInitGradient[] = "StForAllInitGradient";
static const char __pyx_k_StForAllInitIniBackP[] = "StForAllInitIniBackP";
static const char __pyx_k_StForAllInitLeverage[] = "StForAllInitLeverage";
static const char __pyx_k_StForAllSynapsesjF77[] = "StForAllSynapsesjF77";
static const char __pyx_k_StForAllSynapsesjF90[] = "StForAllSynapsesjF90";
static const char __pyx_k_StReturnOutputNameP2[] = "StReturnOutputNameP2";
static const char __pyx_k_StTransGradStatCallF[] = "StTransGradStatCallF";
static const char __pyx_k_StTransGradStatCallP[] = "StTransGradStatCallP";
static const char __pyx_k_StTransGradStatProto[] = "StTransGradStatProto";
static const char __pyx_k_StTransferStatProto2[] = "StTransferStatProto2";
static const char __pyx_k_char_soutputname_int[] = "char *%soutputname(int)";
static const char __pyx_k_define_sTRANSFERGR_H[] = "#define %sTRANSFERGR_H";
static const char __pyx_k_double_node_sNODES_2[] = "double node[%sNODES][2]";
static const char __pyx_k_double_node_sNODES_3[] = "double node[%sNODES][3]";
static const char __pyx_k_double_node_sOUTPUTS[] = "double node[%sOUTPUTS]";
static const char __pyx_k_following_conditions[] = "following conditions :";
static const char __pyx_k_for_i_1_i_sOUTPUTS_i[] = "for (i=1; i < %sOUTPUTS; i++)";
static const char __pyx_k_ifndef_sTRANSFERGR_H[] = "#ifndef %sTRANSFERGR_H";
static const char __pyx_k_max_sINPUTS_sOUTPUTS[] = "max[%sINPUTS + %sOUTPUTS]";
static const char __pyx_k_min_sINPUTS_sOUTPUTS[] = "min[%sINPUTS + %sOUTPUTS]";
static const char __pyx_k_sMATRIX_K_GRADIENT_J[] = "%sMATRIX(K) * GRADIENT(J)";
static const char __pyx_k_smatrix_k_gradient_j[] = "%smatrix[k] * gradient[j]";
static const char __pyx_k_DATA_MEMSYN_sMEMSYN_1[] = "DATA MEMSYN /%sMEMSYN * 1./";
static const char __pyx_k_StForAllInitGradientP[] = "StForAllInitGradientP";
static const char __pyx_k_StForAllInitLeverageP[] = "StForAllInitLeverageP";
static const char __pyx_k_define_sTRANSFERLEV_H[] = "#define %sTRANSFERLEV_H";
static const char __pyx_k_double_memsyn_sMEMSYN[] = "double memsyn[%sMEMSYN]";
static const char __pyx_k_for_i_0_i_sSYNAPSES_i[] = "for (i=0; i < %sSYNAPSES; i++)";
static const char __pyx_k_for_j_0_j_sSYNAPSES_j[] = "for (j=0; j < %sSYNAPSES; j++)";
static const char __pyx_k_ifndef_sTRANSFERLEV_H[] = "#ifndef %sTRANSFERLEV_H";
static const char __pyx_k_is_clearly_mentionned[] = "is clearly mentionned.";
static const char __pyx_k_weights_weightsredir_d[] = "(*weights[weightsredir[%d]])";
static const char __pyx_k_DefaultDecimalSeparator[] = "DefaultDecimalSeparator";
static const char __pyx_k_double_iniback_sOUTPUTS[] = "double iniback[%sOUTPUTS]";
static const char __pyx_k_else_Result_names_index[] = "else Result := names[index]";
static const char __pyx_k_lutilisation_de_ce_code[] = "lutilisation de ce code.";
static const char __pyx_k_singleinterfacetraindll[] = "singleinterfacetraindll";
static const char __pyx_k_for_i_1_to_sOUTPUTS_1_do[] = "for i := 1 to %sOUTPUTS-1 do";
static const char __pyx_k_FUNCTION_sINPUTNAME_INDEX[] = "FUNCTION %sINPUTNAME(INDEX)";
static const char __pyx_k_LartenNetral_JLP_04_05_47[] = "LartenNetral_JLP_04_05_47";
static const char __pyx_k_char_sinputname_int_index[] = "char *%sinputname(int index)";
static const char __pyx_k_double_gradient_sSYNAPSES[] = "double gradient[%sSYNAPSES]";
static const char __pyx_k_for_i_0_to_sSYNAPSES_1_do[] = "for i := 0 to %sSYNAPSES-1 do";
static const char __pyx_k_for_j_0_to_sSYNAPSES_1_do[] = "for j := 0 to %sSYNAPSES-1 do";
static const char __pyx_k_static_char_names_sINPUTS[] = "static char *names[%sINPUTS] = {";
  static const char __pyx_k_Array_0_sNODES_1_of_double[] = "Array[0..%sNODES-1] of double";
  static const char __pyx_k_CHARACTER_32_names_sINPUTS[] = "CHARACTER*32 names(%sINPUTS)";
  static const char __pyx_k_FUNCTION_sOUTPUTNAME_INDEX[] = "FUNCTION %sOUTPUTNAME(INDEX)";
  static const char __pyx_k_char_soutputname_int_index[] = "char *%soutputname(int index)";
  static const char __pyx_k_static_char_names_sOUTPUTS[] = "static char *names[%sOUTPUTS] = {";
    static const char __pyx_k_Array_0_sMEMSYN_1_of_double[] = "Array[0..%sMEMSYN-1] of double";
    static const char __pyx_k_CHARACTER_32_names_sOUTPUTS[] = "CHARACTER*32 names(%sOUTPUTS)";
    static const char __pyx_k_Array_0_sOUTPUTS_1_of_double[] = "Array[0..%sOUTPUTS-1] of double";
    static const char __pyx_k_DIMENSION_REAL_8_NODE_sNODES[] = "DIMENSION REAL*8 NODE (%sNODES)";
    static const char __pyx_k_DIMENSION_REAL_8_s_sMATRIXES[] = "DIMENSION REAL*8 %s (%sMATRIXES)";
    static const char __pyx_k_DIMENSION_REAL_8_s_sSYNAPSES[] = "DIMENSION REAL*8 %s (%sSYNAPSES)";
    static const char __pyx_k_REAL_8_DIMENSION_sMATRIXES_s[] = "REAL*8, DIMENSION(%sMATRIXES) :: %s";
    static const char __pyx_k_REAL_8_DIMENSION_sNODES_NODE[] = "REAL*8, DIMENSION(%sNODES):: NODE";
    static const char __pyx_k_REAL_8_DIMENSION_sSYNAPSES_s[] = "REAL*8, DIMENSION(%sSYNAPSES) :: %s";
    static const char __pyx_k_for_i_0_i_sMEMSYN_memsyn_i_1[] = "for (i=0; i < %sMEMSYN; memsyn[i++] = 1)";
    static const char __pyx_k_void_stransfer_double_double[] = "void %stransfer(double*, double*)";
    static const char __pyx_k_Array_0_sMATRIXES_1_of_double[] = "Array[0..%sMATRIXES-1] of double";
    static const char __pyx_k_Array_0_sSYNAPSES_1_of_double[] = "Array[0..%sSYNAPSES-1] of double";
    static const char __pyx_k_smatrix_mindex_i_j_gradient_j[] = "%smatrix[mindex(i,j)] * gradient[j]";
    static const char __pyx_k_Array_0_sNODES_1_0_1_of_double[] = "Array[0..%sNODES-1, 0..1] of double";
    static const char __pyx_k_Array_0_sNODES_1_0_2_of_double[] = "Array[0..%sNODES-1, 0..2] of double";
    static const char __pyx_k_DATA_GRADIENT_INIBACK_LEVERAGE[] = "DATA GRADIENT, INIBACK, LEVERAGE /";
    static const char __pyx_k_DIMENSION_REAL_8_NODE_2_sNODES[] = "DIMENSION REAL*8 NODE (2, %sNODES)";
    static const char __pyx_k_DIMENSION_REAL_8_NODE_3_sNODES[] = "DIMENSION REAL*8 NODE (3, %sNODES)";
    static const char __pyx_k_DIMENSION_REAL_8_STATES_sORDER[] = "DIMENSION REAL*8 STATES (%sORDER)";
    static const char __pyx_k_REAL_8_DIMENSION_2_sNODES_NODE[] = "REAL*8, DIMENSION(2, %sNODES) :: NODE";
    static const char __pyx_k_REAL_8_DIMENSION_3_sNODES_NODE[] = "REAL*8, DIMENSION(3, %sNODES) :: NODE";
    static const char __pyx_k_REAL_8_DIMENSION_sORDER_STATES[] = "REAL*8, DIMENSION(%sORDER):: STATES";
    static const char __pyx_k_for_i_0_i_sOUTPUTS_iniback_i_0[] = "for (i=0; i < %sOUTPUTS; iniback[i++] = 0)";
    static const char __pyx_k_ADouble_Array_0_65535_Of_Double[] = "ADouble = Array [0..65535] Of Double";
    static const char __pyx_k_Ce_code_a_t_cre_automatiquement[] = "Ce code a \303\257\302\277\305\223t\303\257\302\277\305\223 cr\303\257\302\277\305\223e automatiquement le paquet monal.lcode";
    static const char __pyx_k_DIMENSION_REAL_8_INPUTS_sINPUTS[] = "DIMENSION REAL*8 INPUTS (%sINPUTS)";
    static const char __pyx_k_DIMENSION_REAL_8_MEMSYN_sMEMSYN[] = "DIMENSION REAL*8 MEMSYN (%sMEMSYN)";
    static const char __pyx_k_REAL_8_DIMENSION_sINPUTS_INPUTS[] = "REAL*8, DIMENSION(%sINPUTS) :: INPUTS";
    static const char __pyx_k_REAL_8_DIMENSION_sMEMSYN_MEMSYN[] = "REAL*8, DIMENSION(%sMEMSYN)   :: MEMSYN=(/(1.,i=1,%sMEMSYN)/)";
    static const char __pyx_k_define_mindex_i_j_i_j_j_i_i_1_2[] = "#define mindex(i,j) (i>j)?(j+i*(i+1)/2):(i+j*(j+1)/2)";
    static const char __pyx_k_for_i_0_i_sOUTPUTS_leverage_i_0[] = "for (i=0; i < %sOUTPUTS; leverage[i++] = 0)";
    static const char __pyx_k_sTRANSFERGRADIENT_INPUTS_OUTPUT[] = "%sTRANSFERGRADIENT (INPUTS, OUTPUTS, STATES, GRADIENT, INIBACK)";
    static const char __pyx_k_stransfergradient_inputs_output[] = "%stransfergradient(inputs, outputs, gradient, iniback)";
    static const char __pyx_k_void_stransferlev_double_double[] = "void %stransferlev(double*, double*, double*, double*, double*)";
    static const char __pyx_k_void_stransferlev_double_inputs[] = "void %stransferlev(double *inputs, double *outputs, double *states, double *leverage)";
    static const char __pyx_k_Array_0_sINPUTS_sOUTPUTS_1_of_do[] = "Array[0..%sINPUTS + %sOUTPUTS-1] of double";
    static const char __pyx_k_CALL_sTRANSFERGRADIENT_INPUTS_OU[] = "CALL %sTRANSFERGRADIENT (INPUTS, OUTPUTS, GRADIENT, INIBACK)";
    static const char __pyx_k_CHARACTER_LEN_32_DIMENSION_sINPU[] = "CHARACTER(LEN=32), DIMENSION(%sINPUTS) :: names";
    static const char __pyx_k_CHARACTER_LEN_32_DIMENSION_sOUTP[] = "CHARACTER(LEN=32), DIMENSION(%sOUTPUTS) :: names";
    static const char __pyx_k_Ce_code_est_destin_uniquement_au[] = "Ce code est destin\303\257\302\277\305\223 uniquement au test, et ne peut en aucun cas ";
    static const char __pyx_k_DIMENSION_REAL_8_GRADIENT_sSYNAP[] = "DIMENSION REAL*8 GRADIENT (%sSYNAPSES)";
    static const char __pyx_k_DIMENSION_REAL_8_INIBACK_sOUTPUT[] = "DIMENSION REAL*8 INIBACK (%sOUTPUTS)";
    static const char __pyx_k_DIMENSION_REAL_8_LEVERAGE_sOUTPU[] = "DIMENSION REAL*8 LEVERAGE (%sOUTPUTS)";
    static const char __pyx_k_DIMENSION_REAL_8_OUTPUTS_sOUTPUT[] = "DIMENSION REAL*8 OUTPUTS (%sOUTPUTS)";
    static const char __pyx_k_FUNCTION_sinputname_index_intege[] = "FUNCTION %sinputname(index: integer): String";
    static const char __pyx_k_FUNCTION_soutputname_index_integ[] = "FUNCTION %soutputname(index: integer): String";
    static const char __pyx_k_NETRAL_ne_peut_pas_tre_tenu_pour[] = "NETRAL ne peut pas \303\257\302\277\305\223tre tenu pour responsable des cons\303\257\302\277\305\223quences de";
    static const char __pyx_k_Netral_is_not_responsible_of_any[] = "Netral is not responsible of any consequence of the use if this";
    static const char __pyx_k_PROCEDURE_stranserlev_inputs_out[] = "PROCEDURE %stranserlev(inputs, outputs, states, leverage: PADouble)";
    static const char __pyx_k_PROCEDURE_stransfer_inputs_outpu[] = "PROCEDURE %stransfer(inputs, outputs: PADouble)";
    static const char __pyx_k_PROCEDURE_stransfergradient_inpu[] = "PROCEDURE %stransfergradient(inputs, outputs, gradient, iniback: PADouble)";
    static const char __pyx_k_PROCEDURE_stransferlev_inputs_ou[] = "PROCEDURE %stransferlev(inputs, outputs, leverage: PADouble)";
    static const char __pyx_k_Please_check_name_of_included_he[] = "Please check name of included header file";
    static const char __pyx_k_REAL_8_DIMENSION_sOUTPUTS_INIBAC[] = "REAL*8, DIMENSION(%sOUTPUTS):: INIBACK=(/(0.,i=1,%sOUTPUTS)/)";
    static const char __pyx_k_REAL_8_DIMENSION_sOUTPUTS_LEVERA[] = "REAL*8, DIMENSION(%sOUTPUTS):: LEVERAGE=(/(0.,i=1,%sOUTPUTS)/)";
    static const char __pyx_k_REAL_8_DIMENSION_sOUTPUTS_OUTPUT[] = "REAL*8, DIMENSION(%sOUTPUTS):: OUTPUTS";
    static const char __pyx_k_REAL_8_DIMENSION_sSYNAPSES_GRADI[] = "REAL*8, DIMENSION(%sSYNAPSES) :: GRADIENT=(/(0.,i=1,%sSYNAPSES)/)";
    static const char __pyx_k_SUBROUTINE_sTRANSFERGRADIENT_INP[] = "SUBROUTINE %sTRANSFERGRADIENT (INPUTS, OUTPUTS, GRADIENT, INIBACK)";
    static const char __pyx_k_SUBROUTINE_sTRANSFERLEV_INPUTS_O[] = "SUBROUTINE %sTRANSFERLEV (INPUTS, OUTPUTS, LEVERAGE)";
    static const char __pyx_k_SUBROUTINE_sTRANSFER_INPUTS_OUTP[] = "SUBROUTINE %sTRANSFER (INPUTS, OUTPUTS)";
    static const char __pyx_k_This_code_can_be_freely_used_wit[] = "This code can be freely used without any royalties with the";
    static const char __pyx_k_This_code_has_been_automatically[] = "This code has been automatically created the by the package monal.lcode";
    static const char __pyx_k_code_The_transmission_of_this_co[] = "code. The transmission of this code to any other person or";
    static const char __pyx_k_for_i_0_i_sSYNAPSES_gradient_i_0[] = "for (i=0; i < %sSYNAPSES; gradient[i++] = 0)";
    static const char __pyx_k_for_i_0_to_sMEMSYN_1_do_memsyn_i[] = "for i := 0 to %sMEMSYN-1 do memsyn[i] := 1";
    static const char __pyx_k_for_i_0_to_sOUTPUTS_1_do_Leverag[] = "for i := 0 to %sOUTPUTS-1 do Leverage[i] := 0";
    static const char __pyx_k_for_i_0_to_sOUTPUTS_1_do_iniback[] = "for i := 0 to %sOUTPUTS-1 do iniback[i] := 0";
    static const char __pyx_k_for_i_0_to_sSYNAPSES_1_do_gradie[] = "for i := 0 to %sSYNAPSES-1 do gradient[i] := 0";
    static const char __pyx_k_if_index_0_or_index_sINPUTS_then[] = "if (index < 0) or (index >= %sINPUTS) then Result := ';\n  StReturnInputNameP2 = 'else Result := names[index]';\n  StReturnOutputNameP1 =\n    'if (index < 0) or (index >= %sOUTPUTS) then Result := ";
    static const char __pyx_k_names_Array_0_sINPUTS_1_of_Strin[] = "names: Array[0..%sINPUTS-1] of String = (";
    static const char __pyx_k_names_Array_0_sOUTPUTS_1_of_Stri[] = "names: Array[0..%sOUTPUTS-1] of String = (";
    static const char __pyx_k_organisation_is_free_provided_th[] = "organisation is free, provided that the Netral origin of the code";
    static const char __pyx_k_return_index_0_index_sINPUTS_NUL[] = "return (index < 0 || index >= %sINPUTS) ? NULL : names[index]";
    static const char __pyx_k_return_index_0_index_sOUTPUTS_NU[] = "return (index < 0 || index >= %sOUTPUTS) ? NULL : names[index]";
    static const char __pyx_k_tre_utilis_de_faon_permanente_su[] = "\303\257\302\277\305\223tre utilis\303\257\302\277\305\223 de fa\303\257\302\277\305\223on permanente sur une installation industrielle.";
    static const char __pyx_k_void_stransfer_double_double_dou[] = "void %stransfer(double*, double*, double*)";
    static const char __pyx_k_void_stransfer_double_inputs_dou[] = "void %stransfer(double *inputs, double *outputs)";
    static const char __pyx_k_void_stransfergradient_double_do[] = "void %stransfergradient(double*, double*, double*, double*)";
    static const char __pyx_k_void_stransfergradient_double_in[] = "void %stransfergradient(double *inputs, double *outputs, double *gradient, double *iniback)";
    static const char __pyx_k_void_stransferlev_double_weights[] = "void %stransferlev(double *weights, double *zmat, double *inputs, double *outputs, double *leverage)";
    static const char __pyx_k_stransfergradient_inputs_output_2[] = "%stransfergradient(inputs, outputs, states, gradient, iniback)";
    static const char __pyx_k_stransfergradient_inputs_output_3[] = "%stransfergradient(inputs, outputs, @gradient, @iniback)";
    static const char __pyx_k_stransfergradient_inputs_output_4[] = "%stransfergradient(inputs, outputs, states, @gradient, @iniback)";
    static const char __pyx_k_void_stransferlev_double_double_2[] = "void %stransferlev(double*, double*, double*, double*)";
    static const char __pyx_k_PROCEDURE_stransfer_inputs_outpu_2[] = "PROCEDURE %stransfer(inputs, outputs, states: PADouble)";
    static const char __pyx_k_PROCEDURE_stransfergradient_inpu_2[] = "PROCEDURE %stransfergradient(inputs, outputs, states, gradient, iniback: PADouble)";
    static const char __pyx_k_REAL_8_DIMENSION_sOUTPUTS_INIBAC_2[] = "REAL*8, DIMENSION(%sOUTPUTS):: INIBACK";
    static const char __pyx_k_REAL_8_DIMENSION_sOUTPUTS_LEVERA_2[] = "REAL*8, DIMENSION(%sOUTPUTS):: LEVERAGE";
    static const char __pyx_k_REAL_8_DIMENSION_sSYNAPSES_GRADI_2[] = "REAL*8, DIMENSION(%sSYNAPSES) :: GRADIENT";
    static const char __pyx_k_SUBROUTINE_sTRANSFERGRADIENT_INP_2[] = "SUBROUTINE %sTRANSFERGRADIENT (INPUTS, OUTPUTS, STATES, GRADIENT, XINIBACK)";
    static const char __pyx_k_SUBROUTINE_sTRANSFERLEV_INPUTS_O_2[] = "SUBROUTINE %sTRANSFERLEV (INPUTS, OUTPUTS, STATES, LEVERAGE)";
    static const char __pyx_k_SUBROUTINE_sTRANSFER_INPUTS_OUTP_2[] = "SUBROUTINE %sTRANSFER (INPUTS, OUTPUTS, STATES)";
    static const char __pyx_k_void_stransfer_double_inputs_dou_2[] = "void %stransfer(double *inputs, double *outputs, double *states)";
    static const char __pyx_k_void_stransfergradient_double_do_2[] = "void %stransfergradient(double*, double*, double*, double*, double*)";
    static const char __pyx_k_void_stransfergradient_double_in_2[] = "void %stransfergradient(double *inputs, double *outputs, double *states, double *gradient, double *iniback)";
    static PyObject *__pyx_kp_u_ADouble_Array_0_65535_Of_Double;
    static PyObject *__pyx_kp_u_Array_0_sINPUTS_sOUTPUTS_1_of_do;
    static PyObject *__pyx_kp_u_Array_0_sMATRIXES_1_of_double;
    static PyObject *__pyx_kp_u_Array_0_sMEMSYN_1_of_double;
    static PyObject *__pyx_kp_u_Array_0_sNODES_1_0_1_of_double;
    static PyObject *__pyx_kp_u_Array_0_sNODES_1_0_2_of_double;
    static PyObject *__pyx_kp_u_Array_0_sNODES_1_of_double;
    static PyObject *__pyx_kp_u_Array_0_sOUTPUTS_1_of_double;
    static PyObject *__pyx_kp_u_Array_0_sSYNAPSES_1_of_double;
    static PyObject *__pyx_kp_u_CALL_sTRANSFERGRADIENT_INPUTS_OU;
    static PyObject *__pyx_kp_u_CHARACTER_32_names_sINPUTS;
    static PyObject *__pyx_kp_u_CHARACTER_32_names_sOUTPUTS;
    static PyObject *__pyx_kp_u_CHARACTER_LEN_32_DIMENSION_sINPU;
    static PyObject *__pyx_kp_u_CHARACTER_LEN_32_DIMENSION_sOUTP;
    static PyObject *__pyx_kp_u_Ce_code_a_t_cre_automatiquement;
    static PyObject *__pyx_kp_u_Ce_code_est_destin_uniquement_au;
    static PyObject *__pyx_kp_u_Computer_s;
    static PyObject *__pyx_kp_u_DATA_GRADIENT_INIBACK_LEVERAGE;
    static PyObject *__pyx_kp_u_DATA_MEMSYN_sMEMSYN_1;
    static PyObject *__pyx_n_s_DEq;
    static PyObject *__pyx_kp_u_DIMENSION_REAL_8_GRADIENT_sSYNAP;
    static PyObject *__pyx_kp_u_DIMENSION_REAL_8_INIBACK_sOUTPUT;
    static PyObject *__pyx_kp_u_DIMENSION_REAL_8_INPUTS_sINPUTS;
    static PyObject *__pyx_kp_u_DIMENSION_REAL_8_LEVERAGE_sOUTPU;
    static PyObject *__pyx_kp_u_DIMENSION_REAL_8_MEMSYN_sMEMSYN;
    static PyObject *__pyx_kp_u_DIMENSION_REAL_8_NODE_2_sNODES;
    static PyObject *__pyx_kp_u_DIMENSION_REAL_8_NODE_3_sNODES;
    static PyObject *__pyx_kp_u_DIMENSION_REAL_8_NODE_sNODES;
    static PyObject *__pyx_kp_u_DIMENSION_REAL_8_OUTPUTS_sOUTPUT;
    static PyObject *__pyx_kp_u_DIMENSION_REAL_8_STATES_sORDER;
    static PyObject *__pyx_kp_u_DIMENSION_REAL_8_s_sMATRIXES;
    static PyObject *__pyx_kp_u_DIMENSION_REAL_8_s_sSYNAPSES;
    static PyObject *__pyx_n_u_DLM;
    static PyObject *__pyx_kp_u_DO_140_I_1_sSYNAPSES;
    static PyObject *__pyx_kp_u_DO_140_I_2_sOUTPUTS;
    static PyObject *__pyx_kp_u_DO_150_J_1_sSYNAPSES;
    static PyObject *__pyx_kp_u_DO_I_1_sSYNAPSES;
    static PyObject *__pyx_kp_u_DO_I_2_sOUTPUTS;
    static PyObject *__pyx_kp_u_DO_J_1_sSYNAPSES;
    static PyObject *__pyx_kp_u_Date_s;
    static PyObject *__pyx_n_s_DefaultDecimalSeparator;
    static PyObject *__pyx_n_u_Desc;
    static PyObject *__pyx_n_s_Eq;
    static PyObject *__pyx_kp_u_FUNCTION_sINPUTNAME_INDEX;
    static PyObject *__pyx_kp_u_FUNCTION_sOUTPUTNAME_INDEX;
    static PyObject *__pyx_kp_u_FUNCTION_sinputname_index_intege;
    static PyObject *__pyx_kp_u_FUNCTION_soutputname_index_integ;
    static PyObject *__pyx_n_s_GetComputerName;
    static PyObject *__pyx_n_s_GetFileName;
    static PyObject *__pyx_n_s_GetLicenseNb;
    static PyObject *__pyx_n_s_GetStudent95;
    static PyObject *__pyx_n_s_GetUseCount;
    static PyObject *__pyx_n_s_GetUserName;
    static PyObject *__pyx_n_u_Gradient;
    static PyObject *__pyx_kp_u_INTEGER_PARAMETER;
    static PyObject *__pyx_kp_u_I_s;
    static PyObject *__pyx_n_s_KT_AFTER;
    static PyObject *__pyx_n_s_KT_BEFORE;
    static PyObject *__pyx_n_u_LartenNetral_JLP_04_05_47;
    static PyObject *__pyx_n_u_MATRIX;
    static PyObject *__pyx_n_u_MAX;
    static PyObject *__pyx_n_s_MEq;
    static PyObject *__pyx_n_u_MIN;
    static PyObject *__pyx_n_u_Make;
    static PyObject *__pyx_n_u_MakeDLM;
    static PyObject *__pyx_kp_u_NETRAL_ne_peut_pas_tre_tenu_pour;
    static PyObject *__pyx_kp_u_Netral_is_not_responsible_of_any;
    static PyObject *__pyx_kp_u_PADouble_ADouble;
    static PyObject *__pyx_n_s_PEq;
    static PyObject *__pyx_kp_u_PROCEDURE_stranserlev_inputs_out;
    static PyObject *__pyx_kp_u_PROCEDURE_stransfer_inputs_outpu;
    static PyObject *__pyx_kp_u_PROCEDURE_stransfer_inputs_outpu_2;
    static PyObject *__pyx_kp_u_PROCEDURE_stransfergradient_inpu;
    static PyObject *__pyx_kp_u_PROCEDURE_stransfergradient_inpu_2;
    static PyObject *__pyx_kp_u_PROCEDURE_stransferlev_inputs_ou;
    static PyObject *__pyx_kp_u_P_lower_l_l;
    static PyObject *__pyx_kp_u_P_std;
    static PyObject *__pyx_kp_u_P_upper_U_U;
    static PyObject *__pyx_kp_u_Please_check_name_of_included_he;
    static PyObject *__pyx_n_s_PyCodeList;
    static PyObject *__pyx_n_s_PyCodeList0;
    static PyObject *__pyx_n_s_PyCodeList2;
    static PyObject *__pyx_n_s_PyCodeListFull;
    static PyObject *__pyx_n_s_PyCodeListShort;
    static PyObject *__pyx_n_s_PyCodeMulti;
    static PyObject *__pyx_n_s_PyCodeMultiFull;
    static PyObject *__pyx_n_s_PyCodeSingleTrain;
    static PyObject *__pyx_n_u_PyInterfaceCMulti;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_2_sNODES_NODE;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_3_sNODES_NODE;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_sINPUTS_INPUTS;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_sMATRIXES_s;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_sMEMSYN_MEMSYN;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_sNODES_NODE;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_sORDER_STATES;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_INIBAC;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_INIBAC_2;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_LEVERA;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_LEVERA_2;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_OUTPUT;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_sSYNAPSES_GRADI;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_sSYNAPSES_GRADI_2;
    static PyObject *__pyx_kp_u_REAL_8_DIMENSION_sSYNAPSES_s;
    static PyObject *__pyx_kp_u_REAL_8_POTENTIAL;
    static PyObject *__pyx_kp_u_REAL_8_X;
    static PyObject *__pyx_kp_u_REAL_8_X_POTENTIAL;
    static PyObject *__pyx_n_u_STDTRDEV;
    static PyObject *__pyx_kp_u_SUBROUTINE_sTRANSFERGRADIENT_INP;
    static PyObject *__pyx_kp_u_SUBROUTINE_sTRANSFERGRADIENT_INP_2;
    static PyObject *__pyx_kp_u_SUBROUTINE_sTRANSFERLEV_INPUTS_O;
    static PyObject *__pyx_kp_u_SUBROUTINE_sTRANSFERLEV_INPUTS_O_2;
    static PyObject *__pyx_kp_u_SUBROUTINE_sTRANSFER_INPUTS_OUTP;
    static PyObject *__pyx_kp_u_SUBROUTINE_sTRANSFER_INPUTS_OUTP_2;
    static PyObject *__pyx_n_u_SetupPy;
    static PyObject *__pyx_n_u_Setupcfg;
    static PyObject *__pyx_n_s_StAdvert;
    static PyObject *__pyx_n_s_StAdvertAdv;
    static PyObject *__pyx_n_s_StArrayInOutP;
    static PyObject *__pyx_n_s_StArrayMatrixF77;
    static PyObject *__pyx_n_s_StArrayMatrixF90;
    static PyObject *__pyx_n_s_StArrayNodes2P;
    static PyObject *__pyx_n_s_StArrayNodes3P;
    static PyObject *__pyx_n_s_StArrayNodesP;
    static PyObject *__pyx_n_s_StArrayOutputsP;
    static PyObject *__pyx_n_s_StArraySynapsesF77;
    static PyObject *__pyx_n_s_StArraySynapsesF90;
    static PyObject *__pyx_n_s_StArraySynapsesP;
    static PyObject *__pyx_n_s_StConstConstants;
    static PyObject *__pyx_n_s_StConstConstantsF;
    static PyObject *__pyx_n_s_StConstDimension;
    static PyObject *__pyx_n_s_StConstDimn;
    static PyObject *__pyx_n_s_StConstInputs;
    static PyObject *__pyx_n_s_StConstInputsF;
    static PyObject *__pyx_n_s_StConstOrder;
    static PyObject *__pyx_n_s_StConstOrderF;
    static PyObject *__pyx_n_s_StConstOutputs;
    static PyObject *__pyx_n_s_StConstOutputsF;
    static PyObject *__pyx_n_s_StConstStdDev;
    static PyObject *__pyx_n_s_StConstStdDev1;
    static PyObject *__pyx_n_s_StConstStud95;
    static PyObject *__pyx_n_s_StDataGrIniLevF77;
    static PyObject *__pyx_n_s_StDataMemSynF77;
    static PyObject *__pyx_n_s_StDefineConstants;
    static PyObject *__pyx_n_s_StDefineInputs;
    static PyObject *__pyx_n_s_StDefineMIndex;
    static PyObject *__pyx_n_s_StDefineOrder;
    static PyObject *__pyx_n_s_StDefineOutputs;
    static PyObject *__pyx_n_s_StDefineTransfer;
    static PyObject *__pyx_n_s_StDefineTransferGr;
    static PyObject *__pyx_n_s_StDefineTransferLev;
    static PyObject *__pyx_n_s_StForAllInitGradient;
    static PyObject *__pyx_n_s_StForAllInitGradientP;
    static PyObject *__pyx_n_s_StForAllInitIniBack;
    static PyObject *__pyx_n_s_StForAllInitIniBackP;
    static PyObject *__pyx_n_s_StForAllInitLeverage;
    static PyObject *__pyx_n_s_StForAllInitLeverageP;
    static PyObject *__pyx_n_s_StForAllInitMemSyn;
    static PyObject *__pyx_n_s_StForAllInitMemSynP;
    static PyObject *__pyx_n_s_StForAllOutputs;
    static PyObject *__pyx_n_s_StForAllOutputsF77;
    static PyObject *__pyx_n_s_StForAllOutputsF90;
    static PyObject *__pyx_n_s_StForAllOutputsP;
    static PyObject *__pyx_n_s_StForAllSynapses;
    static PyObject *__pyx_n_s_StForAllSynapsesF77;
    static PyObject *__pyx_n_s_StForAllSynapsesF90;
    static PyObject *__pyx_n_s_StForAllSynapsesP;
    static PyObject *__pyx_n_s_StForAllSynapsesj;
    static PyObject *__pyx_n_s_StForAllSynapsesjF77;
    static PyObject *__pyx_n_s_StForAllSynapsesjF90;
    static PyObject *__pyx_n_s_StForAllSynapsesjP;
    static PyObject *__pyx_n_s_StIfNDefTransfer;
    static PyObject *__pyx_n_s_StIfNDefTransferGr;
    static PyObject *__pyx_n_s_StIfNDefTransferLev;
    static PyObject *__pyx_n_s_StIniBackdS;
    static PyObject *__pyx_n_s_StIniBackdSP;
    static PyObject *__pyx_n_s_StInputName;
    static PyObject *__pyx_n_s_StInputNameDecl;
    static PyObject *__pyx_n_s_StInputNameDeclF77;
    static PyObject *__pyx_n_s_StInputNameDeclF90;
    static PyObject *__pyx_n_s_StInputNameDeclP;
    static PyObject *__pyx_n_s_StInputNameF;
    static PyObject *__pyx_n_s_StInputNameFU;
    static PyObject *__pyx_n_s_StInputNameP;
    static PyObject *__pyx_n_s_StInputNameProto;
    static PyObject *__pyx_n_s_StInputsDeclF77;
    static PyObject *__pyx_n_s_StInputsDeclF90;
    static PyObject *__pyx_n_s_StIntParam;
    static PyObject *__pyx_n_s_StMatIniS;
    static PyObject *__pyx_n_s_StMatIniSP;
    static PyObject *__pyx_n_s_StMatrixCompute;
    static PyObject *__pyx_n_s_StMatrixComputeF;
    static PyObject *__pyx_n_s_StMatrixComputeP;
    static PyObject *__pyx_n_s_StMatrixIni;
    static PyObject *__pyx_n_s_StMatrixIniF;
    static PyObject *__pyx_n_s_StMaxIniS;
    static PyObject *__pyx_n_s_StMaxIniSF;
    static PyObject *__pyx_n_s_StMaxIniSP;
    static PyObject *__pyx_n_s_StMinIniS;
    static PyObject *__pyx_n_s_StMinIniSF;
    static PyObject *__pyx_n_s_StMinIniSP;
    static PyObject *__pyx_n_s_StOutPutsDeclF77;
    static PyObject *__pyx_n_s_StOutPutsDeclF90;
    static PyObject *__pyx_n_s_StOutputName;
    static PyObject *__pyx_n_s_StOutputNameDecl;
    static PyObject *__pyx_n_s_StOutputNameDeclF77;
    static PyObject *__pyx_n_s_StOutputNameDeclF90;
    static PyObject *__pyx_n_s_StOutputNameDeclP;
    static PyObject *__pyx_n_s_StOutputNameF;
    static PyObject *__pyx_n_s_StOutputNameFU;
    static PyObject *__pyx_n_s_StOutputNameP;
    static PyObject *__pyx_n_s_StOutputNameProto;
    static PyObject *__pyx_n_s_StReturnInputName;
    static PyObject *__pyx_n_s_StReturnInputNameP1;
    static PyObject *__pyx_n_s_StReturnOutputName;
    static PyObject *__pyx_n_s_StReturnOutputNameP2;
    static PyObject *__pyx_n_s_StStdDEvIniF;
    static PyObject *__pyx_n_s_StStdDEvIniP;
    static PyObject *__pyx_n_s_StStdDEvInis;
    static PyObject *__pyx_n_s_StTransGradDyn;
    static PyObject *__pyx_n_s_StTransGradDynCall;
    static PyObject *__pyx_n_s_StTransGradDynCallF;
    static PyObject *__pyx_n_s_StTransGradDynCallP;
    static PyObject *__pyx_n_s_StTransGradDynF;
    static PyObject *__pyx_n_s_StTransGradDynPP;
    static PyObject *__pyx_n_s_StTransGradDynProto;
    static PyObject *__pyx_n_s_StTransGradStat;
    static PyObject *__pyx_n_s_StTransGradStatCall;
    static PyObject *__pyx_n_s_StTransGradStatCallF;
    static PyObject *__pyx_n_s_StTransGradStatCallP;
    static PyObject *__pyx_n_s_StTransGradStatF;
    static PyObject *__pyx_n_s_StTransGradStatPP;
    static PyObject *__pyx_n_s_StTransGradStatProto;
    static PyObject *__pyx_n_s_StTransferDyn;
    static PyObject *__pyx_n_s_StTransferDyn2;
    static PyObject *__pyx_n_s_StTransferDynF;
    static PyObject *__pyx_n_s_StTransferDynF2;
    static PyObject *__pyx_n_s_StTransferDynPP;
    static PyObject *__pyx_n_s_StTransferDynPP2;
    static PyObject *__pyx_n_s_StTransferDynProto;
    static PyObject *__pyx_n_s_StTransferDynProto2;
    static PyObject *__pyx_n_s_StTransferStat;
    static PyObject *__pyx_n_s_StTransferStat2;
    static PyObject *__pyx_n_s_StTransferStatF;
    static PyObject *__pyx_n_s_StTransferStatF2;
    static PyObject *__pyx_n_s_StTransferStatPP;
    static PyObject *__pyx_n_s_StTransferStatPP2;
    static PyObject *__pyx_n_s_StTransferStatProto;
    static PyObject *__pyx_n_s_StTransferStatProto2;
    static PyObject *__pyx_n_s_StWeightIni;
    static PyObject *__pyx_n_s_StWeightIniF;
    static PyObject *__pyx_n_s_StWeightIniS;
    static PyObject *__pyx_n_s_StWeightIniSP;
    static PyObject *__pyx_n_s_StWeightRedirIni;
    static PyObject *__pyx_kp_u_StdDev_s;
    static PyObject *__pyx_n_s_StdefineStdDev;
    static PyObject *__pyx_n_s_StdefineStud95;
    static PyObject *__pyx_kp_u_T_sInOut;
    static PyObject *__pyx_kp_u_T_sMatrix;
    static PyObject *__pyx_kp_u_T_sMemsyn;
    static PyObject *__pyx_kp_u_T_sNode;
    static PyObject *__pyx_kp_u_T_sOutput;
    static PyObject *__pyx_kp_u_T_sSynapse;
    static PyObject *__pyx_kp_u_This_code_can_be_freely_used_wit;
    static PyObject *__pyx_kp_u_This_code_has_been_automatically;
    static PyObject *__pyx_kp_u_Time_s;
    static PyObject *__pyx_n_u_Transfert;
    static PyObject *__pyx_n_u_TransfertLeverage;
    static PyObject *__pyx_kp_u_Under_construction;
    static PyObject *__pyx_kp_u_User_s;
    static PyObject *__pyx_n_u_WEIGHTS;
    static PyObject *__pyx_kp_u_X_potential_Double;
    static PyObject *__pyx_kp_u__10;
    static PyObject *__pyx_kp_u__11;
    static PyObject *__pyx_kp_u__12;
    static PyObject *__pyx_kp_u__13;
    static PyObject *__pyx_kp_u__14;
    static PyObject *__pyx_kp_u__15;
    static PyObject *__pyx_kp_u__16;
    static PyObject *__pyx_kp_u__17;
    static PyObject *__pyx_kp_u__18;
    static PyObject *__pyx_n_u_base;
    static PyObject *__pyx_kp_u_char_sinputname_int;
    static PyObject *__pyx_kp_u_char_sinputname_int_index;
    static PyObject *__pyx_kp_u_char_soutputname_int;
    static PyObject *__pyx_kp_u_char_soutputname_int_index;
    static PyObject *__pyx_n_s_cline_in_traceback;
    static PyObject *__pyx_n_s_code_Body;
    static PyObject *__pyx_n_s_code_Header;
    static PyObject *__pyx_kp_u_code_The_transmission_of_this_co;
    static PyObject *__pyx_n_s_code_common_maxi;
    static PyObject *__pyx_n_s_code_common_mini;
    static PyObject *__pyx_n_s_codename;
    static PyObject *__pyx_n_s_commondict;
    static PyObject *__pyx_n_s_cs_All;
    static PyObject *__pyx_n_s_cs_Common;
    static PyObject *__pyx_n_s_cs_Model;
    static PyObject *__pyx_n_s_ctC;
    static PyObject *__pyx_n_s_ctCython;
    static PyObject *__pyx_n_s_ctExcel;
    static PyObject *__pyx_n_s_ctFortran77;
    static PyObject *__pyx_n_s_ctFortran90;
    static PyObject *__pyx_n_s_ctPascal;
    static PyObject *__pyx_n_s_ctPython;
    static PyObject *__pyx_n_s_ctUnknown;
    static PyObject *__pyx_n_s_ctVBasic;
    static PyObject *__pyx_n_s_ctXML;
    static PyObject *__pyx_kp_u_define_mindex_i_j_i_j_j_i_i_1_2;
    static PyObject *__pyx_kp_u_define_sCONSTANTS_d;
    static PyObject *__pyx_kp_u_define_sDIMENSION_d;
    static PyObject *__pyx_kp_u_define_sDIM_d_d;
    static PyObject *__pyx_kp_u_define_sINPUTS_d;
    static PyObject *__pyx_kp_u_define_sMATRIXES_d;
    static PyObject *__pyx_kp_u_define_sMEMNEURO_d;
    static PyObject *__pyx_kp_u_define_sMEMSYN_d;
    static PyObject *__pyx_kp_u_define_sNODES_d;
    static PyObject *__pyx_kp_u_define_sORDER_d;
    static PyObject *__pyx_kp_u_define_sOUTPUTS_d;
    static PyObject *__pyx_kp_u_define_sSTDTRDEV_g;
    static PyObject *__pyx_kp_u_define_sSTUDENT95_g;
    static PyObject *__pyx_kp_u_define_sSYNAPSES_d;
    static PyObject *__pyx_kp_u_define_sTRANSFERGR_H;
    static PyObject *__pyx_kp_u_define_sTRANSFERLEV_H;
    static PyObject *__pyx_kp_u_define_sTRANSFER_H;
    static PyObject *__pyx_n_u_dlmdriver;
    static PyObject *__pyx_n_u_dlmdriverhelp;
    static PyObject *__pyx_n_u_dlmuse;
    static PyObject *__pyx_n_u_dlmutil;
    static PyObject *__pyx_kp_u_double_gradient_sSYNAPSES;
    static PyObject *__pyx_kp_u_double_iniback_sOUTPUTS;
    static PyObject *__pyx_kp_u_double_memsyn_sMEMSYN;
    static PyObject *__pyx_kp_u_double_node_sNODES;
    static PyObject *__pyx_kp_u_double_node_sNODES_2;
    static PyObject *__pyx_kp_u_double_node_sNODES_3;
    static PyObject *__pyx_kp_u_double_node_sOUTPUTS;
    static PyObject *__pyx_kp_u_else_Result_names_index;
    static PyObject *__pyx_kp_u_following_conditions;
    static PyObject *__pyx_kp_u_for_i_0_i_sMEMSYN_memsyn_i_1;
    static PyObject *__pyx_kp_u_for_i_0_i_sOUTPUTS_iniback_i_0;
    static PyObject *__pyx_kp_u_for_i_0_i_sOUTPUTS_leverage_i_0;
    static PyObject *__pyx_kp_u_for_i_0_i_sSYNAPSES_gradient_i_0;
    static PyObject *__pyx_kp_u_for_i_0_i_sSYNAPSES_i;
    static PyObject *__pyx_kp_u_for_i_0_to_sMEMSYN_1_do_memsyn_i;
    static PyObject *__pyx_kp_u_for_i_0_to_sOUTPUTS_1_do_Leverag;
    static PyObject *__pyx_kp_u_for_i_0_to_sOUTPUTS_1_do_iniback;
    static PyObject *__pyx_kp_u_for_i_0_to_sSYNAPSES_1_do;
    static PyObject *__pyx_kp_u_for_i_0_to_sSYNAPSES_1_do_gradie;
    static PyObject *__pyx_kp_u_for_i_1_i_sOUTPUTS_i;
    static PyObject *__pyx_kp_u_for_i_1_to_sOUTPUTS_1_do;
    static PyObject *__pyx_kp_u_for_j_0_j_sSYNAPSES_j;
    static PyObject *__pyx_kp_u_for_j_0_to_sSYNAPSES_1_do;
    static PyObject *__pyx_n_s_ftBase;
    static PyObject *__pyx_n_s_ftDLM;
    static PyObject *__pyx_n_s_ftDesc;
    static PyObject *__pyx_n_s_ftGradient;
    static PyObject *__pyx_n_s_ftMake;
    static PyObject *__pyx_n_s_ftMakeDLM;
    static PyObject *__pyx_n_s_ftMax;
    static PyObject *__pyx_n_s_ftOnlyHeader;
    static PyObject *__pyx_n_s_ftSetupMulti;
    static PyObject *__pyx_n_s_ftSetupPy;
    static PyObject *__pyx_n_s_ftSpecialSub;
    static PyObject *__pyx_n_s_ftTransfert;
    static PyObject *__pyx_n_s_ftTransfertLeverage;
    static PyObject *__pyx_n_s_ftWithHeader;
    static PyObject *__pyx_n_s_ftbasetrain;
    static PyObject *__pyx_n_s_ftbaseuse;
    static PyObject *__pyx_n_s_ftdlmdriver;
    static PyObject *__pyx_n_s_ftdlmdriverhelp;
    static PyObject *__pyx_n_s_ftdlmuse;
    static PyObject *__pyx_n_s_ftdlmutil;
    static PyObject *__pyx_n_s_ftgenerictrain;
    static PyObject *__pyx_n_s_ftmathplus;
    static PyObject *__pyx_n_s_ftmodeltrain;
    static PyObject *__pyx_n_s_ftnetraltype;
    static PyObject *__pyx_n_s_ftntdefines;
    static PyObject *__pyx_n_s_ftnttoolbox;
    static PyObject *__pyx_n_s_ftreadfile;
    static PyObject *__pyx_n_s_ftsetupcfg;
    static PyObject *__pyx_n_s_ftsinglebasetrain;
    static PyObject *__pyx_n_s_ftsingletraindll;
    static PyObject *__pyx_n_s_fttrain;
    static PyObject *__pyx_n_s_fttraindll;
    static PyObject *__pyx_n_s_ftusedll;
    static PyObject *__pyx_n_s_ftwithCext;
    static PyObject *__pyx_n_u_generictrain;
    static PyObject *__pyx_kp_u_gradient_T_sSynapse;
    static PyObject *__pyx_kp_u_if_index_0_or_index_sINPUTS_then;
    static PyObject *__pyx_kp_u_ifndef_sTRANSFERGR_H;
    static PyObject *__pyx_kp_u_ifndef_sTRANSFERLEV_H;
    static PyObject *__pyx_kp_u_ifndef_sTRANSFER_H;
    static PyObject *__pyx_kp_u_include_s;
    static PyObject *__pyx_kp_u_iniback_T_sOutput;
    static PyObject *__pyx_n_u_interfacetraindll;
    static PyObject *__pyx_n_u_interfaceusedll;
    static PyObject *__pyx_kp_u_is_clearly_mentionned;
    static PyObject *__pyx_kp_u_jurgb_txt;
    static PyObject *__pyx_n_s_lower_re;
    static PyObject *__pyx_kp_u_lutilisation_de_ce_code;
    static PyObject *__pyx_n_s_main;
    static PyObject *__pyx_n_u_mathplus;
    static PyObject *__pyx_kp_u_matrix_d;
    static PyObject *__pyx_kp_u_matrix_s;
    static PyObject *__pyx_kp_u_matrix_sMATRIXES;
    static PyObject *__pyx_kp_u_max_s;
    static PyObject *__pyx_kp_u_max_sINPUTS_sOUTPUTS;
    static PyObject *__pyx_kp_u_memsyn_T_sMemsyn;
    static PyObject *__pyx_kp_u_min_s;
    static PyObject *__pyx_kp_u_min_sINPUTS_sOUTPUTS;
    static PyObject *__pyx_n_u_modeltrain;
    static PyObject *__pyx_kp_u_names_Array_0_sINPUTS_1_of_Strin;
    static PyObject *__pyx_kp_u_names_Array_0_sOUTPUTS_1_of_Stri;
    static PyObject *__pyx_n_s_ncNone;
    static PyObject *__pyx_n_s_ncPi;
    static PyObject *__pyx_n_s_ncSigma;
    static PyObject *__pyx_n_s_nkDyn;
    static PyObject *__pyx_n_s_nkStat;
    static PyObject *__pyx_n_s_nl;
    static PyObject *__pyx_n_s_nlf;
    static PyObject *__pyx_n_s_nncName;
    static PyObject *__pyx_n_s_nncNode;
    static PyObject *__pyx_kp_u_node_T_sNode;
    static PyObject *__pyx_kp_u_node_T_sNode2;
    static PyObject *__pyx_kp_u_node_T_sNode3;
    static PyObject *__pyx_kp_u_node_T_sOutput;
    static PyObject *__pyx_n_s_npConst;
    static PyObject *__pyx_n_s_npHidden;
    static PyObject *__pyx_n_s_npInput;
    static PyObject *__pyx_n_s_npOutState;
    static PyObject *__pyx_n_s_npOutput;
    static PyObject *__pyx_n_s_npState;
    static PyObject *__pyx_n_u_ntdefines;
    static PyObject *__pyx_n_u_nttoolbox;
    static PyObject *__pyx_n_u_nttype;
    static PyObject *__pyx_kp_u_organisation_is_free_provided_th;
    static PyObject *__pyx_kp_u_potential_Double;
    static PyObject *__pyx_n_s_proprietaryname;
    static PyObject *__pyx_n_s_range;
    static PyObject *__pyx_n_u_readfile;
    static PyObject *__pyx_kp_u_return_index_0_index_sINPUTS_NUL;
    static PyObject *__pyx_kp_u_return_index_0_index_sOUTPUTS_NU;
    static PyObject *__pyx_kp_u_sCONSTANTS_d;
    static PyObject *__pyx_kp_u_sDIMENSION_d;
    static PyObject *__pyx_kp_u_sDIM_d_d;
    static PyObject *__pyx_kp_u_sINPUTNAME;
    static PyObject *__pyx_kp_u_sINPUTS_d;
    static PyObject *__pyx_kp_u_sMATRIXES_d;
    static PyObject *__pyx_kp_u_sMATRIX_K_GRADIENT_J;
    static PyObject *__pyx_kp_u_sMEMSYN_d;
    static PyObject *__pyx_kp_u_sNODES_d;
    static PyObject *__pyx_kp_u_sORDER_d;
    static PyObject *__pyx_kp_u_sOUTPUTNAME;
    static PyObject *__pyx_kp_u_sOUTPUTS_d;
    static PyObject *__pyx_kp_u_sSTDTRDEV_d_g;
    static PyObject *__pyx_kp_u_sSTDTRDEV_g;
    static PyObject *__pyx_kp_u_sSTUDENT95_g;
    static PyObject *__pyx_kp_u_sSYNAPSES_d;
    static PyObject *__pyx_kp_u_sTRANSFERGRADIENT_INPUTS_OUTPUT;
    static PyObject *__pyx_kp_u_setup_cfg;
    static PyObject *__pyx_n_u_singleinterfacetraindll;
    static PyObject *__pyx_n_u_singletrainbase;
    static PyObject *__pyx_kp_u_smatrix_k_gradient_j;
    static PyObject *__pyx_kp_u_smatrix_mindex_i_j_gradient_j;
    static PyObject *__pyx_n_s_stAdvertDate;
    static PyObject *__pyx_n_s_stAdvertMachine;
    static PyObject *__pyx_n_s_stAdvertTime;
    static PyObject *__pyx_n_s_stAdvertUser;
    static PyObject *__pyx_n_s_stAlert;
    static PyObject *__pyx_n_s_stArrayMatrixesP;
    static PyObject *__pyx_n_s_stArrayOfDoubleDef;
    static PyObject *__pyx_n_s_stArrayRetroP;
    static PyObject *__pyx_n_s_stConstLinks;
    static PyObject *__pyx_n_s_stConstLinksF;
    static PyObject *__pyx_n_s_stConstMatrix;
    static PyObject *__pyx_n_s_stConstMatrixF;
    static PyObject *__pyx_n_s_stConstNodes;
    static PyObject *__pyx_n_s_stConstNodesF;
    static PyObject *__pyx_n_s_stConstSynapses;
    static PyObject *__pyx_n_s_stConstSynapsesF;
    static PyObject *__pyx_n_s_stDefineDimension;
    static PyObject *__pyx_n_s_stDefineDimn;
    static PyObject *__pyx_n_s_stDefineLinks;
    static PyObject *__pyx_n_s_stDefineMatrixes;
    static PyObject *__pyx_n_s_stDefineNodes;
    static PyObject *__pyx_n_s_stDefineRetros;
    static PyObject *__pyx_n_s_stDefineSynapses;
    static PyObject *__pyx_n_s_stGradientDeclF77;
    static PyObject *__pyx_n_s_stGradientDeclF90;
    static PyObject *__pyx_n_s_stGradientDeclF90S;
    static PyObject *__pyx_n_s_stGradientdS;
    static PyObject *__pyx_n_s_stGradientdSP;
    static PyObject *__pyx_n_s_stIncludeFile;
    static PyObject *__pyx_n_s_stIncludeFileP;
    static PyObject *__pyx_n_s_stInibackDeclF77;
    static PyObject *__pyx_n_s_stInibackDeclF90;
    static PyObject *__pyx_n_s_stInibackDeclF90S;
    static PyObject *__pyx_n_s_stLeverageDeclF77;
    static PyObject *__pyx_n_s_stLeverageDeclF90;
    static PyObject *__pyx_n_s_stLeverageDeclF90S;
    static PyObject *__pyx_n_s_stMemSyndS;
    static PyObject *__pyx_n_s_stMemSyndSF77;
    static PyObject *__pyx_n_s_stMemSyndSF90;
    static PyObject *__pyx_n_s_stMemSyndSP;
    static PyObject *__pyx_n_s_stNodedS;
    static PyObject *__pyx_n_s_stNodedS2;
    static PyObject *__pyx_n_s_stNodedS2F77;
    static PyObject *__pyx_n_s_stNodedS2F90;
    static PyObject *__pyx_n_s_stNodedS3;
    static PyObject *__pyx_n_s_stNodedS3F77;
    static PyObject *__pyx_n_s_stNodedS3F90;
    static PyObject *__pyx_n_s_stNodedSF77;
    static PyObject *__pyx_n_s_stNodedSF90;
    static PyObject *__pyx_n_s_stNodedSK;
    static PyObject *__pyx_n_s_stNodedSP;
    static PyObject *__pyx_n_s_stNodedSP2;
    static PyObject *__pyx_n_s_stNodedSP3;
    static PyObject *__pyx_n_s_stNodedSPK;
    static PyObject *__pyx_n_s_stPArrayOfDoubleDef;
    static PyObject *__pyx_n_s_stPotentialdSF77;
    static PyObject *__pyx_n_s_stPotentialdSF90;
    static PyObject *__pyx_n_s_stPotentialdSP;
    static PyObject *__pyx_n_s_stStatesDeclF77;
    static PyObject *__pyx_n_s_stStatesDeclF90;
    static PyObject *__pyx_n_s_stTypeMatrixes;
    static PyObject *__pyx_n_s_stTypeMinMax;
    static PyObject *__pyx_n_s_stTypeNodes;
    static PyObject *__pyx_n_s_stTypeOutput;
    static PyObject *__pyx_n_s_stTypeRetroP;
    static PyObject *__pyx_n_s_stTypeSynapse;
    static PyObject *__pyx_n_s_stXPotentialdSF77;
    static PyObject *__pyx_n_s_stXPotentialdSF90;
    static PyObject *__pyx_n_s_stXPotentialdSP;
    static PyObject *__pyx_n_s_stXdSF77;
    static PyObject *__pyx_n_s_stXdSF90;
    static PyObject *__pyx_n_s_st_InDev;
    static PyObject *__pyx_n_s_st_Juridic;
    static PyObject *__pyx_kp_u_static_char_names_sINPUTS;
    static PyObject *__pyx_kp_u_static_char_names_sOUTPUTS;
    static PyObject *__pyx_n_s_std_re;
    static PyObject *__pyx_kp_u_stddev_sOUTPUTS;
    static PyObject *__pyx_kp_u_stransfergradient_inputs_output;
    static PyObject *__pyx_kp_u_stransfergradient_inputs_output_2;
    static PyObject *__pyx_kp_u_stransfergradient_inputs_output_3;
    static PyObject *__pyx_kp_u_stransfergradient_inputs_output_4;
    static PyObject *__pyx_n_s_subpattern;
    static PyObject *__pyx_n_s_test;
    static PyObject *__pyx_n_u_train;
    static PyObject *__pyx_n_u_trainbase;
    static PyObject *__pyx_kp_u_tre_utilis_de_faon_permanente_su;
    static PyObject *__pyx_n_s_upper_re;
    static PyObject *__pyx_n_u_usebase;
    static PyObject *__pyx_n_s_useversion;
    static PyObject *__pyx_n_s_version;
    static PyObject *__pyx_kp_u_void_stransfer_double_double;
    static PyObject *__pyx_kp_u_void_stransfer_double_double_dou;
    static PyObject *__pyx_kp_u_void_stransfer_double_inputs_dou;
    static PyObject *__pyx_kp_u_void_stransfer_double_inputs_dou_2;
    static PyObject *__pyx_kp_u_void_stransfergradient_double_do;
    static PyObject *__pyx_kp_u_void_stransfergradient_double_do_2;
    static PyObject *__pyx_kp_u_void_stransfergradient_double_in;
    static PyObject *__pyx_kp_u_void_stransfergradient_double_in_2;
    static PyObject *__pyx_kp_u_void_stransferlev_double_double;
    static PyObject *__pyx_kp_u_void_stransferlev_double_double_2;
    static PyObject *__pyx_kp_u_void_stransferlev_double_inputs;
    static PyObject *__pyx_kp_u_void_stransferlev_double_weights;
    static PyObject *__pyx_n_s_vtMax;
    static PyObject *__pyx_n_s_vtMin;
    static PyObject *__pyx_n_s_vtStdDev;
    static PyObject *__pyx_n_s_vtWeight;
    static PyObject *__pyx_kp_u_weights_d;
    static PyObject *__pyx_kp_u_weights_s;
    static PyObject *__pyx_kp_u_weights_sSYNAPSES;
    static PyObject *__pyx_kp_u_weights_weightsredir_d;
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_6;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_17;
static PyObject *__pyx_int_21;
static PyObject *__pyx_int_26;
static PyObject *__pyx_int_33;
static PyObject *__pyx_int_93;
static PyObject *__pyx_int_256;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
/* Late includes */

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_codeconst(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_codeconst},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "codeconst",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_ADouble_Array_0_65535_Of_Double, __pyx_k_ADouble_Array_0_65535_Of_Double, sizeof(__pyx_k_ADouble_Array_0_65535_Of_Double), 0, 1, 0, 0},
  {&__pyx_kp_u_Array_0_sINPUTS_sOUTPUTS_1_of_do, __pyx_k_Array_0_sINPUTS_sOUTPUTS_1_of_do, sizeof(__pyx_k_Array_0_sINPUTS_sOUTPUTS_1_of_do), 0, 1, 0, 0},
  {&__pyx_kp_u_Array_0_sMATRIXES_1_of_double, __pyx_k_Array_0_sMATRIXES_1_of_double, sizeof(__pyx_k_Array_0_sMATRIXES_1_of_double), 0, 1, 0, 0},
  {&__pyx_kp_u_Array_0_sMEMSYN_1_of_double, __pyx_k_Array_0_sMEMSYN_1_of_double, sizeof(__pyx_k_Array_0_sMEMSYN_1_of_double), 0, 1, 0, 0},
  {&__pyx_kp_u_Array_0_sNODES_1_0_1_of_double, __pyx_k_Array_0_sNODES_1_0_1_of_double, sizeof(__pyx_k_Array_0_sNODES_1_0_1_of_double), 0, 1, 0, 0},
  {&__pyx_kp_u_Array_0_sNODES_1_0_2_of_double, __pyx_k_Array_0_sNODES_1_0_2_of_double, sizeof(__pyx_k_Array_0_sNODES_1_0_2_of_double), 0, 1, 0, 0},
  {&__pyx_kp_u_Array_0_sNODES_1_of_double, __pyx_k_Array_0_sNODES_1_of_double, sizeof(__pyx_k_Array_0_sNODES_1_of_double), 0, 1, 0, 0},
  {&__pyx_kp_u_Array_0_sOUTPUTS_1_of_double, __pyx_k_Array_0_sOUTPUTS_1_of_double, sizeof(__pyx_k_Array_0_sOUTPUTS_1_of_double), 0, 1, 0, 0},
  {&__pyx_kp_u_Array_0_sSYNAPSES_1_of_double, __pyx_k_Array_0_sSYNAPSES_1_of_double, sizeof(__pyx_k_Array_0_sSYNAPSES_1_of_double), 0, 1, 0, 0},
  {&__pyx_kp_u_CALL_sTRANSFERGRADIENT_INPUTS_OU, __pyx_k_CALL_sTRANSFERGRADIENT_INPUTS_OU, sizeof(__pyx_k_CALL_sTRANSFERGRADIENT_INPUTS_OU), 0, 1, 0, 0},
  {&__pyx_kp_u_CHARACTER_32_names_sINPUTS, __pyx_k_CHARACTER_32_names_sINPUTS, sizeof(__pyx_k_CHARACTER_32_names_sINPUTS), 0, 1, 0, 0},
  {&__pyx_kp_u_CHARACTER_32_names_sOUTPUTS, __pyx_k_CHARACTER_32_names_sOUTPUTS, sizeof(__pyx_k_CHARACTER_32_names_sOUTPUTS), 0, 1, 0, 0},
  {&__pyx_kp_u_CHARACTER_LEN_32_DIMENSION_sINPU, __pyx_k_CHARACTER_LEN_32_DIMENSION_sINPU, sizeof(__pyx_k_CHARACTER_LEN_32_DIMENSION_sINPU), 0, 1, 0, 0},
  {&__pyx_kp_u_CHARACTER_LEN_32_DIMENSION_sOUTP, __pyx_k_CHARACTER_LEN_32_DIMENSION_sOUTP, sizeof(__pyx_k_CHARACTER_LEN_32_DIMENSION_sOUTP), 0, 1, 0, 0},
  {&__pyx_kp_u_Ce_code_a_t_cre_automatiquement, __pyx_k_Ce_code_a_t_cre_automatiquement, sizeof(__pyx_k_Ce_code_a_t_cre_automatiquement), 0, 1, 0, 0},
  {&__pyx_kp_u_Ce_code_est_destin_uniquement_au, __pyx_k_Ce_code_est_destin_uniquement_au, sizeof(__pyx_k_Ce_code_est_destin_uniquement_au), 0, 1, 0, 0},
  {&__pyx_kp_u_Computer_s, __pyx_k_Computer_s, sizeof(__pyx_k_Computer_s), 0, 1, 0, 0},
  {&__pyx_kp_u_DATA_GRADIENT_INIBACK_LEVERAGE, __pyx_k_DATA_GRADIENT_INIBACK_LEVERAGE, sizeof(__pyx_k_DATA_GRADIENT_INIBACK_LEVERAGE), 0, 1, 0, 0},
  {&__pyx_kp_u_DATA_MEMSYN_sMEMSYN_1, __pyx_k_DATA_MEMSYN_sMEMSYN_1, sizeof(__pyx_k_DATA_MEMSYN_sMEMSYN_1), 0, 1, 0, 0},
  {&__pyx_n_s_DEq, __pyx_k_DEq, sizeof(__pyx_k_DEq), 0, 0, 1, 1},
  {&__pyx_kp_u_DIMENSION_REAL_8_GRADIENT_sSYNAP, __pyx_k_DIMENSION_REAL_8_GRADIENT_sSYNAP, sizeof(__pyx_k_DIMENSION_REAL_8_GRADIENT_sSYNAP), 0, 1, 0, 0},
  {&__pyx_kp_u_DIMENSION_REAL_8_INIBACK_sOUTPUT, __pyx_k_DIMENSION_REAL_8_INIBACK_sOUTPUT, sizeof(__pyx_k_DIMENSION_REAL_8_INIBACK_sOUTPUT), 0, 1, 0, 0},
  {&__pyx_kp_u_DIMENSION_REAL_8_INPUTS_sINPUTS, __pyx_k_DIMENSION_REAL_8_INPUTS_sINPUTS, sizeof(__pyx_k_DIMENSION_REAL_8_INPUTS_sINPUTS), 0, 1, 0, 0},
  {&__pyx_kp_u_DIMENSION_REAL_8_LEVERAGE_sOUTPU, __pyx_k_DIMENSION_REAL_8_LEVERAGE_sOUTPU, sizeof(__pyx_k_DIMENSION_REAL_8_LEVERAGE_sOUTPU), 0, 1, 0, 0},
  {&__pyx_kp_u_DIMENSION_REAL_8_MEMSYN_sMEMSYN, __pyx_k_DIMENSION_REAL_8_MEMSYN_sMEMSYN, sizeof(__pyx_k_DIMENSION_REAL_8_MEMSYN_sMEMSYN), 0, 1, 0, 0},
  {&__pyx_kp_u_DIMENSION_REAL_8_NODE_2_sNODES, __pyx_k_DIMENSION_REAL_8_NODE_2_sNODES, sizeof(__pyx_k_DIMENSION_REAL_8_NODE_2_sNODES), 0, 1, 0, 0},
  {&__pyx_kp_u_DIMENSION_REAL_8_NODE_3_sNODES, __pyx_k_DIMENSION_REAL_8_NODE_3_sNODES, sizeof(__pyx_k_DIMENSION_REAL_8_NODE_3_sNODES), 0, 1, 0, 0},
  {&__pyx_kp_u_DIMENSION_REAL_8_NODE_sNODES, __pyx_k_DIMENSION_REAL_8_NODE_sNODES, sizeof(__pyx_k_DIMENSION_REAL_8_NODE_sNODES), 0, 1, 0, 0},
  {&__pyx_kp_u_DIMENSION_REAL_8_OUTPUTS_sOUTPUT, __pyx_k_DIMENSION_REAL_8_OUTPUTS_sOUTPUT, sizeof(__pyx_k_DIMENSION_REAL_8_OUTPUTS_sOUTPUT), 0, 1, 0, 0},
  {&__pyx_kp_u_DIMENSION_REAL_8_STATES_sORDER, __pyx_k_DIMENSION_REAL_8_STATES_sORDER, sizeof(__pyx_k_DIMENSION_REAL_8_STATES_sORDER), 0, 1, 0, 0},
  {&__pyx_kp_u_DIMENSION_REAL_8_s_sMATRIXES, __pyx_k_DIMENSION_REAL_8_s_sMATRIXES, sizeof(__pyx_k_DIMENSION_REAL_8_s_sMATRIXES), 0, 1, 0, 0},
  {&__pyx_kp_u_DIMENSION_REAL_8_s_sSYNAPSES, __pyx_k_DIMENSION_REAL_8_s_sSYNAPSES, sizeof(__pyx_k_DIMENSION_REAL_8_s_sSYNAPSES), 0, 1, 0, 0},
  {&__pyx_n_u_DLM, __pyx_k_DLM, sizeof(__pyx_k_DLM), 0, 1, 0, 1},
  {&__pyx_kp_u_DO_140_I_1_sSYNAPSES, __pyx_k_DO_140_I_1_sSYNAPSES, sizeof(__pyx_k_DO_140_I_1_sSYNAPSES), 0, 1, 0, 0},
  {&__pyx_kp_u_DO_140_I_2_sOUTPUTS, __pyx_k_DO_140_I_2_sOUTPUTS, sizeof(__pyx_k_DO_140_I_2_sOUTPUTS), 0, 1, 0, 0},
  {&__pyx_kp_u_DO_150_J_1_sSYNAPSES, __pyx_k_DO_150_J_1_sSYNAPSES, sizeof(__pyx_k_DO_150_J_1_sSYNAPSES), 0, 1, 0, 0},
  {&__pyx_kp_u_DO_I_1_sSYNAPSES, __pyx_k_DO_I_1_sSYNAPSES, sizeof(__pyx_k_DO_I_1_sSYNAPSES), 0, 1, 0, 0},
  {&__pyx_kp_u_DO_I_2_sOUTPUTS, __pyx_k_DO_I_2_sOUTPUTS, sizeof(__pyx_k_DO_I_2_sOUTPUTS), 0, 1, 0, 0},
  {&__pyx_kp_u_DO_J_1_sSYNAPSES, __pyx_k_DO_J_1_sSYNAPSES, sizeof(__pyx_k_DO_J_1_sSYNAPSES), 0, 1, 0, 0},
  {&__pyx_kp_u_Date_s, __pyx_k_Date_s, sizeof(__pyx_k_Date_s), 0, 1, 0, 0},
  {&__pyx_n_s_DefaultDecimalSeparator, __pyx_k_DefaultDecimalSeparator, sizeof(__pyx_k_DefaultDecimalSeparator), 0, 0, 1, 1},
  {&__pyx_n_u_Desc, __pyx_k_Desc, sizeof(__pyx_k_Desc), 0, 1, 0, 1},
  {&__pyx_n_s_Eq, __pyx_k_Eq, sizeof(__pyx_k_Eq), 0, 0, 1, 1},
  {&__pyx_kp_u_FUNCTION_sINPUTNAME_INDEX, __pyx_k_FUNCTION_sINPUTNAME_INDEX, sizeof(__pyx_k_FUNCTION_sINPUTNAME_INDEX), 0, 1, 0, 0},
  {&__pyx_kp_u_FUNCTION_sOUTPUTNAME_INDEX, __pyx_k_FUNCTION_sOUTPUTNAME_INDEX, sizeof(__pyx_k_FUNCTION_sOUTPUTNAME_INDEX), 0, 1, 0, 0},
  {&__pyx_kp_u_FUNCTION_sinputname_index_intege, __pyx_k_FUNCTION_sinputname_index_intege, sizeof(__pyx_k_FUNCTION_sinputname_index_intege), 0, 1, 0, 0},
  {&__pyx_kp_u_FUNCTION_soutputname_index_integ, __pyx_k_FUNCTION_soutputname_index_integ, sizeof(__pyx_k_FUNCTION_soutputname_index_integ), 0, 1, 0, 0},
  {&__pyx_n_s_GetComputerName, __pyx_k_GetComputerName, sizeof(__pyx_k_GetComputerName), 0, 0, 1, 1},
  {&__pyx_n_s_GetFileName, __pyx_k_GetFileName, sizeof(__pyx_k_GetFileName), 0, 0, 1, 1},
  {&__pyx_n_s_GetLicenseNb, __pyx_k_GetLicenseNb, sizeof(__pyx_k_GetLicenseNb), 0, 0, 1, 1},
  {&__pyx_n_s_GetStudent95, __pyx_k_GetStudent95, sizeof(__pyx_k_GetStudent95), 0, 0, 1, 1},
  {&__pyx_n_s_GetUseCount, __pyx_k_GetUseCount, sizeof(__pyx_k_GetUseCount), 0, 0, 1, 1},
  {&__pyx_n_s_GetUserName, __pyx_k_GetUserName, sizeof(__pyx_k_GetUserName), 0, 0, 1, 1},
  {&__pyx_n_u_Gradient, __pyx_k_Gradient, sizeof(__pyx_k_Gradient), 0, 1, 0, 1},
  {&__pyx_kp_u_INTEGER_PARAMETER, __pyx_k_INTEGER_PARAMETER, sizeof(__pyx_k_INTEGER_PARAMETER), 0, 1, 0, 0},
  {&__pyx_kp_u_I_s, __pyx_k_I_s, sizeof(__pyx_k_I_s), 0, 1, 0, 0},
  {&__pyx_n_s_KT_AFTER, __pyx_k_KT_AFTER, sizeof(__pyx_k_KT_AFTER), 0, 0, 1, 1},
  {&__pyx_n_s_KT_BEFORE, __pyx_k_KT_BEFORE, sizeof(__pyx_k_KT_BEFORE), 0, 0, 1, 1},
  {&__pyx_n_u_LartenNetral_JLP_04_05_47, __pyx_k_LartenNetral_JLP_04_05_47, sizeof(__pyx_k_LartenNetral_JLP_04_05_47), 0, 1, 0, 1},
  {&__pyx_n_u_MATRIX, __pyx_k_MATRIX, sizeof(__pyx_k_MATRIX), 0, 1, 0, 1},
  {&__pyx_n_u_MAX, __pyx_k_MAX, sizeof(__pyx_k_MAX), 0, 1, 0, 1},
  {&__pyx_n_s_MEq, __pyx_k_MEq, sizeof(__pyx_k_MEq), 0, 0, 1, 1},
  {&__pyx_n_u_MIN, __pyx_k_MIN, sizeof(__pyx_k_MIN), 0, 1, 0, 1},
  {&__pyx_n_u_Make, __pyx_k_Make, sizeof(__pyx_k_Make), 0, 1, 0, 1},
  {&__pyx_n_u_MakeDLM, __pyx_k_MakeDLM, sizeof(__pyx_k_MakeDLM), 0, 1, 0, 1},
  {&__pyx_kp_u_NETRAL_ne_peut_pas_tre_tenu_pour, __pyx_k_NETRAL_ne_peut_pas_tre_tenu_pour, sizeof(__pyx_k_NETRAL_ne_peut_pas_tre_tenu_pour), 0, 1, 0, 0},
  {&__pyx_kp_u_Netral_is_not_responsible_of_any, __pyx_k_Netral_is_not_responsible_of_any, sizeof(__pyx_k_Netral_is_not_responsible_of_any), 0, 1, 0, 0},
  {&__pyx_kp_u_PADouble_ADouble, __pyx_k_PADouble_ADouble, sizeof(__pyx_k_PADouble_ADouble), 0, 1, 0, 0},
  {&__pyx_n_s_PEq, __pyx_k_PEq, sizeof(__pyx_k_PEq), 0, 0, 1, 1},
  {&__pyx_kp_u_PROCEDURE_stranserlev_inputs_out, __pyx_k_PROCEDURE_stranserlev_inputs_out, sizeof(__pyx_k_PROCEDURE_stranserlev_inputs_out), 0, 1, 0, 0},
  {&__pyx_kp_u_PROCEDURE_stransfer_inputs_outpu, __pyx_k_PROCEDURE_stransfer_inputs_outpu, sizeof(__pyx_k_PROCEDURE_stransfer_inputs_outpu), 0, 1, 0, 0},
  {&__pyx_kp_u_PROCEDURE_stransfer_inputs_outpu_2, __pyx_k_PROCEDURE_stransfer_inputs_outpu_2, sizeof(__pyx_k_PROCEDURE_stransfer_inputs_outpu_2), 0, 1, 0, 0},
  {&__pyx_kp_u_PROCEDURE_stransfergradient_inpu, __pyx_k_PROCEDURE_stransfergradient_inpu, sizeof(__pyx_k_PROCEDURE_stransfergradient_inpu), 0, 1, 0, 0},
  {&__pyx_kp_u_PROCEDURE_stransfergradient_inpu_2, __pyx_k_PROCEDURE_stransfergradient_inpu_2, sizeof(__pyx_k_PROCEDURE_stransfergradient_inpu_2), 0, 1, 0, 0},
  {&__pyx_kp_u_PROCEDURE_stransferlev_inputs_ou, __pyx_k_PROCEDURE_stransferlev_inputs_ou, sizeof(__pyx_k_PROCEDURE_stransferlev_inputs_ou), 0, 1, 0, 0},
  {&__pyx_kp_u_P_lower_l_l, __pyx_k_P_lower_l_l, sizeof(__pyx_k_P_lower_l_l), 0, 1, 0, 0},
  {&__pyx_kp_u_P_std, __pyx_k_P_std, sizeof(__pyx_k_P_std), 0, 1, 0, 0},
  {&__pyx_kp_u_P_upper_U_U, __pyx_k_P_upper_U_U, sizeof(__pyx_k_P_upper_U_U), 0, 1, 0, 0},
  {&__pyx_kp_u_Please_check_name_of_included_he, __pyx_k_Please_check_name_of_included_he, sizeof(__pyx_k_Please_check_name_of_included_he), 0, 1, 0, 0},
  {&__pyx_n_s_PyCodeList, __pyx_k_PyCodeList, sizeof(__pyx_k_PyCodeList), 0, 0, 1, 1},
  {&__pyx_n_s_PyCodeList0, __pyx_k_PyCodeList0, sizeof(__pyx_k_PyCodeList0), 0, 0, 1, 1},
  {&__pyx_n_s_PyCodeList2, __pyx_k_PyCodeList2, sizeof(__pyx_k_PyCodeList2), 0, 0, 1, 1},
  {&__pyx_n_s_PyCodeListFull, __pyx_k_PyCodeListFull, sizeof(__pyx_k_PyCodeListFull), 0, 0, 1, 1},
  {&__pyx_n_s_PyCodeListShort, __pyx_k_PyCodeListShort, sizeof(__pyx_k_PyCodeListShort), 0, 0, 1, 1},
  {&__pyx_n_s_PyCodeMulti, __pyx_k_PyCodeMulti, sizeof(__pyx_k_PyCodeMulti), 0, 0, 1, 1},
  {&__pyx_n_s_PyCodeMultiFull, __pyx_k_PyCodeMultiFull, sizeof(__pyx_k_PyCodeMultiFull), 0, 0, 1, 1},
  {&__pyx_n_s_PyCodeSingleTrain, __pyx_k_PyCodeSingleTrain, sizeof(__pyx_k_PyCodeSingleTrain), 0, 0, 1, 1},
  {&__pyx_n_u_PyInterfaceCMulti, __pyx_k_PyInterfaceCMulti, sizeof(__pyx_k_PyInterfaceCMulti), 0, 1, 0, 1},
  {&__pyx_kp_u_REAL_8_DIMENSION_2_sNODES_NODE, __pyx_k_REAL_8_DIMENSION_2_sNODES_NODE, sizeof(__pyx_k_REAL_8_DIMENSION_2_sNODES_NODE), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_DIMENSION_3_sNODES_NODE, __pyx_k_REAL_8_DIMENSION_3_sNODES_NODE, sizeof(__pyx_k_REAL_8_DIMENSION_3_sNODES_NODE), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_DIMENSION_sINPUTS_INPUTS, __pyx_k_REAL_8_DIMENSION_sINPUTS_INPUTS, sizeof(__pyx_k_REAL_8_DIMENSION_sINPUTS_INPUTS), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_DIMENSION_sMATRIXES_s, __pyx_k_REAL_8_DIMENSION_sMATRIXES_s, sizeof(__pyx_k_REAL_8_DIMENSION_sMATRIXES_s), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_DIMENSION_sMEMSYN_MEMSYN, __pyx_k_REAL_8_DIMENSION_sMEMSYN_MEMSYN, sizeof(__pyx_k_REAL_8_DIMENSION_sMEMSYN_MEMSYN), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_DIMENSION_sNODES_NODE, __pyx_k_REAL_8_DIMENSION_sNODES_NODE, sizeof(__pyx_k_REAL_8_DIMENSION_sNODES_NODE), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_DIMENSION_sORDER_STATES, __pyx_k_REAL_8_DIMENSION_sORDER_STATES, sizeof(__pyx_k_REAL_8_DIMENSION_sORDER_STATES), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_INIBAC, __pyx_k_REAL_8_DIMENSION_sOUTPUTS_INIBAC, sizeof(__pyx_k_REAL_8_DIMENSION_sOUTPUTS_INIBAC), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_INIBAC_2, __pyx_k_REAL_8_DIMENSION_sOUTPUTS_INIBAC_2, sizeof(__pyx_k_REAL_8_DIMENSION_sOUTPUTS_INIBAC_2), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_LEVERA, __pyx_k_REAL_8_DIMENSION_sOUTPUTS_LEVERA, sizeof(__pyx_k_REAL_8_DIMENSION_sOUTPUTS_LEVERA), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_LEVERA_2, __pyx_k_REAL_8_DIMENSION_sOUTPUTS_LEVERA_2, sizeof(__pyx_k_REAL_8_DIMENSION_sOUTPUTS_LEVERA_2), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_OUTPUT, __pyx_k_REAL_8_DIMENSION_sOUTPUTS_OUTPUT, sizeof(__pyx_k_REAL_8_DIMENSION_sOUTPUTS_OUTPUT), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_DIMENSION_sSYNAPSES_GRADI, __pyx_k_REAL_8_DIMENSION_sSYNAPSES_GRADI, sizeof(__pyx_k_REAL_8_DIMENSION_sSYNAPSES_GRADI), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_DIMENSION_sSYNAPSES_GRADI_2, __pyx_k_REAL_8_DIMENSION_sSYNAPSES_GRADI_2, sizeof(__pyx_k_REAL_8_DIMENSION_sSYNAPSES_GRADI_2), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_DIMENSION_sSYNAPSES_s, __pyx_k_REAL_8_DIMENSION_sSYNAPSES_s, sizeof(__pyx_k_REAL_8_DIMENSION_sSYNAPSES_s), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_POTENTIAL, __pyx_k_REAL_8_POTENTIAL, sizeof(__pyx_k_REAL_8_POTENTIAL), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_X, __pyx_k_REAL_8_X, sizeof(__pyx_k_REAL_8_X), 0, 1, 0, 0},
  {&__pyx_kp_u_REAL_8_X_POTENTIAL, __pyx_k_REAL_8_X_POTENTIAL, sizeof(__pyx_k_REAL_8_X_POTENTIAL), 0, 1, 0, 0},
  {&__pyx_n_u_STDTRDEV, __pyx_k_STDTRDEV, sizeof(__pyx_k_STDTRDEV), 0, 1, 0, 1},
  {&__pyx_kp_u_SUBROUTINE_sTRANSFERGRADIENT_INP, __pyx_k_SUBROUTINE_sTRANSFERGRADIENT_INP, sizeof(__pyx_k_SUBROUTINE_sTRANSFERGRADIENT_INP), 0, 1, 0, 0},
  {&__pyx_kp_u_SUBROUTINE_sTRANSFERGRADIENT_INP_2, __pyx_k_SUBROUTINE_sTRANSFERGRADIENT_INP_2, sizeof(__pyx_k_SUBROUTINE_sTRANSFERGRADIENT_INP_2), 0, 1, 0, 0},
  {&__pyx_kp_u_SUBROUTINE_sTRANSFERLEV_INPUTS_O, __pyx_k_SUBROUTINE_sTRANSFERLEV_INPUTS_O, sizeof(__pyx_k_SUBROUTINE_sTRANSFERLEV_INPUTS_O), 0, 1, 0, 0},
  {&__pyx_kp_u_SUBROUTINE_sTRANSFERLEV_INPUTS_O_2, __pyx_k_SUBROUTINE_sTRANSFERLEV_INPUTS_O_2, sizeof(__pyx_k_SUBROUTINE_sTRANSFERLEV_INPUTS_O_2), 0, 1, 0, 0},
  {&__pyx_kp_u_SUBROUTINE_sTRANSFER_INPUTS_OUTP, __pyx_k_SUBROUTINE_sTRANSFER_INPUTS_OUTP, sizeof(__pyx_k_SUBROUTINE_sTRANSFER_INPUTS_OUTP), 0, 1, 0, 0},
  {&__pyx_kp_u_SUBROUTINE_sTRANSFER_INPUTS_OUTP_2, __pyx_k_SUBROUTINE_sTRANSFER_INPUTS_OUTP_2, sizeof(__pyx_k_SUBROUTINE_sTRANSFER_INPUTS_OUTP_2), 0, 1, 0, 0},
  {&__pyx_n_u_SetupPy, __pyx_k_SetupPy, sizeof(__pyx_k_SetupPy), 0, 1, 0, 1},
  {&__pyx_n_u_Setupcfg, __pyx_k_Setupcfg, sizeof(__pyx_k_Setupcfg), 0, 1, 0, 1},
  {&__pyx_n_s_StAdvert, __pyx_k_StAdvert, sizeof(__pyx_k_StAdvert), 0, 0, 1, 1},
  {&__pyx_n_s_StAdvertAdv, __pyx_k_StAdvertAdv, sizeof(__pyx_k_StAdvertAdv), 0, 0, 1, 1},
  {&__pyx_n_s_StArrayInOutP, __pyx_k_StArrayInOutP, sizeof(__pyx_k_StArrayInOutP), 0, 0, 1, 1},
  {&__pyx_n_s_StArrayMatrixF77, __pyx_k_StArrayMatrixF77, sizeof(__pyx_k_StArrayMatrixF77), 0, 0, 1, 1},
  {&__pyx_n_s_StArrayMatrixF90, __pyx_k_StArrayMatrixF90, sizeof(__pyx_k_StArrayMatrixF90), 0, 0, 1, 1},
  {&__pyx_n_s_StArrayNodes2P, __pyx_k_StArrayNodes2P, sizeof(__pyx_k_StArrayNodes2P), 0, 0, 1, 1},
  {&__pyx_n_s_StArrayNodes3P, __pyx_k_StArrayNodes3P, sizeof(__pyx_k_StArrayNodes3P), 0, 0, 1, 1},
  {&__pyx_n_s_StArrayNodesP, __pyx_k_StArrayNodesP, sizeof(__pyx_k_StArrayNodesP), 0, 0, 1, 1},
  {&__pyx_n_s_StArrayOutputsP, __pyx_k_StArrayOutputsP, sizeof(__pyx_k_StArrayOutputsP), 0, 0, 1, 1},
  {&__pyx_n_s_StArraySynapsesF77, __pyx_k_StArraySynapsesF77, sizeof(__pyx_k_StArraySynapsesF77), 0, 0, 1, 1},
  {&__pyx_n_s_StArraySynapsesF90, __pyx_k_StArraySynapsesF90, sizeof(__pyx_k_StArraySynapsesF90), 0, 0, 1, 1},
  {&__pyx_n_s_StArraySynapsesP, __pyx_k_StArraySynapsesP, sizeof(__pyx_k_StArraySynapsesP), 0, 0, 1, 1},
  {&__pyx_n_s_StConstConstants, __pyx_k_StConstConstants, sizeof(__pyx_k_StConstConstants), 0, 0, 1, 1},
  {&__pyx_n_s_StConstConstantsF, __pyx_k_StConstConstantsF, sizeof(__pyx_k_StConstConstantsF), 0, 0, 1, 1},
  {&__pyx_n_s_StConstDimension, __pyx_k_StConstDimension, sizeof(__pyx_k_StConstDimension), 0, 0, 1, 1},
  {&__pyx_n_s_StConstDimn, __pyx_k_StConstDimn, sizeof(__pyx_k_StConstDimn), 0, 0, 1, 1},
  {&__pyx_n_s_StConstInputs, __pyx_k_StConstInputs, sizeof(__pyx_k_StConstInputs), 0, 0, 1, 1},
  {&__pyx_n_s_StConstInputsF, __pyx_k_StConstInputsF, sizeof(__pyx_k_StConstInputsF), 0, 0, 1, 1},
  {&__pyx_n_s_StConstOrder, __pyx_k_StConstOrder, sizeof(__pyx_k_StConstOrder), 0, 0, 1, 1},
  {&__pyx_n_s_StConstOrderF, __pyx_k_StConstOrderF, sizeof(__pyx_k_StConstOrderF), 0, 0, 1, 1},
  {&__pyx_n_s_StConstOutputs, __pyx_k_StConstOutputs, sizeof(__pyx_k_StConstOutputs), 0, 0, 1, 1},
  {&__pyx_n_s_StConstOutputsF, __pyx_k_StConstOutputsF, sizeof(__pyx_k_StConstOutputsF), 0, 0, 1, 1},
  {&__pyx_n_s_StConstStdDev, __pyx_k_StConstStdDev, sizeof(__pyx_k_StConstStdDev), 0, 0, 1, 1},
  {&__pyx_n_s_StConstStdDev1, __pyx_k_StConstStdDev1, sizeof(__pyx_k_StConstStdDev1), 0, 0, 1, 1},
  {&__pyx_n_s_StConstStud95, __pyx_k_StConstStud95, sizeof(__pyx_k_StConstStud95), 0, 0, 1, 1},
  {&__pyx_n_s_StDataGrIniLevF77, __pyx_k_StDataGrIniLevF77, sizeof(__pyx_k_StDataGrIniLevF77), 0, 0, 1, 1},
  {&__pyx_n_s_StDataMemSynF77, __pyx_k_StDataMemSynF77, sizeof(__pyx_k_StDataMemSynF77), 0, 0, 1, 1},
  {&__pyx_n_s_StDefineConstants, __pyx_k_StDefineConstants, sizeof(__pyx_k_StDefineConstants), 0, 0, 1, 1},
  {&__pyx_n_s_StDefineInputs, __pyx_k_StDefineInputs, sizeof(__pyx_k_StDefineInputs), 0, 0, 1, 1},
  {&__pyx_n_s_StDefineMIndex, __pyx_k_StDefineMIndex, sizeof(__pyx_k_StDefineMIndex), 0, 0, 1, 1},
  {&__pyx_n_s_StDefineOrder, __pyx_k_StDefineOrder, sizeof(__pyx_k_StDefineOrder), 0, 0, 1, 1},
  {&__pyx_n_s_StDefineOutputs, __pyx_k_StDefineOutputs, sizeof(__pyx_k_StDefineOutputs), 0, 0, 1, 1},
  {&__pyx_n_s_StDefineTransfer, __pyx_k_StDefineTransfer, sizeof(__pyx_k_StDefineTransfer), 0, 0, 1, 1},
  {&__pyx_n_s_StDefineTransferGr, __pyx_k_StDefineTransferGr, sizeof(__pyx_k_StDefineTransferGr), 0, 0, 1, 1},
  {&__pyx_n_s_StDefineTransferLev, __pyx_k_StDefineTransferLev, sizeof(__pyx_k_StDefineTransferLev), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllInitGradient, __pyx_k_StForAllInitGradient, sizeof(__pyx_k_StForAllInitGradient), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllInitGradientP, __pyx_k_StForAllInitGradientP, sizeof(__pyx_k_StForAllInitGradientP), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllInitIniBack, __pyx_k_StForAllInitIniBack, sizeof(__pyx_k_StForAllInitIniBack), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllInitIniBackP, __pyx_k_StForAllInitIniBackP, sizeof(__pyx_k_StForAllInitIniBackP), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllInitLeverage, __pyx_k_StForAllInitLeverage, sizeof(__pyx_k_StForAllInitLeverage), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllInitLeverageP, __pyx_k_StForAllInitLeverageP, sizeof(__pyx_k_StForAllInitLeverageP), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllInitMemSyn, __pyx_k_StForAllInitMemSyn, sizeof(__pyx_k_StForAllInitMemSyn), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllInitMemSynP, __pyx_k_StForAllInitMemSynP, sizeof(__pyx_k_StForAllInitMemSynP), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllOutputs, __pyx_k_StForAllOutputs, sizeof(__pyx_k_StForAllOutputs), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllOutputsF77, __pyx_k_StForAllOutputsF77, sizeof(__pyx_k_StForAllOutputsF77), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllOutputsF90, __pyx_k_StForAllOutputsF90, sizeof(__pyx_k_StForAllOutputsF90), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllOutputsP, __pyx_k_StForAllOutputsP, sizeof(__pyx_k_StForAllOutputsP), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllSynapses, __pyx_k_StForAllSynapses, sizeof(__pyx_k_StForAllSynapses), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllSynapsesF77, __pyx_k_StForAllSynapsesF77, sizeof(__pyx_k_StForAllSynapsesF77), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllSynapsesF90, __pyx_k_StForAllSynapsesF90, sizeof(__pyx_k_StForAllSynapsesF90), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllSynapsesP, __pyx_k_StForAllSynapsesP, sizeof(__pyx_k_StForAllSynapsesP), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllSynapsesj, __pyx_k_StForAllSynapsesj, sizeof(__pyx_k_StForAllSynapsesj), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllSynapsesjF77, __pyx_k_StForAllSynapsesjF77, sizeof(__pyx_k_StForAllSynapsesjF77), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllSynapsesjF90, __pyx_k_StForAllSynapsesjF90, sizeof(__pyx_k_StForAllSynapsesjF90), 0, 0, 1, 1},
  {&__pyx_n_s_StForAllSynapsesjP, __pyx_k_StForAllSynapsesjP, sizeof(__pyx_k_StForAllSynapsesjP), 0, 0, 1, 1},
  {&__pyx_n_s_StIfNDefTransfer, __pyx_k_StIfNDefTransfer, sizeof(__pyx_k_StIfNDefTransfer), 0, 0, 1, 1},
  {&__pyx_n_s_StIfNDefTransferGr, __pyx_k_StIfNDefTransferGr, sizeof(__pyx_k_StIfNDefTransferGr), 0, 0, 1, 1},
  {&__pyx_n_s_StIfNDefTransferLev, __pyx_k_StIfNDefTransferLev, sizeof(__pyx_k_StIfNDefTransferLev), 0, 0, 1, 1},
  {&__pyx_n_s_StIniBackdS, __pyx_k_StIniBackdS, sizeof(__pyx_k_StIniBackdS), 0, 0, 1, 1},
  {&__pyx_n_s_StIniBackdSP, __pyx_k_StIniBackdSP, sizeof(__pyx_k_StIniBackdSP), 0, 0, 1, 1},
  {&__pyx_n_s_StInputName, __pyx_k_StInputName, sizeof(__pyx_k_StInputName), 0, 0, 1, 1},
  {&__pyx_n_s_StInputNameDecl, __pyx_k_StInputNameDecl, sizeof(__pyx_k_StInputNameDecl), 0, 0, 1, 1},
  {&__pyx_n_s_StInputNameDeclF77, __pyx_k_StInputNameDeclF77, sizeof(__pyx_k_StInputNameDeclF77), 0, 0, 1, 1},
  {&__pyx_n_s_StInputNameDeclF90, __pyx_k_StInputNameDeclF90, sizeof(__pyx_k_StInputNameDeclF90), 0, 0, 1, 1},
  {&__pyx_n_s_StInputNameDeclP, __pyx_k_StInputNameDeclP, sizeof(__pyx_k_StInputNameDeclP), 0, 0, 1, 1},
  {&__pyx_n_s_StInputNameF, __pyx_k_StInputNameF, sizeof(__pyx_k_StInputNameF), 0, 0, 1, 1},
  {&__pyx_n_s_StInputNameFU, __pyx_k_StInputNameFU, sizeof(__pyx_k_StInputNameFU), 0, 0, 1, 1},
  {&__pyx_n_s_StInputNameP, __pyx_k_StInputNameP, sizeof(__pyx_k_StInputNameP), 0, 0, 1, 1},
  {&__pyx_n_s_StInputNameProto, __pyx_k_StInputNameProto, sizeof(__pyx_k_StInputNameProto), 0, 0, 1, 1},
  {&__pyx_n_s_StInputsDeclF77, __pyx_k_StInputsDeclF77, sizeof(__pyx_k_StInputsDeclF77), 0, 0, 1, 1},
  {&__pyx_n_s_StInputsDeclF90, __pyx_k_StInputsDeclF90, sizeof(__pyx_k_StInputsDeclF90), 0, 0, 1, 1},
  {&__pyx_n_s_StIntParam, __pyx_k_StIntParam, sizeof(__pyx_k_StIntParam), 0, 0, 1, 1},
  {&__pyx_n_s_StMatIniS, __pyx_k_StMatIniS, sizeof(__pyx_k_StMatIniS), 0, 0, 1, 1},
  {&__pyx_n_s_StMatIniSP, __pyx_k_StMatIniSP, sizeof(__pyx_k_StMatIniSP), 0, 0, 1, 1},
  {&__pyx_n_s_StMatrixCompute, __pyx_k_StMatrixCompute, sizeof(__pyx_k_StMatrixCompute), 0, 0, 1, 1},
  {&__pyx_n_s_StMatrixComputeF, __pyx_k_StMatrixComputeF, sizeof(__pyx_k_StMatrixComputeF), 0, 0, 1, 1},
  {&__pyx_n_s_StMatrixComputeP, __pyx_k_StMatrixComputeP, sizeof(__pyx_k_StMatrixComputeP), 0, 0, 1, 1},
  {&__pyx_n_s_StMatrixIni, __pyx_k_StMatrixIni, sizeof(__pyx_k_StMatrixIni), 0, 0, 1, 1},
  {&__pyx_n_s_StMatrixIniF, __pyx_k_StMatrixIniF, sizeof(__pyx_k_StMatrixIniF), 0, 0, 1, 1},
  {&__pyx_n_s_StMaxIniS, __pyx_k_StMaxIniS, sizeof(__pyx_k_StMaxIniS), 0, 0, 1, 1},
  {&__pyx_n_s_StMaxIniSF, __pyx_k_StMaxIniSF, sizeof(__pyx_k_StMaxIniSF), 0, 0, 1, 1},
  {&__pyx_n_s_StMaxIniSP, __pyx_k_StMaxIniSP, sizeof(__pyx_k_StMaxIniSP), 0, 0, 1, 1},
  {&__pyx_n_s_StMinIniS, __pyx_k_StMinIniS, sizeof(__pyx_k_StMinIniS), 0, 0, 1, 1},
  {&__pyx_n_s_StMinIniSF, __pyx_k_StMinIniSF, sizeof(__pyx_k_StMinIniSF), 0, 0, 1, 1},
  {&__pyx_n_s_StMinIniSP, __pyx_k_StMinIniSP, sizeof(__pyx_k_StMinIniSP), 0, 0, 1, 1},
  {&__pyx_n_s_StOutPutsDeclF77, __pyx_k_StOutPutsDeclF77, sizeof(__pyx_k_StOutPutsDeclF77), 0, 0, 1, 1},
  {&__pyx_n_s_StOutPutsDeclF90, __pyx_k_StOutPutsDeclF90, sizeof(__pyx_k_StOutPutsDeclF90), 0, 0, 1, 1},
  {&__pyx_n_s_StOutputName, __pyx_k_StOutputName, sizeof(__pyx_k_StOutputName), 0, 0, 1, 1},
  {&__pyx_n_s_StOutputNameDecl, __pyx_k_StOutputNameDecl, sizeof(__pyx_k_StOutputNameDecl), 0, 0, 1, 1},
  {&__pyx_n_s_StOutputNameDeclF77, __pyx_k_StOutputNameDeclF77, sizeof(__pyx_k_StOutputNameDeclF77), 0, 0, 1, 1},
  {&__pyx_n_s_StOutputNameDeclF90, __pyx_k_StOutputNameDeclF90, sizeof(__pyx_k_StOutputNameDeclF90), 0, 0, 1, 1},
  {&__pyx_n_s_StOutputNameDeclP, __pyx_k_StOutputNameDeclP, sizeof(__pyx_k_StOutputNameDeclP), 0, 0, 1, 1},
  {&__pyx_n_s_StOutputNameF, __pyx_k_StOutputNameF, sizeof(__pyx_k_StOutputNameF), 0, 0, 1, 1},
  {&__pyx_n_s_StOutputNameFU, __pyx_k_StOutputNameFU, sizeof(__pyx_k_StOutputNameFU), 0, 0, 1, 1},
  {&__pyx_n_s_StOutputNameP, __pyx_k_StOutputNameP, sizeof(__pyx_k_StOutputNameP), 0, 0, 1, 1},
  {&__pyx_n_s_StOutputNameProto, __pyx_k_StOutputNameProto, sizeof(__pyx_k_StOutputNameProto), 0, 0, 1, 1},
  {&__pyx_n_s_StReturnInputName, __pyx_k_StReturnInputName, sizeof(__pyx_k_StReturnInputName), 0, 0, 1, 1},
  {&__pyx_n_s_StReturnInputNameP1, __pyx_k_StReturnInputNameP1, sizeof(__pyx_k_StReturnInputNameP1), 0, 0, 1, 1},
  {&__pyx_n_s_StReturnOutputName, __pyx_k_StReturnOutputName, sizeof(__pyx_k_StReturnOutputName), 0, 0, 1, 1},
  {&__pyx_n_s_StReturnOutputNameP2, __pyx_k_StReturnOutputNameP2, sizeof(__pyx_k_StReturnOutputNameP2), 0, 0, 1, 1},
  {&__pyx_n_s_StStdDEvIniF, __pyx_k_StStdDEvIniF, sizeof(__pyx_k_StStdDEvIniF), 0, 0, 1, 1},
  {&__pyx_n_s_StStdDEvIniP, __pyx_k_StStdDEvIniP, sizeof(__pyx_k_StStdDEvIniP), 0, 0, 1, 1},
  {&__pyx_n_s_StStdDEvInis, __pyx_k_StStdDEvInis, sizeof(__pyx_k_StStdDEvInis), 0, 0, 1, 1},
  {&__pyx_n_s_StTransGradDyn, __pyx_k_StTransGradDyn, sizeof(__pyx_k_StTransGradDyn), 0, 0, 1, 1},
  {&__pyx_n_s_StTransGradDynCall, __pyx_k_StTransGradDynCall, sizeof(__pyx_k_StTransGradDynCall), 0, 0, 1, 1},
  {&__pyx_n_s_StTransGradDynCallF, __pyx_k_StTransGradDynCallF, sizeof(__pyx_k_StTransGradDynCallF), 0, 0, 1, 1},
  {&__pyx_n_s_StTransGradDynCallP, __pyx_k_StTransGradDynCallP, sizeof(__pyx_k_StTransGradDynCallP), 0, 0, 1, 1},
  {&__pyx_n_s_StTransGradDynF, __pyx_k_StTransGradDynF, sizeof(__pyx_k_StTransGradDynF), 0, 0, 1, 1},
  {&__pyx_n_s_StTransGradDynPP, __pyx_k_StTransGradDynPP, sizeof(__pyx_k_StTransGradDynPP), 0, 0, 1, 1},
  {&__pyx_n_s_StTransGradDynProto, __pyx_k_StTransGradDynProto, sizeof(__pyx_k_StTransGradDynProto), 0, 0, 1, 1},
  {&__pyx_n_s_StTransGradStat, __pyx_k_StTransGradStat, sizeof(__pyx_k_StTransGradStat), 0, 0, 1, 1},
  {&__pyx_n_s_StTransGradStatCall, __pyx_k_StTransGradStatCall, sizeof(__pyx_k_StTransGradStatCall), 0, 0, 1, 1},
  {&__pyx_n_s_StTransGradStatCallF, __pyx_k_StTransGradStatCallF, sizeof(__pyx_k_StTransGradStatCallF), 0, 0, 1, 1},
  {&__pyx_n_s_StTransGradStatCallP, __pyx_k_StTransGradStatCallP, sizeof(__pyx_k_StTransGradStatCallP), 0, 0, 1, 1},
  {&__pyx_n_s_StTransGradStatF, __pyx_k_StTransGradStatF, sizeof(__pyx_k_StTransGradStatF), 0, 0, 1, 1},
  {&__pyx_n_s_StTransGradStatPP, __pyx_k_StTransGradStatPP, sizeof(__pyx_k_StTransGradStatPP), 0, 0, 1, 1},
  {&__pyx_n_s_StTransGradStatProto, __pyx_k_StTransGradStatProto, sizeof(__pyx_k_StTransGradStatProto), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferDyn, __pyx_k_StTransferDyn, sizeof(__pyx_k_StTransferDyn), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferDyn2, __pyx_k_StTransferDyn2, sizeof(__pyx_k_StTransferDyn2), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferDynF, __pyx_k_StTransferDynF, sizeof(__pyx_k_StTransferDynF), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferDynF2, __pyx_k_StTransferDynF2, sizeof(__pyx_k_StTransferDynF2), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferDynPP, __pyx_k_StTransferDynPP, sizeof(__pyx_k_StTransferDynPP), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferDynPP2, __pyx_k_StTransferDynPP2, sizeof(__pyx_k_StTransferDynPP2), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferDynProto, __pyx_k_StTransferDynProto, sizeof(__pyx_k_StTransferDynProto), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferDynProto2, __pyx_k_StTransferDynProto2, sizeof(__pyx_k_StTransferDynProto2), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferStat, __pyx_k_StTransferStat, sizeof(__pyx_k_StTransferStat), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferStat2, __pyx_k_StTransferStat2, sizeof(__pyx_k_StTransferStat2), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferStatF, __pyx_k_StTransferStatF, sizeof(__pyx_k_StTransferStatF), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferStatF2, __pyx_k_StTransferStatF2, sizeof(__pyx_k_StTransferStatF2), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferStatPP, __pyx_k_StTransferStatPP, sizeof(__pyx_k_StTransferStatPP), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferStatPP2, __pyx_k_StTransferStatPP2, sizeof(__pyx_k_StTransferStatPP2), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferStatProto, __pyx_k_StTransferStatProto, sizeof(__pyx_k_StTransferStatProto), 0, 0, 1, 1},
  {&__pyx_n_s_StTransferStatProto2, __pyx_k_StTransferStatProto2, sizeof(__pyx_k_StTransferStatProto2), 0, 0, 1, 1},
  {&__pyx_n_s_StWeightIni, __pyx_k_StWeightIni, sizeof(__pyx_k_StWeightIni), 0, 0, 1, 1},
  {&__pyx_n_s_StWeightIniF, __pyx_k_StWeightIniF, sizeof(__pyx_k_StWeightIniF), 0, 0, 1, 1},
  {&__pyx_n_s_StWeightIniS, __pyx_k_StWeightIniS, sizeof(__pyx_k_StWeightIniS), 0, 0, 1, 1},
  {&__pyx_n_s_StWeightIniSP, __pyx_k_StWeightIniSP, sizeof(__pyx_k_StWeightIniSP), 0, 0, 1, 1},
  {&__pyx_n_s_StWeightRedirIni, __pyx_k_StWeightRedirIni, sizeof(__pyx_k_StWeightRedirIni), 0, 0, 1, 1},
  {&__pyx_kp_u_StdDev_s, __pyx_k_StdDev_s, sizeof(__pyx_k_StdDev_s), 0, 1, 0, 0},
  {&__pyx_n_s_StdefineStdDev, __pyx_k_StdefineStdDev, sizeof(__pyx_k_StdefineStdDev), 0, 0, 1, 1},
  {&__pyx_n_s_StdefineStud95, __pyx_k_StdefineStud95, sizeof(__pyx_k_StdefineStud95), 0, 0, 1, 1},
  {&__pyx_kp_u_T_sInOut, __pyx_k_T_sInOut, sizeof(__pyx_k_T_sInOut), 0, 1, 0, 0},
  {&__pyx_kp_u_T_sMatrix, __pyx_k_T_sMatrix, sizeof(__pyx_k_T_sMatrix), 0, 1, 0, 0},
  {&__pyx_kp_u_T_sMemsyn, __pyx_k_T_sMemsyn, sizeof(__pyx_k_T_sMemsyn), 0, 1, 0, 0},
  {&__pyx_kp_u_T_sNode, __pyx_k_T_sNode, sizeof(__pyx_k_T_sNode), 0, 1, 0, 0},
  {&__pyx_kp_u_T_sOutput, __pyx_k_T_sOutput, sizeof(__pyx_k_T_sOutput), 0, 1, 0, 0},
  {&__pyx_kp_u_T_sSynapse, __pyx_k_T_sSynapse, sizeof(__pyx_k_T_sSynapse), 0, 1, 0, 0},
  {&__pyx_kp_u_This_code_can_be_freely_used_wit, __pyx_k_This_code_can_be_freely_used_wit, sizeof(__pyx_k_This_code_can_be_freely_used_wit), 0, 1, 0, 0},
  {&__pyx_kp_u_This_code_has_been_automatically, __pyx_k_This_code_has_been_automatically, sizeof(__pyx_k_This_code_has_been_automatically), 0, 1, 0, 0},
  {&__pyx_kp_u_Time_s, __pyx_k_Time_s, sizeof(__pyx_k_Time_s), 0, 1, 0, 0},
  {&__pyx_n_u_Transfert, __pyx_k_Transfert, sizeof(__pyx_k_Transfert), 0, 1, 0, 1},
  {&__pyx_n_u_TransfertLeverage, __pyx_k_TransfertLeverage, sizeof(__pyx_k_TransfertLeverage), 0, 1, 0, 1},
  {&__pyx_kp_u_Under_construction, __pyx_k_Under_construction, sizeof(__pyx_k_Under_construction), 0, 1, 0, 0},
  {&__pyx_kp_u_User_s, __pyx_k_User_s, sizeof(__pyx_k_User_s), 0, 1, 0, 0},
  {&__pyx_n_u_WEIGHTS, __pyx_k_WEIGHTS, sizeof(__pyx_k_WEIGHTS), 0, 1, 0, 1},
  {&__pyx_kp_u_X_potential_Double, __pyx_k_X_potential_Double, sizeof(__pyx_k_X_potential_Double), 0, 1, 0, 0},
  {&__pyx_kp_u__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 1, 0, 0},
  {&__pyx_kp_u__11, __pyx_k__11, sizeof(__pyx_k__11), 0, 1, 0, 0},
  {&__pyx_kp_u__12, __pyx_k__12, sizeof(__pyx_k__12), 0, 1, 0, 0},
  {&__pyx_kp_u__13, __pyx_k__13, sizeof(__pyx_k__13), 0, 1, 0, 0},
  {&__pyx_kp_u__14, __pyx_k__14, sizeof(__pyx_k__14), 0, 1, 0, 0},
  {&__pyx_kp_u__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 1, 0, 0},
  {&__pyx_kp_u__16, __pyx_k__16, sizeof(__pyx_k__16), 0, 1, 0, 0},
  {&__pyx_kp_u__17, __pyx_k__17, sizeof(__pyx_k__17), 0, 1, 0, 0},
  {&__pyx_kp_u__18, __pyx_k__18, sizeof(__pyx_k__18), 0, 1, 0, 0},
  {&__pyx_n_u_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 1, 0, 1},
  {&__pyx_kp_u_char_sinputname_int, __pyx_k_char_sinputname_int, sizeof(__pyx_k_char_sinputname_int), 0, 1, 0, 0},
  {&__pyx_kp_u_char_sinputname_int_index, __pyx_k_char_sinputname_int_index, sizeof(__pyx_k_char_sinputname_int_index), 0, 1, 0, 0},
  {&__pyx_kp_u_char_soutputname_int, __pyx_k_char_soutputname_int, sizeof(__pyx_k_char_soutputname_int), 0, 1, 0, 0},
  {&__pyx_kp_u_char_soutputname_int_index, __pyx_k_char_soutputname_int_index, sizeof(__pyx_k_char_soutputname_int_index), 0, 1, 0, 0},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_code_Body, __pyx_k_code_Body, sizeof(__pyx_k_code_Body), 0, 0, 1, 1},
  {&__pyx_n_s_code_Header, __pyx_k_code_Header, sizeof(__pyx_k_code_Header), 0, 0, 1, 1},
  {&__pyx_kp_u_code_The_transmission_of_this_co, __pyx_k_code_The_transmission_of_this_co, sizeof(__pyx_k_code_The_transmission_of_this_co), 0, 1, 0, 0},
  {&__pyx_n_s_code_common_maxi, __pyx_k_code_common_maxi, sizeof(__pyx_k_code_common_maxi), 0, 0, 1, 1},
  {&__pyx_n_s_code_common_mini, __pyx_k_code_common_mini, sizeof(__pyx_k_code_common_mini), 0, 0, 1, 1},
  {&__pyx_n_s_codename, __pyx_k_codename, sizeof(__pyx_k_codename), 0, 0, 1, 1},
  {&__pyx_n_s_commondict, __pyx_k_commondict, sizeof(__pyx_k_commondict), 0, 0, 1, 1},
  {&__pyx_n_s_cs_All, __pyx_k_cs_All, sizeof(__pyx_k_cs_All), 0, 0, 1, 1},
  {&__pyx_n_s_cs_Common, __pyx_k_cs_Common, sizeof(__pyx_k_cs_Common), 0, 0, 1, 1},
  {&__pyx_n_s_cs_Model, __pyx_k_cs_Model, sizeof(__pyx_k_cs_Model), 0, 0, 1, 1},
  {&__pyx_n_s_ctC, __pyx_k_ctC, sizeof(__pyx_k_ctC), 0, 0, 1, 1},
  {&__pyx_n_s_ctCython, __pyx_k_ctCython, sizeof(__pyx_k_ctCython), 0, 0, 1, 1},
  {&__pyx_n_s_ctExcel, __pyx_k_ctExcel, sizeof(__pyx_k_ctExcel), 0, 0, 1, 1},
  {&__pyx_n_s_ctFortran77, __pyx_k_ctFortran77, sizeof(__pyx_k_ctFortran77), 0, 0, 1, 1},
  {&__pyx_n_s_ctFortran90, __pyx_k_ctFortran90, sizeof(__pyx_k_ctFortran90), 0, 0, 1, 1},
  {&__pyx_n_s_ctPascal, __pyx_k_ctPascal, sizeof(__pyx_k_ctPascal), 0, 0, 1, 1},
  {&__pyx_n_s_ctPython, __pyx_k_ctPython, sizeof(__pyx_k_ctPython), 0, 0, 1, 1},
  {&__pyx_n_s_ctUnknown, __pyx_k_ctUnknown, sizeof(__pyx_k_ctUnknown), 0, 0, 1, 1},
  {&__pyx_n_s_ctVBasic, __pyx_k_ctVBasic, sizeof(__pyx_k_ctVBasic), 0, 0, 1, 1},
  {&__pyx_n_s_ctXML, __pyx_k_ctXML, sizeof(__pyx_k_ctXML), 0, 0, 1, 1},
  {&__pyx_kp_u_define_mindex_i_j_i_j_j_i_i_1_2, __pyx_k_define_mindex_i_j_i_j_j_i_i_1_2, sizeof(__pyx_k_define_mindex_i_j_i_j_j_i_i_1_2), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sCONSTANTS_d, __pyx_k_define_sCONSTANTS_d, sizeof(__pyx_k_define_sCONSTANTS_d), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sDIMENSION_d, __pyx_k_define_sDIMENSION_d, sizeof(__pyx_k_define_sDIMENSION_d), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sDIM_d_d, __pyx_k_define_sDIM_d_d, sizeof(__pyx_k_define_sDIM_d_d), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sINPUTS_d, __pyx_k_define_sINPUTS_d, sizeof(__pyx_k_define_sINPUTS_d), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sMATRIXES_d, __pyx_k_define_sMATRIXES_d, sizeof(__pyx_k_define_sMATRIXES_d), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sMEMNEURO_d, __pyx_k_define_sMEMNEURO_d, sizeof(__pyx_k_define_sMEMNEURO_d), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sMEMSYN_d, __pyx_k_define_sMEMSYN_d, sizeof(__pyx_k_define_sMEMSYN_d), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sNODES_d, __pyx_k_define_sNODES_d, sizeof(__pyx_k_define_sNODES_d), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sORDER_d, __pyx_k_define_sORDER_d, sizeof(__pyx_k_define_sORDER_d), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sOUTPUTS_d, __pyx_k_define_sOUTPUTS_d, sizeof(__pyx_k_define_sOUTPUTS_d), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sSTDTRDEV_g, __pyx_k_define_sSTDTRDEV_g, sizeof(__pyx_k_define_sSTDTRDEV_g), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sSTUDENT95_g, __pyx_k_define_sSTUDENT95_g, sizeof(__pyx_k_define_sSTUDENT95_g), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sSYNAPSES_d, __pyx_k_define_sSYNAPSES_d, sizeof(__pyx_k_define_sSYNAPSES_d), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sTRANSFERGR_H, __pyx_k_define_sTRANSFERGR_H, sizeof(__pyx_k_define_sTRANSFERGR_H), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sTRANSFERLEV_H, __pyx_k_define_sTRANSFERLEV_H, sizeof(__pyx_k_define_sTRANSFERLEV_H), 0, 1, 0, 0},
  {&__pyx_kp_u_define_sTRANSFER_H, __pyx_k_define_sTRANSFER_H, sizeof(__pyx_k_define_sTRANSFER_H), 0, 1, 0, 0},
  {&__pyx_n_u_dlmdriver, __pyx_k_dlmdriver, sizeof(__pyx_k_dlmdriver), 0, 1, 0, 1},
  {&__pyx_n_u_dlmdriverhelp, __pyx_k_dlmdriverhelp, sizeof(__pyx_k_dlmdriverhelp), 0, 1, 0, 1},
  {&__pyx_n_u_dlmuse, __pyx_k_dlmuse, sizeof(__pyx_k_dlmuse), 0, 1, 0, 1},
  {&__pyx_n_u_dlmutil, __pyx_k_dlmutil, sizeof(__pyx_k_dlmutil), 0, 1, 0, 1},
  {&__pyx_kp_u_double_gradient_sSYNAPSES, __pyx_k_double_gradient_sSYNAPSES, sizeof(__pyx_k_double_gradient_sSYNAPSES), 0, 1, 0, 0},
  {&__pyx_kp_u_double_iniback_sOUTPUTS, __pyx_k_double_iniback_sOUTPUTS, sizeof(__pyx_k_double_iniback_sOUTPUTS), 0, 1, 0, 0},
  {&__pyx_kp_u_double_memsyn_sMEMSYN, __pyx_k_double_memsyn_sMEMSYN, sizeof(__pyx_k_double_memsyn_sMEMSYN), 0, 1, 0, 0},
  {&__pyx_kp_u_double_node_sNODES, __pyx_k_double_node_sNODES, sizeof(__pyx_k_double_node_sNODES), 0, 1, 0, 0},
  {&__pyx_kp_u_double_node_sNODES_2, __pyx_k_double_node_sNODES_2, sizeof(__pyx_k_double_node_sNODES_2), 0, 1, 0, 0},
  {&__pyx_kp_u_double_node_sNODES_3, __pyx_k_double_node_sNODES_3, sizeof(__pyx_k_double_node_sNODES_3), 0, 1, 0, 0},
  {&__pyx_kp_u_double_node_sOUTPUTS, __pyx_k_double_node_sOUTPUTS, sizeof(__pyx_k_double_node_sOUTPUTS), 0, 1, 0, 0},
  {&__pyx_kp_u_else_Result_names_index, __pyx_k_else_Result_names_index, sizeof(__pyx_k_else_Result_names_index), 0, 1, 0, 0},
  {&__pyx_kp_u_following_conditions, __pyx_k_following_conditions, sizeof(__pyx_k_following_conditions), 0, 1, 0, 0},
  {&__pyx_kp_u_for_i_0_i_sMEMSYN_memsyn_i_1, __pyx_k_for_i_0_i_sMEMSYN_memsyn_i_1, sizeof(__pyx_k_for_i_0_i_sMEMSYN_memsyn_i_1), 0, 1, 0, 0},
  {&__pyx_kp_u_for_i_0_i_sOUTPUTS_iniback_i_0, __pyx_k_for_i_0_i_sOUTPUTS_iniback_i_0, sizeof(__pyx_k_for_i_0_i_sOUTPUTS_iniback_i_0), 0, 1, 0, 0},
  {&__pyx_kp_u_for_i_0_i_sOUTPUTS_leverage_i_0, __pyx_k_for_i_0_i_sOUTPUTS_leverage_i_0, sizeof(__pyx_k_for_i_0_i_sOUTPUTS_leverage_i_0), 0, 1, 0, 0},
  {&__pyx_kp_u_for_i_0_i_sSYNAPSES_gradient_i_0, __pyx_k_for_i_0_i_sSYNAPSES_gradient_i_0, sizeof(__pyx_k_for_i_0_i_sSYNAPSES_gradient_i_0), 0, 1, 0, 0},
  {&__pyx_kp_u_for_i_0_i_sSYNAPSES_i, __pyx_k_for_i_0_i_sSYNAPSES_i, sizeof(__pyx_k_for_i_0_i_sSYNAPSES_i), 0, 1, 0, 0},
  {&__pyx_kp_u_for_i_0_to_sMEMSYN_1_do_memsyn_i, __pyx_k_for_i_0_to_sMEMSYN_1_do_memsyn_i, sizeof(__pyx_k_for_i_0_to_sMEMSYN_1_do_memsyn_i), 0, 1, 0, 0},
  {&__pyx_kp_u_for_i_0_to_sOUTPUTS_1_do_Leverag, __pyx_k_for_i_0_to_sOUTPUTS_1_do_Leverag, sizeof(__pyx_k_for_i_0_to_sOUTPUTS_1_do_Leverag), 0, 1, 0, 0},
  {&__pyx_kp_u_for_i_0_to_sOUTPUTS_1_do_iniback, __pyx_k_for_i_0_to_sOUTPUTS_1_do_iniback, sizeof(__pyx_k_for_i_0_to_sOUTPUTS_1_do_iniback), 0, 1, 0, 0},
  {&__pyx_kp_u_for_i_0_to_sSYNAPSES_1_do, __pyx_k_for_i_0_to_sSYNAPSES_1_do, sizeof(__pyx_k_for_i_0_to_sSYNAPSES_1_do), 0, 1, 0, 0},
  {&__pyx_kp_u_for_i_0_to_sSYNAPSES_1_do_gradie, __pyx_k_for_i_0_to_sSYNAPSES_1_do_gradie, sizeof(__pyx_k_for_i_0_to_sSYNAPSES_1_do_gradie), 0, 1, 0, 0},
  {&__pyx_kp_u_for_i_1_i_sOUTPUTS_i, __pyx_k_for_i_1_i_sOUTPUTS_i, sizeof(__pyx_k_for_i_1_i_sOUTPUTS_i), 0, 1, 0, 0},
  {&__pyx_kp_u_for_i_1_to_sOUTPUTS_1_do, __pyx_k_for_i_1_to_sOUTPUTS_1_do, sizeof(__pyx_k_for_i_1_to_sOUTPUTS_1_do), 0, 1, 0, 0},
  {&__pyx_kp_u_for_j_0_j_sSYNAPSES_j, __pyx_k_for_j_0_j_sSYNAPSES_j, sizeof(__pyx_k_for_j_0_j_sSYNAPSES_j), 0, 1, 0, 0},
  {&__pyx_kp_u_for_j_0_to_sSYNAPSES_1_do, __pyx_k_for_j_0_to_sSYNAPSES_1_do, sizeof(__pyx_k_for_j_0_to_sSYNAPSES_1_do), 0, 1, 0, 0},
  {&__pyx_n_s_ftBase, __pyx_k_ftBase, sizeof(__pyx_k_ftBase), 0, 0, 1, 1},
  {&__pyx_n_s_ftDLM, __pyx_k_ftDLM, sizeof(__pyx_k_ftDLM), 0, 0, 1, 1},
  {&__pyx_n_s_ftDesc, __pyx_k_ftDesc, sizeof(__pyx_k_ftDesc), 0, 0, 1, 1},
  {&__pyx_n_s_ftGradient, __pyx_k_ftGradient, sizeof(__pyx_k_ftGradient), 0, 0, 1, 1},
  {&__pyx_n_s_ftMake, __pyx_k_ftMake, sizeof(__pyx_k_ftMake), 0, 0, 1, 1},
  {&__pyx_n_s_ftMakeDLM, __pyx_k_ftMakeDLM, sizeof(__pyx_k_ftMakeDLM), 0, 0, 1, 1},
  {&__pyx_n_s_ftMax, __pyx_k_ftMax, sizeof(__pyx_k_ftMax), 0, 0, 1, 1},
  {&__pyx_n_s_ftOnlyHeader, __pyx_k_ftOnlyHeader, sizeof(__pyx_k_ftOnlyHeader), 0, 0, 1, 1},
  {&__pyx_n_s_ftSetupMulti, __pyx_k_ftSetupMulti, sizeof(__pyx_k_ftSetupMulti), 0, 0, 1, 1},
  {&__pyx_n_s_ftSetupPy, __pyx_k_ftSetupPy, sizeof(__pyx_k_ftSetupPy), 0, 0, 1, 1},
  {&__pyx_n_s_ftSpecialSub, __pyx_k_ftSpecialSub, sizeof(__pyx_k_ftSpecialSub), 0, 0, 1, 1},
  {&__pyx_n_s_ftTransfert, __pyx_k_ftTransfert, sizeof(__pyx_k_ftTransfert), 0, 0, 1, 1},
  {&__pyx_n_s_ftTransfertLeverage, __pyx_k_ftTransfertLeverage, sizeof(__pyx_k_ftTransfertLeverage), 0, 0, 1, 1},
  {&__pyx_n_s_ftWithHeader, __pyx_k_ftWithHeader, sizeof(__pyx_k_ftWithHeader), 0, 0, 1, 1},
  {&__pyx_n_s_ftbasetrain, __pyx_k_ftbasetrain, sizeof(__pyx_k_ftbasetrain), 0, 0, 1, 1},
  {&__pyx_n_s_ftbaseuse, __pyx_k_ftbaseuse, sizeof(__pyx_k_ftbaseuse), 0, 0, 1, 1},
  {&__pyx_n_s_ftdlmdriver, __pyx_k_ftdlmdriver, sizeof(__pyx_k_ftdlmdriver), 0, 0, 1, 1},
  {&__pyx_n_s_ftdlmdriverhelp, __pyx_k_ftdlmdriverhelp, sizeof(__pyx_k_ftdlmdriverhelp), 0, 0, 1, 1},
  {&__pyx_n_s_ftdlmuse, __pyx_k_ftdlmuse, sizeof(__pyx_k_ftdlmuse), 0, 0, 1, 1},
  {&__pyx_n_s_ftdlmutil, __pyx_k_ftdlmutil, sizeof(__pyx_k_ftdlmutil), 0, 0, 1, 1},
  {&__pyx_n_s_ftgenerictrain, __pyx_k_ftgenerictrain, sizeof(__pyx_k_ftgenerictrain), 0, 0, 1, 1},
  {&__pyx_n_s_ftmathplus, __pyx_k_ftmathplus, sizeof(__pyx_k_ftmathplus), 0, 0, 1, 1},
  {&__pyx_n_s_ftmodeltrain, __pyx_k_ftmodeltrain, sizeof(__pyx_k_ftmodeltrain), 0, 0, 1, 1},
  {&__pyx_n_s_ftnetraltype, __pyx_k_ftnetraltype, sizeof(__pyx_k_ftnetraltype), 0, 0, 1, 1},
  {&__pyx_n_s_ftntdefines, __pyx_k_ftntdefines, sizeof(__pyx_k_ftntdefines), 0, 0, 1, 1},
  {&__pyx_n_s_ftnttoolbox, __pyx_k_ftnttoolbox, sizeof(__pyx_k_ftnttoolbox), 0, 0, 1, 1},
  {&__pyx_n_s_ftreadfile, __pyx_k_ftreadfile, sizeof(__pyx_k_ftreadfile), 0, 0, 1, 1},
  {&__pyx_n_s_ftsetupcfg, __pyx_k_ftsetupcfg, sizeof(__pyx_k_ftsetupcfg), 0, 0, 1, 1},
  {&__pyx_n_s_ftsinglebasetrain, __pyx_k_ftsinglebasetrain, sizeof(__pyx_k_ftsinglebasetrain), 0, 0, 1, 1},
  {&__pyx_n_s_ftsingletraindll, __pyx_k_ftsingletraindll, sizeof(__pyx_k_ftsingletraindll), 0, 0, 1, 1},
  {&__pyx_n_s_fttrain, __pyx_k_fttrain, sizeof(__pyx_k_fttrain), 0, 0, 1, 1},
  {&__pyx_n_s_fttraindll, __pyx_k_fttraindll, sizeof(__pyx_k_fttraindll), 0, 0, 1, 1},
  {&__pyx_n_s_ftusedll, __pyx_k_ftusedll, sizeof(__pyx_k_ftusedll), 0, 0, 1, 1},
  {&__pyx_n_s_ftwithCext, __pyx_k_ftwithCext, sizeof(__pyx_k_ftwithCext), 0, 0, 1, 1},
  {&__pyx_n_u_generictrain, __pyx_k_generictrain, sizeof(__pyx_k_generictrain), 0, 1, 0, 1},
  {&__pyx_kp_u_gradient_T_sSynapse, __pyx_k_gradient_T_sSynapse, sizeof(__pyx_k_gradient_T_sSynapse), 0, 1, 0, 0},
  {&__pyx_kp_u_if_index_0_or_index_sINPUTS_then, __pyx_k_if_index_0_or_index_sINPUTS_then, sizeof(__pyx_k_if_index_0_or_index_sINPUTS_then), 0, 1, 0, 0},
  {&__pyx_kp_u_ifndef_sTRANSFERGR_H, __pyx_k_ifndef_sTRANSFERGR_H, sizeof(__pyx_k_ifndef_sTRANSFERGR_H), 0, 1, 0, 0},
  {&__pyx_kp_u_ifndef_sTRANSFERLEV_H, __pyx_k_ifndef_sTRANSFERLEV_H, sizeof(__pyx_k_ifndef_sTRANSFERLEV_H), 0, 1, 0, 0},
  {&__pyx_kp_u_ifndef_sTRANSFER_H, __pyx_k_ifndef_sTRANSFER_H, sizeof(__pyx_k_ifndef_sTRANSFER_H), 0, 1, 0, 0},
  {&__pyx_kp_u_include_s, __pyx_k_include_s, sizeof(__pyx_k_include_s), 0, 1, 0, 0},
  {&__pyx_kp_u_iniback_T_sOutput, __pyx_k_iniback_T_sOutput, sizeof(__pyx_k_iniback_T_sOutput), 0, 1, 0, 0},
  {&__pyx_n_u_interfacetraindll, __pyx_k_interfacetraindll, sizeof(__pyx_k_interfacetraindll), 0, 1, 0, 1},
  {&__pyx_n_u_interfaceusedll, __pyx_k_interfaceusedll, sizeof(__pyx_k_interfaceusedll), 0, 1, 0, 1},
  {&__pyx_kp_u_is_clearly_mentionned, __pyx_k_is_clearly_mentionned, sizeof(__pyx_k_is_clearly_mentionned), 0, 1, 0, 0},
  {&__pyx_kp_u_jurgb_txt, __pyx_k_jurgb_txt, sizeof(__pyx_k_jurgb_txt), 0, 1, 0, 0},
  {&__pyx_n_s_lower_re, __pyx_k_lower_re, sizeof(__pyx_k_lower_re), 0, 0, 1, 1},
  {&__pyx_kp_u_lutilisation_de_ce_code, __pyx_k_lutilisation_de_ce_code, sizeof(__pyx_k_lutilisation_de_ce_code), 0, 1, 0, 0},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_u_mathplus, __pyx_k_mathplus, sizeof(__pyx_k_mathplus), 0, 1, 0, 1},
  {&__pyx_kp_u_matrix_d, __pyx_k_matrix_d, sizeof(__pyx_k_matrix_d), 0, 1, 0, 0},
  {&__pyx_kp_u_matrix_s, __pyx_k_matrix_s, sizeof(__pyx_k_matrix_s), 0, 1, 0, 0},
  {&__pyx_kp_u_matrix_sMATRIXES, __pyx_k_matrix_sMATRIXES, sizeof(__pyx_k_matrix_sMATRIXES), 0, 1, 0, 0},
  {&__pyx_kp_u_max_s, __pyx_k_max_s, sizeof(__pyx_k_max_s), 0, 1, 0, 0},
  {&__pyx_kp_u_max_sINPUTS_sOUTPUTS, __pyx_k_max_sINPUTS_sOUTPUTS, sizeof(__pyx_k_max_sINPUTS_sOUTPUTS), 0, 1, 0, 0},
  {&__pyx_kp_u_memsyn_T_sMemsyn, __pyx_k_memsyn_T_sMemsyn, sizeof(__pyx_k_memsyn_T_sMemsyn), 0, 1, 0, 0},
  {&__pyx_kp_u_min_s, __pyx_k_min_s, sizeof(__pyx_k_min_s), 0, 1, 0, 0},
  {&__pyx_kp_u_min_sINPUTS_sOUTPUTS, __pyx_k_min_sINPUTS_sOUTPUTS, sizeof(__pyx_k_min_sINPUTS_sOUTPUTS), 0, 1, 0, 0},
  {&__pyx_n_u_modeltrain, __pyx_k_modeltrain, sizeof(__pyx_k_modeltrain), 0, 1, 0, 1},
  {&__pyx_kp_u_names_Array_0_sINPUTS_1_of_Strin, __pyx_k_names_Array_0_sINPUTS_1_of_Strin, sizeof(__pyx_k_names_Array_0_sINPUTS_1_of_Strin), 0, 1, 0, 0},
  {&__pyx_kp_u_names_Array_0_sOUTPUTS_1_of_Stri, __pyx_k_names_Array_0_sOUTPUTS_1_of_Stri, sizeof(__pyx_k_names_Array_0_sOUTPUTS_1_of_Stri), 0, 1, 0, 0},
  {&__pyx_n_s_ncNone, __pyx_k_ncNone, sizeof(__pyx_k_ncNone), 0, 0, 1, 1},
  {&__pyx_n_s_ncPi, __pyx_k_ncPi, sizeof(__pyx_k_ncPi), 0, 0, 1, 1},
  {&__pyx_n_s_ncSigma, __pyx_k_ncSigma, sizeof(__pyx_k_ncSigma), 0, 0, 1, 1},
  {&__pyx_n_s_nkDyn, __pyx_k_nkDyn, sizeof(__pyx_k_nkDyn), 0, 0, 1, 1},
  {&__pyx_n_s_nkStat, __pyx_k_nkStat, sizeof(__pyx_k_nkStat), 0, 0, 1, 1},
  {&__pyx_n_s_nl, __pyx_k_nl, sizeof(__pyx_k_nl), 0, 0, 1, 1},
  {&__pyx_n_s_nlf, __pyx_k_nlf, sizeof(__pyx_k_nlf), 0, 0, 1, 1},
  {&__pyx_n_s_nncName, __pyx_k_nncName, sizeof(__pyx_k_nncName), 0, 0, 1, 1},
  {&__pyx_n_s_nncNode, __pyx_k_nncNode, sizeof(__pyx_k_nncNode), 0, 0, 1, 1},
  {&__pyx_kp_u_node_T_sNode, __pyx_k_node_T_sNode, sizeof(__pyx_k_node_T_sNode), 0, 1, 0, 0},
  {&__pyx_kp_u_node_T_sNode2, __pyx_k_node_T_sNode2, sizeof(__pyx_k_node_T_sNode2), 0, 1, 0, 0},
  {&__pyx_kp_u_node_T_sNode3, __pyx_k_node_T_sNode3, sizeof(__pyx_k_node_T_sNode3), 0, 1, 0, 0},
  {&__pyx_kp_u_node_T_sOutput, __pyx_k_node_T_sOutput, sizeof(__pyx_k_node_T_sOutput), 0, 1, 0, 0},
  {&__pyx_n_s_npConst, __pyx_k_npConst, sizeof(__pyx_k_npConst), 0, 0, 1, 1},
  {&__pyx_n_s_npHidden, __pyx_k_npHidden, sizeof(__pyx_k_npHidden), 0, 0, 1, 1},
  {&__pyx_n_s_npInput, __pyx_k_npInput, sizeof(__pyx_k_npInput), 0, 0, 1, 1},
  {&__pyx_n_s_npOutState, __pyx_k_npOutState, sizeof(__pyx_k_npOutState), 0, 0, 1, 1},
  {&__pyx_n_s_npOutput, __pyx_k_npOutput, sizeof(__pyx_k_npOutput), 0, 0, 1, 1},
  {&__pyx_n_s_npState, __pyx_k_npState, sizeof(__pyx_k_npState), 0, 0, 1, 1},
  {&__pyx_n_u_ntdefines, __pyx_k_ntdefines, sizeof(__pyx_k_ntdefines), 0, 1, 0, 1},
  {&__pyx_n_u_nttoolbox, __pyx_k_nttoolbox, sizeof(__pyx_k_nttoolbox), 0, 1, 0, 1},
  {&__pyx_n_u_nttype, __pyx_k_nttype, sizeof(__pyx_k_nttype), 0, 1, 0, 1},
  {&__pyx_kp_u_organisation_is_free_provided_th, __pyx_k_organisation_is_free_provided_th, sizeof(__pyx_k_organisation_is_free_provided_th), 0, 1, 0, 0},
  {&__pyx_kp_u_potential_Double, __pyx_k_potential_Double, sizeof(__pyx_k_potential_Double), 0, 1, 0, 0},
  {&__pyx_n_s_proprietaryname, __pyx_k_proprietaryname, sizeof(__pyx_k_proprietaryname), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_u_readfile, __pyx_k_readfile, sizeof(__pyx_k_readfile), 0, 1, 0, 1},
  {&__pyx_kp_u_return_index_0_index_sINPUTS_NUL, __pyx_k_return_index_0_index_sINPUTS_NUL, sizeof(__pyx_k_return_index_0_index_sINPUTS_NUL), 0, 1, 0, 0},
  {&__pyx_kp_u_return_index_0_index_sOUTPUTS_NU, __pyx_k_return_index_0_index_sOUTPUTS_NU, sizeof(__pyx_k_return_index_0_index_sOUTPUTS_NU), 0, 1, 0, 0},
  {&__pyx_kp_u_sCONSTANTS_d, __pyx_k_sCONSTANTS_d, sizeof(__pyx_k_sCONSTANTS_d), 0, 1, 0, 0},
  {&__pyx_kp_u_sDIMENSION_d, __pyx_k_sDIMENSION_d, sizeof(__pyx_k_sDIMENSION_d), 0, 1, 0, 0},
  {&__pyx_kp_u_sDIM_d_d, __pyx_k_sDIM_d_d, sizeof(__pyx_k_sDIM_d_d), 0, 1, 0, 0},
  {&__pyx_kp_u_sINPUTNAME, __pyx_k_sINPUTNAME, sizeof(__pyx_k_sINPUTNAME), 0, 1, 0, 0},
  {&__pyx_kp_u_sINPUTS_d, __pyx_k_sINPUTS_d, sizeof(__pyx_k_sINPUTS_d), 0, 1, 0, 0},
  {&__pyx_kp_u_sMATRIXES_d, __pyx_k_sMATRIXES_d, sizeof(__pyx_k_sMATRIXES_d), 0, 1, 0, 0},
  {&__pyx_kp_u_sMATRIX_K_GRADIENT_J, __pyx_k_sMATRIX_K_GRADIENT_J, sizeof(__pyx_k_sMATRIX_K_GRADIENT_J), 0, 1, 0, 0},
  {&__pyx_kp_u_sMEMSYN_d, __pyx_k_sMEMSYN_d, sizeof(__pyx_k_sMEMSYN_d), 0, 1, 0, 0},
  {&__pyx_kp_u_sNODES_d, __pyx_k_sNODES_d, sizeof(__pyx_k_sNODES_d), 0, 1, 0, 0},
  {&__pyx_kp_u_sORDER_d, __pyx_k_sORDER_d, sizeof(__pyx_k_sORDER_d), 0, 1, 0, 0},
  {&__pyx_kp_u_sOUTPUTNAME, __pyx_k_sOUTPUTNAME, sizeof(__pyx_k_sOUTPUTNAME), 0, 1, 0, 0},
  {&__pyx_kp_u_sOUTPUTS_d, __pyx_k_sOUTPUTS_d, sizeof(__pyx_k_sOUTPUTS_d), 0, 1, 0, 0},
  {&__pyx_kp_u_sSTDTRDEV_d_g, __pyx_k_sSTDTRDEV_d_g, sizeof(__pyx_k_sSTDTRDEV_d_g), 0, 1, 0, 0},
  {&__pyx_kp_u_sSTDTRDEV_g, __pyx_k_sSTDTRDEV_g, sizeof(__pyx_k_sSTDTRDEV_g), 0, 1, 0, 0},
  {&__pyx_kp_u_sSTUDENT95_g, __pyx_k_sSTUDENT95_g, sizeof(__pyx_k_sSTUDENT95_g), 0, 1, 0, 0},
  {&__pyx_kp_u_sSYNAPSES_d, __pyx_k_sSYNAPSES_d, sizeof(__pyx_k_sSYNAPSES_d), 0, 1, 0, 0},
  {&__pyx_kp_u_sTRANSFERGRADIENT_INPUTS_OUTPUT, __pyx_k_sTRANSFERGRADIENT_INPUTS_OUTPUT, sizeof(__pyx_k_sTRANSFERGRADIENT_INPUTS_OUTPUT), 0, 1, 0, 0},
  {&__pyx_kp_u_setup_cfg, __pyx_k_setup_cfg, sizeof(__pyx_k_setup_cfg), 0, 1, 0, 0},
  {&__pyx_n_u_singleinterfacetraindll, __pyx_k_singleinterfacetraindll, sizeof(__pyx_k_singleinterfacetraindll), 0, 1, 0, 1},
  {&__pyx_n_u_singletrainbase, __pyx_k_singletrainbase, sizeof(__pyx_k_singletrainbase), 0, 1, 0, 1},
  {&__pyx_kp_u_smatrix_k_gradient_j, __pyx_k_smatrix_k_gradient_j, sizeof(__pyx_k_smatrix_k_gradient_j), 0, 1, 0, 0},
  {&__pyx_kp_u_smatrix_mindex_i_j_gradient_j, __pyx_k_smatrix_mindex_i_j_gradient_j, sizeof(__pyx_k_smatrix_mindex_i_j_gradient_j), 0, 1, 0, 0},
  {&__pyx_n_s_stAdvertDate, __pyx_k_stAdvertDate, sizeof(__pyx_k_stAdvertDate), 0, 0, 1, 1},
  {&__pyx_n_s_stAdvertMachine, __pyx_k_stAdvertMachine, sizeof(__pyx_k_stAdvertMachine), 0, 0, 1, 1},
  {&__pyx_n_s_stAdvertTime, __pyx_k_stAdvertTime, sizeof(__pyx_k_stAdvertTime), 0, 0, 1, 1},
  {&__pyx_n_s_stAdvertUser, __pyx_k_stAdvertUser, sizeof(__pyx_k_stAdvertUser), 0, 0, 1, 1},
  {&__pyx_n_s_stAlert, __pyx_k_stAlert, sizeof(__pyx_k_stAlert), 0, 0, 1, 1},
  {&__pyx_n_s_stArrayMatrixesP, __pyx_k_stArrayMatrixesP, sizeof(__pyx_k_stArrayMatrixesP), 0, 0, 1, 1},
  {&__pyx_n_s_stArrayOfDoubleDef, __pyx_k_stArrayOfDoubleDef, sizeof(__pyx_k_stArrayOfDoubleDef), 0, 0, 1, 1},
  {&__pyx_n_s_stArrayRetroP, __pyx_k_stArrayRetroP, sizeof(__pyx_k_stArrayRetroP), 0, 0, 1, 1},
  {&__pyx_n_s_stConstLinks, __pyx_k_stConstLinks, sizeof(__pyx_k_stConstLinks), 0, 0, 1, 1},
  {&__pyx_n_s_stConstLinksF, __pyx_k_stConstLinksF, sizeof(__pyx_k_stConstLinksF), 0, 0, 1, 1},
  {&__pyx_n_s_stConstMatrix, __pyx_k_stConstMatrix, sizeof(__pyx_k_stConstMatrix), 0, 0, 1, 1},
  {&__pyx_n_s_stConstMatrixF, __pyx_k_stConstMatrixF, sizeof(__pyx_k_stConstMatrixF), 0, 0, 1, 1},
  {&__pyx_n_s_stConstNodes, __pyx_k_stConstNodes, sizeof(__pyx_k_stConstNodes), 0, 0, 1, 1},
  {&__pyx_n_s_stConstNodesF, __pyx_k_stConstNodesF, sizeof(__pyx_k_stConstNodesF), 0, 0, 1, 1},
  {&__pyx_n_s_stConstSynapses, __pyx_k_stConstSynapses, sizeof(__pyx_k_stConstSynapses), 0, 0, 1, 1},
  {&__pyx_n_s_stConstSynapsesF, __pyx_k_stConstSynapsesF, sizeof(__pyx_k_stConstSynapsesF), 0, 0, 1, 1},
  {&__pyx_n_s_stDefineDimension, __pyx_k_stDefineDimension, sizeof(__pyx_k_stDefineDimension), 0, 0, 1, 1},
  {&__pyx_n_s_stDefineDimn, __pyx_k_stDefineDimn, sizeof(__pyx_k_stDefineDimn), 0, 0, 1, 1},
  {&__pyx_n_s_stDefineLinks, __pyx_k_stDefineLinks, sizeof(__pyx_k_stDefineLinks), 0, 0, 1, 1},
  {&__pyx_n_s_stDefineMatrixes, __pyx_k_stDefineMatrixes, sizeof(__pyx_k_stDefineMatrixes), 0, 0, 1, 1},
  {&__pyx_n_s_stDefineNodes, __pyx_k_stDefineNodes, sizeof(__pyx_k_stDefineNodes), 0, 0, 1, 1},
  {&__pyx_n_s_stDefineRetros, __pyx_k_stDefineRetros, sizeof(__pyx_k_stDefineRetros), 0, 0, 1, 1},
  {&__pyx_n_s_stDefineSynapses, __pyx_k_stDefineSynapses, sizeof(__pyx_k_stDefineSynapses), 0, 0, 1, 1},
  {&__pyx_n_s_stGradientDeclF77, __pyx_k_stGradientDeclF77, sizeof(__pyx_k_stGradientDeclF77), 0, 0, 1, 1},
  {&__pyx_n_s_stGradientDeclF90, __pyx_k_stGradientDeclF90, sizeof(__pyx_k_stGradientDeclF90), 0, 0, 1, 1},
  {&__pyx_n_s_stGradientDeclF90S, __pyx_k_stGradientDeclF90S, sizeof(__pyx_k_stGradientDeclF90S), 0, 0, 1, 1},
  {&__pyx_n_s_stGradientdS, __pyx_k_stGradientdS, sizeof(__pyx_k_stGradientdS), 0, 0, 1, 1},
  {&__pyx_n_s_stGradientdSP, __pyx_k_stGradientdSP, sizeof(__pyx_k_stGradientdSP), 0, 0, 1, 1},
  {&__pyx_n_s_stIncludeFile, __pyx_k_stIncludeFile, sizeof(__pyx_k_stIncludeFile), 0, 0, 1, 1},
  {&__pyx_n_s_stIncludeFileP, __pyx_k_stIncludeFileP, sizeof(__pyx_k_stIncludeFileP), 0, 0, 1, 1},
  {&__pyx_n_s_stInibackDeclF77, __pyx_k_stInibackDeclF77, sizeof(__pyx_k_stInibackDeclF77), 0, 0, 1, 1},
  {&__pyx_n_s_stInibackDeclF90, __pyx_k_stInibackDeclF90, sizeof(__pyx_k_stInibackDeclF90), 0, 0, 1, 1},
  {&__pyx_n_s_stInibackDeclF90S, __pyx_k_stInibackDeclF90S, sizeof(__pyx_k_stInibackDeclF90S), 0, 0, 1, 1},
  {&__pyx_n_s_stLeverageDeclF77, __pyx_k_stLeverageDeclF77, sizeof(__pyx_k_stLeverageDeclF77), 0, 0, 1, 1},
  {&__pyx_n_s_stLeverageDeclF90, __pyx_k_stLeverageDeclF90, sizeof(__pyx_k_stLeverageDeclF90), 0, 0, 1, 1},
  {&__pyx_n_s_stLeverageDeclF90S, __pyx_k_stLeverageDeclF90S, sizeof(__pyx_k_stLeverageDeclF90S), 0, 0, 1, 1},
  {&__pyx_n_s_stMemSyndS, __pyx_k_stMemSyndS, sizeof(__pyx_k_stMemSyndS), 0, 0, 1, 1},
  {&__pyx_n_s_stMemSyndSF77, __pyx_k_stMemSyndSF77, sizeof(__pyx_k_stMemSyndSF77), 0, 0, 1, 1},
  {&__pyx_n_s_stMemSyndSF90, __pyx_k_stMemSyndSF90, sizeof(__pyx_k_stMemSyndSF90), 0, 0, 1, 1},
  {&__pyx_n_s_stMemSyndSP, __pyx_k_stMemSyndSP, sizeof(__pyx_k_stMemSyndSP), 0, 0, 1, 1},
  {&__pyx_n_s_stNodedS, __pyx_k_stNodedS, sizeof(__pyx_k_stNodedS), 0, 0, 1, 1},
  {&__pyx_n_s_stNodedS2, __pyx_k_stNodedS2, sizeof(__pyx_k_stNodedS2), 0, 0, 1, 1},
  {&__pyx_n_s_stNodedS2F77, __pyx_k_stNodedS2F77, sizeof(__pyx_k_stNodedS2F77), 0, 0, 1, 1},
  {&__pyx_n_s_stNodedS2F90, __pyx_k_stNodedS2F90, sizeof(__pyx_k_stNodedS2F90), 0, 0, 1, 1},
  {&__pyx_n_s_stNodedS3, __pyx_k_stNodedS3, sizeof(__pyx_k_stNodedS3), 0, 0, 1, 1},
  {&__pyx_n_s_stNodedS3F77, __pyx_k_stNodedS3F77, sizeof(__pyx_k_stNodedS3F77), 0, 0, 1, 1},
  {&__pyx_n_s_stNodedS3F90, __pyx_k_stNodedS3F90, sizeof(__pyx_k_stNodedS3F90), 0, 0, 1, 1},
  {&__pyx_n_s_stNodedSF77, __pyx_k_stNodedSF77, sizeof(__pyx_k_stNodedSF77), 0, 0, 1, 1},
  {&__pyx_n_s_stNodedSF90, __pyx_k_stNodedSF90, sizeof(__pyx_k_stNodedSF90), 0, 0, 1, 1},
  {&__pyx_n_s_stNodedSK, __pyx_k_stNodedSK, sizeof(__pyx_k_stNodedSK), 0, 0, 1, 1},
  {&__pyx_n_s_stNodedSP, __pyx_k_stNodedSP, sizeof(__pyx_k_stNodedSP), 0, 0, 1, 1},
  {&__pyx_n_s_stNodedSP2, __pyx_k_stNodedSP2, sizeof(__pyx_k_stNodedSP2), 0, 0, 1, 1},
  {&__pyx_n_s_stNodedSP3, __pyx_k_stNodedSP3, sizeof(__pyx_k_stNodedSP3), 0, 0, 1, 1},
  {&__pyx_n_s_stNodedSPK, __pyx_k_stNodedSPK, sizeof(__pyx_k_stNodedSPK), 0, 0, 1, 1},
  {&__pyx_n_s_stPArrayOfDoubleDef, __pyx_k_stPArrayOfDoubleDef, sizeof(__pyx_k_stPArrayOfDoubleDef), 0, 0, 1, 1},
  {&__pyx_n_s_stPotentialdSF77, __pyx_k_stPotentialdSF77, sizeof(__pyx_k_stPotentialdSF77), 0, 0, 1, 1},
  {&__pyx_n_s_stPotentialdSF90, __pyx_k_stPotentialdSF90, sizeof(__pyx_k_stPotentialdSF90), 0, 0, 1, 1},
  {&__pyx_n_s_stPotentialdSP, __pyx_k_stPotentialdSP, sizeof(__pyx_k_stPotentialdSP), 0, 0, 1, 1},
  {&__pyx_n_s_stStatesDeclF77, __pyx_k_stStatesDeclF77, sizeof(__pyx_k_stStatesDeclF77), 0, 0, 1, 1},
  {&__pyx_n_s_stStatesDeclF90, __pyx_k_stStatesDeclF90, sizeof(__pyx_k_stStatesDeclF90), 0, 0, 1, 1},
  {&__pyx_n_s_stTypeMatrixes, __pyx_k_stTypeMatrixes, sizeof(__pyx_k_stTypeMatrixes), 0, 0, 1, 1},
  {&__pyx_n_s_stTypeMinMax, __pyx_k_stTypeMinMax, sizeof(__pyx_k_stTypeMinMax), 0, 0, 1, 1},
  {&__pyx_n_s_stTypeNodes, __pyx_k_stTypeNodes, sizeof(__pyx_k_stTypeNodes), 0, 0, 1, 1},
  {&__pyx_n_s_stTypeOutput, __pyx_k_stTypeOutput, sizeof(__pyx_k_stTypeOutput), 0, 0, 1, 1},
  {&__pyx_n_s_stTypeRetroP, __pyx_k_stTypeRetroP, sizeof(__pyx_k_stTypeRetroP), 0, 0, 1, 1},
  {&__pyx_n_s_stTypeSynapse, __pyx_k_stTypeSynapse, sizeof(__pyx_k_stTypeSynapse), 0, 0, 1, 1},
  {&__pyx_n_s_stXPotentialdSF77, __pyx_k_stXPotentialdSF77, sizeof(__pyx_k_stXPotentialdSF77), 0, 0, 1, 1},
  {&__pyx_n_s_stXPotentialdSF90, __pyx_k_stXPotentialdSF90, sizeof(__pyx_k_stXPotentialdSF90), 0, 0, 1, 1},
  {&__pyx_n_s_stXPotentialdSP, __pyx_k_stXPotentialdSP, sizeof(__pyx_k_stXPotentialdSP), 0, 0, 1, 1},
  {&__pyx_n_s_stXdSF77, __pyx_k_stXdSF77, sizeof(__pyx_k_stXdSF77), 0, 0, 1, 1},
  {&__pyx_n_s_stXdSF90, __pyx_k_stXdSF90, sizeof(__pyx_k_stXdSF90), 0, 0, 1, 1},
  {&__pyx_n_s_st_InDev, __pyx_k_st_InDev, sizeof(__pyx_k_st_InDev), 0, 0, 1, 1},
  {&__pyx_n_s_st_Juridic, __pyx_k_st_Juridic, sizeof(__pyx_k_st_Juridic), 0, 0, 1, 1},
  {&__pyx_kp_u_static_char_names_sINPUTS, __pyx_k_static_char_names_sINPUTS, sizeof(__pyx_k_static_char_names_sINPUTS), 0, 1, 0, 0},
  {&__pyx_kp_u_static_char_names_sOUTPUTS, __pyx_k_static_char_names_sOUTPUTS, sizeof(__pyx_k_static_char_names_sOUTPUTS), 0, 1, 0, 0},
  {&__pyx_n_s_std_re, __pyx_k_std_re, sizeof(__pyx_k_std_re), 0, 0, 1, 1},
  {&__pyx_kp_u_stddev_sOUTPUTS, __pyx_k_stddev_sOUTPUTS, sizeof(__pyx_k_stddev_sOUTPUTS), 0, 1, 0, 0},
  {&__pyx_kp_u_stransfergradient_inputs_output, __pyx_k_stransfergradient_inputs_output, sizeof(__pyx_k_stransfergradient_inputs_output), 0, 1, 0, 0},
  {&__pyx_kp_u_stransfergradient_inputs_output_2, __pyx_k_stransfergradient_inputs_output_2, sizeof(__pyx_k_stransfergradient_inputs_output_2), 0, 1, 0, 0},
  {&__pyx_kp_u_stransfergradient_inputs_output_3, __pyx_k_stransfergradient_inputs_output_3, sizeof(__pyx_k_stransfergradient_inputs_output_3), 0, 1, 0, 0},
  {&__pyx_kp_u_stransfergradient_inputs_output_4, __pyx_k_stransfergradient_inputs_output_4, sizeof(__pyx_k_stransfergradient_inputs_output_4), 0, 1, 0, 0},
  {&__pyx_n_s_subpattern, __pyx_k_subpattern, sizeof(__pyx_k_subpattern), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_u_train, __pyx_k_train, sizeof(__pyx_k_train), 0, 1, 0, 1},
  {&__pyx_n_u_trainbase, __pyx_k_trainbase, sizeof(__pyx_k_trainbase), 0, 1, 0, 1},
  {&__pyx_kp_u_tre_utilis_de_faon_permanente_su, __pyx_k_tre_utilis_de_faon_permanente_su, sizeof(__pyx_k_tre_utilis_de_faon_permanente_su), 0, 1, 0, 0},
  {&__pyx_n_s_upper_re, __pyx_k_upper_re, sizeof(__pyx_k_upper_re), 0, 0, 1, 1},
  {&__pyx_n_u_usebase, __pyx_k_usebase, sizeof(__pyx_k_usebase), 0, 1, 0, 1},
  {&__pyx_n_s_useversion, __pyx_k_useversion, sizeof(__pyx_k_useversion), 0, 0, 1, 1},
  {&__pyx_n_s_version, __pyx_k_version, sizeof(__pyx_k_version), 0, 0, 1, 1},
  {&__pyx_kp_u_void_stransfer_double_double, __pyx_k_void_stransfer_double_double, sizeof(__pyx_k_void_stransfer_double_double), 0, 1, 0, 0},
  {&__pyx_kp_u_void_stransfer_double_double_dou, __pyx_k_void_stransfer_double_double_dou, sizeof(__pyx_k_void_stransfer_double_double_dou), 0, 1, 0, 0},
  {&__pyx_kp_u_void_stransfer_double_inputs_dou, __pyx_k_void_stransfer_double_inputs_dou, sizeof(__pyx_k_void_stransfer_double_inputs_dou), 0, 1, 0, 0},
  {&__pyx_kp_u_void_stransfer_double_inputs_dou_2, __pyx_k_void_stransfer_double_inputs_dou_2, sizeof(__pyx_k_void_stransfer_double_inputs_dou_2), 0, 1, 0, 0},
  {&__pyx_kp_u_void_stransfergradient_double_do, __pyx_k_void_stransfergradient_double_do, sizeof(__pyx_k_void_stransfergradient_double_do), 0, 1, 0, 0},
  {&__pyx_kp_u_void_stransfergradient_double_do_2, __pyx_k_void_stransfergradient_double_do_2, sizeof(__pyx_k_void_stransfergradient_double_do_2), 0, 1, 0, 0},
  {&__pyx_kp_u_void_stransfergradient_double_in, __pyx_k_void_stransfergradient_double_in, sizeof(__pyx_k_void_stransfergradient_double_in), 0, 1, 0, 0},
  {&__pyx_kp_u_void_stransfergradient_double_in_2, __pyx_k_void_stransfergradient_double_in_2, sizeof(__pyx_k_void_stransfergradient_double_in_2), 0, 1, 0, 0},
  {&__pyx_kp_u_void_stransferlev_double_double, __pyx_k_void_stransferlev_double_double, sizeof(__pyx_k_void_stransferlev_double_double), 0, 1, 0, 0},
  {&__pyx_kp_u_void_stransferlev_double_double_2, __pyx_k_void_stransferlev_double_double_2, sizeof(__pyx_k_void_stransferlev_double_double_2), 0, 1, 0, 0},
  {&__pyx_kp_u_void_stransferlev_double_inputs, __pyx_k_void_stransferlev_double_inputs, sizeof(__pyx_k_void_stransferlev_double_inputs), 0, 1, 0, 0},
  {&__pyx_kp_u_void_stransferlev_double_weights, __pyx_k_void_stransferlev_double_weights, sizeof(__pyx_k_void_stransferlev_double_weights), 0, 1, 0, 0},
  {&__pyx_n_s_vtMax, __pyx_k_vtMax, sizeof(__pyx_k_vtMax), 0, 0, 1, 1},
  {&__pyx_n_s_vtMin, __pyx_k_vtMin, sizeof(__pyx_k_vtMin), 0, 0, 1, 1},
  {&__pyx_n_s_vtStdDev, __pyx_k_vtStdDev, sizeof(__pyx_k_vtStdDev), 0, 0, 1, 1},
  {&__pyx_n_s_vtWeight, __pyx_k_vtWeight, sizeof(__pyx_k_vtWeight), 0, 0, 1, 1},
  {&__pyx_kp_u_weights_d, __pyx_k_weights_d, sizeof(__pyx_k_weights_d), 0, 1, 0, 0},
  {&__pyx_kp_u_weights_s, __pyx_k_weights_s, sizeof(__pyx_k_weights_s), 0, 1, 0, 0},
  {&__pyx_kp_u_weights_sSYNAPSES, __pyx_k_weights_sSYNAPSES, sizeof(__pyx_k_weights_sSYNAPSES), 0, 1, 0, 0},
  {&__pyx_kp_u_weights_weightsredir_d, __pyx_k_weights_weightsredir_d, sizeof(__pyx_k_weights_weightsredir_d), 0, 1, 0, 0},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 103, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "metaphor/monal/lcode/codeconst.py":103
 * 
 * #tNeuronPos =
 * (npConst, npInput, npState, npHidden, npOutput, npOutState) = list(range(6))             # <<<<<<<<<<<<<<
 * 
 * code_Header = 0x100
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_int_6); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "metaphor/monal/lcode/codeconst.py":124
 * ftusedll, # 14
 * ftsinglebasetrain, # 15
 * ftsingletraindll) = list(range(17))  # 16             # <<<<<<<<<<<<<<
 * 
 * ftMax = ftSetupPy
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_int_17); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "metaphor/monal/lcode/codeconst.py":139
 * ftdlmdriverhelp,  # = 30
 * ftMake,
 * ftsetupcfg) = list(range(21, 33))             # <<<<<<<<<<<<<<
 * 
 * code_common_mini = ftntdefines
 */
  __pyx_tuple__3 = PyTuple_Pack(2, __pyx_int_21, __pyx_int_33); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "metaphor/monal/lcode/codeconst.py":170
 * ftOnlyHeader = [ftntdefines]
 * ftSpecialSub = [ftbasetrain]
 * (vtWeight, vtMin, vtMax, vtStdDev) = list(range(4))             # <<<<<<<<<<<<<<
 * (nkStat, nkDyn) = list(range(2))  #, nkKoh
 * 
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_int_4); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "metaphor/monal/lcode/codeconst.py":171
 * ftSpecialSub = [ftbasetrain]
 * (vtWeight, vtMin, vtMax, vtStdDev) = list(range(4))
 * (nkStat, nkDyn) = list(range(2))  #, nkKoh             # <<<<<<<<<<<<<<
 * 
 * codename = {
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_int_2); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "metaphor/monal/lcode/codeconst.py":217
 * PyCodeSingleTrain = [ftsinglebasetrain, ftsingletraindll]
 * 
 * (nncNode, nncName) = list(range(2))             # <<<<<<<<<<<<<<
 * 
 * (ncNone, ncSigma, ncPi) = list(range(3))  #, ncSigma2, ncSigmaPlus
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_int_2); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "metaphor/monal/lcode/codeconst.py":219
 * (nncNode, nncName) = list(range(2))
 * 
 * (ncNone, ncSigma, ncPi) = list(range(3))  #, ncSigma2, ncSigmaPlus             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_int_3); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "metaphor/monal/lcode/codeconst.py":223
 * 
 * (_GetLicenseNb, _GetUserName, _GetComputerName, _GetFileName,
 *     _GetStudent95, _GetUseCount) = list(range(6))             # <<<<<<<<<<<<<<
 * 
 * (ctUnknown, ctC, ctPascal, ctFortran77, ctFortran90, ctVBasic,
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_int_6); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "metaphor/monal/lcode/codeconst.py":226
 * 
 * (ctUnknown, ctC, ctPascal, ctFortran77, ctFortran90, ctVBasic,
 *     ctXML, ctExcel, ctPython, ctCython) = list(range(10))             # <<<<<<<<<<<<<<
 * 
 * cs_Common = 1
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_int_10); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 2, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_17 = PyInt_FromLong(17); if (unlikely(!__pyx_int_17)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_21 = PyInt_FromLong(21); if (unlikely(!__pyx_int_21)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_26 = PyInt_FromLong(26); if (unlikely(!__pyx_int_26)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_33 = PyInt_FromLong(33); if (unlikely(!__pyx_int_33)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_93 = PyInt_FromLong(93); if (unlikely(!__pyx_int_93)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_int_256 = PyInt_FromLong(256); if (unlikely(!__pyx_int_256)) __PYX_ERR(0, 2, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_modinit_global_init_code(void); /*proto*/
static int __Pyx_modinit_variable_export_code(void); /*proto*/
static int __Pyx_modinit_function_export_code(void); /*proto*/
static int __Pyx_modinit_type_init_code(void); /*proto*/
static int __Pyx_modinit_type_import_code(void); /*proto*/
static int __Pyx_modinit_variable_import_code(void); /*proto*/
static int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (!(defined(__cplusplus)) || (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ > 4)))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcodeconst(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcodeconst(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_codeconst(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_codeconst(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        result = PyDict_SetItemString(moddict, to_name, value);
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__") < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static int __pyx_pymod_exec_codeconst(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  int __pyx_t_24;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m && __pyx_m == __pyx_pyinit_module) return 0;
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_codeconst(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 2, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("codeconst", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 2, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 2, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 2, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 2, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_metaphor__monal__lcode__codeconst) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 2, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "metaphor.monal.lcode.codeconst")) {
      if (unlikely(PyDict_SetItemString(modules, "metaphor.monal.lcode.codeconst", __pyx_m) < 0)) __PYX_ERR(0, 2, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  (void)__Pyx_modinit_type_init_code();
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  #endif

  /* "metaphor/monal/lcode/codeconst.py":9
 * #===============================================================================
 * 
 * __version__ = 93             # <<<<<<<<<<<<<<
 * 
 * # 93 30-01-2020 add
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_version, __pyx_int_93) < 0) __PYX_ERR(0, 9, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":87
 * # 1  premiere introduction
 * 
 * __useversion__ = 26 #             # <<<<<<<<<<<<<<
 * # 25  add transfezleverageex()
 * # 24  add moduletype()
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_useversion, __pyx_int_26) < 0) __PYX_ERR(0, 87, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":96
 * # 19  introduction de useversion
 * 
 * KT_AFTER = 1             # <<<<<<<<<<<<<<
 * KT_BEFORE = 2
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_KT_AFTER, __pyx_int_1) < 0) __PYX_ERR(0, 96, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":97
 * 
 * KT_AFTER = 1
 * KT_BEFORE = 2             # <<<<<<<<<<<<<<
 * 
 * st_Juridic = 'jurgb.txt'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_KT_BEFORE, __pyx_int_2) < 0) __PYX_ERR(0, 97, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":99
 * KT_BEFORE = 2
 * 
 * st_Juridic = 'jurgb.txt'             # <<<<<<<<<<<<<<
 * #st_Juridic = 'jurFR.txt'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_st_Juridic, __pyx_kp_u_jurgb_txt) < 0) __PYX_ERR(0, 99, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":103
 * 
 * #tNeuronPos =
 * (npConst, npInput, npState, npHidden, npOutput, npOutState) = list(range(6))             # <<<<<<<<<<<<<<
 * 
 * code_Header = 0x100
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PySequence_List(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (1) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 6)) {
      if (size > 6) __Pyx_RaiseTooManyValuesError(6);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 103, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
    __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
    __pyx_t_5 = PyList_GET_ITEM(sequence, 3); 
    __pyx_t_6 = PyList_GET_ITEM(sequence, 4); 
    __pyx_t_7 = PyList_GET_ITEM(sequence, 5); 
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_7);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[6] = {&__pyx_t_1,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7};
      for (i=0; i < 6; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 103, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_npConst, __pyx_t_1) < 0) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_npInput, __pyx_t_3) < 0) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_npState, __pyx_t_4) < 0) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_npHidden, __pyx_t_5) < 0) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_npOutput, __pyx_t_6) < 0) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_npOutState, __pyx_t_7) < 0) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "metaphor/monal/lcode/codeconst.py":105
 * (npConst, npInput, npState, npHidden, npOutput, npOutState) = list(range(6))
 * 
 * code_Header = 0x100             # <<<<<<<<<<<<<<
 * code_Body = 0
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_code_Header, __pyx_int_256) < 0) __PYX_ERR(0, 105, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":106
 * 
 * code_Header = 0x100
 * code_Body = 0             # <<<<<<<<<<<<<<
 * 
 * (ftTransfert, # 0
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_code_Body, __pyx_int_0) < 0) __PYX_ERR(0, 106, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":124
 * ftusedll, # 14
 * ftsinglebasetrain, # 15
 * ftsingletraindll) = list(range(17))  # 16             # <<<<<<<<<<<<<<
 * 
 * ftMax = ftSetupPy
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_7 = PySequence_List(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (1) {
    PyObject* sequence = __pyx_t_7;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 17)) {
      if (size > 17) __Pyx_RaiseTooManyValuesError(17);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 108, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
    __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
    __pyx_t_5 = PyList_GET_ITEM(sequence, 2); 
    __pyx_t_4 = PyList_GET_ITEM(sequence, 3); 
    __pyx_t_3 = PyList_GET_ITEM(sequence, 4); 
    __pyx_t_1 = PyList_GET_ITEM(sequence, 5); 
    __pyx_t_8 = PyList_GET_ITEM(sequence, 6); 
    __pyx_t_9 = PyList_GET_ITEM(sequence, 7); 
    __pyx_t_10 = PyList_GET_ITEM(sequence, 8); 
    __pyx_t_11 = PyList_GET_ITEM(sequence, 9); 
    __pyx_t_12 = PyList_GET_ITEM(sequence, 10); 
    __pyx_t_13 = PyList_GET_ITEM(sequence, 11); 
    __pyx_t_14 = PyList_GET_ITEM(sequence, 12); 
    __pyx_t_15 = PyList_GET_ITEM(sequence, 13); 
    __pyx_t_16 = PyList_GET_ITEM(sequence, 14); 
    __pyx_t_17 = PyList_GET_ITEM(sequence, 15); 
    __pyx_t_18 = PyList_GET_ITEM(sequence, 16); 
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_t_14);
    __Pyx_INCREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_t_16);
    __Pyx_INCREF(__pyx_t_17);
    __Pyx_INCREF(__pyx_t_18);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[17] = {&__pyx_t_2,&__pyx_t_6,&__pyx_t_5,&__pyx_t_4,&__pyx_t_3,&__pyx_t_1,&__pyx_t_8,&__pyx_t_9,&__pyx_t_10,&__pyx_t_11,&__pyx_t_12,&__pyx_t_13,&__pyx_t_14,&__pyx_t_15,&__pyx_t_16,&__pyx_t_17,&__pyx_t_18};
      for (i=0; i < 17; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 108, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }

  /* "metaphor/monal/lcode/codeconst.py":108
 * code_Body = 0
 * 
 * (ftTransfert, # 0             # <<<<<<<<<<<<<<
 * ftGradient, # 1
 * ftTransfertLeverage,  # 2
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftTransfert, __pyx_t_2) < 0) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftGradient, __pyx_t_6) < 0) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftTransfertLeverage, __pyx_t_5) < 0) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftDLM, __pyx_t_4) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftMakeDLM, __pyx_t_3) < 0) __PYX_ERR(0, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftDesc, __pyx_t_1) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftmodeltrain, __pyx_t_8) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftSetupPy, __pyx_t_9) < 0) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftSetupMulti, __pyx_t_10) < 0) __PYX_ERR(0, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftBase, __pyx_t_11) < 0) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftbasetrain, __pyx_t_12) < 0) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fttrain, __pyx_t_13) < 0) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftbaseuse, __pyx_t_14) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fttraindll, __pyx_t_15) < 0) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftusedll, __pyx_t_16) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftsinglebasetrain, __pyx_t_17) < 0) __PYX_ERR(0, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftsingletraindll, __pyx_t_18) < 0) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "metaphor/monal/lcode/codeconst.py":126
 * ftsingletraindll) = list(range(17))  # 16
 * 
 * ftMax = ftSetupPy             # <<<<<<<<<<<<<<
 * 
 * (ftntdefines,  #21
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftSetupPy); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftMax, __pyx_t_7) < 0) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "metaphor/monal/lcode/codeconst.py":139
 * ftdlmdriverhelp,  # = 30
 * ftMake,
 * ftsetupcfg) = list(range(21, 33))             # <<<<<<<<<<<<<<
 * 
 * code_common_mini = ftntdefines
 */
  __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_18 = PySequence_List(__pyx_t_7); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (1) {
    PyObject* sequence = __pyx_t_18;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 12)) {
      if (size > 12) __Pyx_RaiseTooManyValuesError(12);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 128, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
    __pyx_t_17 = PyList_GET_ITEM(sequence, 1); 
    __pyx_t_16 = PyList_GET_ITEM(sequence, 2); 
    __pyx_t_15 = PyList_GET_ITEM(sequence, 3); 
    __pyx_t_14 = PyList_GET_ITEM(sequence, 4); 
    __pyx_t_13 = PyList_GET_ITEM(sequence, 5); 
    __pyx_t_12 = PyList_GET_ITEM(sequence, 6); 
    __pyx_t_11 = PyList_GET_ITEM(sequence, 7); 
    __pyx_t_10 = PyList_GET_ITEM(sequence, 8); 
    __pyx_t_9 = PyList_GET_ITEM(sequence, 9); 
    __pyx_t_8 = PyList_GET_ITEM(sequence, 10); 
    __pyx_t_1 = PyList_GET_ITEM(sequence, 11); 
    __Pyx_INCREF(__pyx_t_7);
    __Pyx_INCREF(__pyx_t_17);
    __Pyx_INCREF(__pyx_t_16);
    __Pyx_INCREF(__pyx_t_15);
    __Pyx_INCREF(__pyx_t_14);
    __Pyx_INCREF(__pyx_t_13);
    __Pyx_INCREF(__pyx_t_12);
    __Pyx_INCREF(__pyx_t_11);
    __Pyx_INCREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_t_9);
    __Pyx_INCREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_t_1);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[12] = {&__pyx_t_7,&__pyx_t_17,&__pyx_t_16,&__pyx_t_15,&__pyx_t_14,&__pyx_t_13,&__pyx_t_12,&__pyx_t_11,&__pyx_t_10,&__pyx_t_9,&__pyx_t_8,&__pyx_t_1};
      for (i=0; i < 12; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 128, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
  }

  /* "metaphor/monal/lcode/codeconst.py":128
 * ftMax = ftSetupPy
 * 
 * (ftntdefines,  #21             # <<<<<<<<<<<<<<
 * ftmathplus,  # = 22
 * ftreadfile,  # = 23
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftntdefines, __pyx_t_7) < 0) __PYX_ERR(0, 128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftmathplus, __pyx_t_17) < 0) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftreadfile, __pyx_t_16) < 0) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftnetraltype, __pyx_t_15) < 0) __PYX_ERR(0, 131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftnttoolbox, __pyx_t_14) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftdlmutil, __pyx_t_13) < 0) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftdlmuse, __pyx_t_12) < 0) __PYX_ERR(0, 134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftgenerictrain, __pyx_t_11) < 0) __PYX_ERR(0, 135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftdlmdriver, __pyx_t_10) < 0) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftdlmdriverhelp, __pyx_t_9) < 0) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftMake, __pyx_t_8) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftsetupcfg, __pyx_t_1) < 0) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metaphor/monal/lcode/codeconst.py":141
 * ftsetupcfg) = list(range(21, 33))
 * 
 * code_common_mini = ftntdefines             # <<<<<<<<<<<<<<
 * code_common_maxi = ftMake
 * 
 */
  __pyx_t_18 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftntdefines); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_code_common_mini, __pyx_t_18) < 0) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "metaphor/monal/lcode/codeconst.py":142
 * 
 * code_common_mini = ftntdefines
 * code_common_maxi = ftMake             # <<<<<<<<<<<<<<
 * 
 * commondict = {
 */
  __pyx_t_18 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftMake); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_code_common_maxi, __pyx_t_18) < 0) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "metaphor/monal/lcode/codeconst.py":145
 * 
 * commondict = {
 *     ftntdefines : "ntdefines",  #12             # <<<<<<<<<<<<<<
 *     ftmathplus : "mathplus",  # = 13
 *     ftreadfile : "readfile",  # = 14
 */
  __pyx_t_18 = __Pyx_PyDict_NewPresized(12); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftntdefines); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_18, __pyx_t_1, __pyx_n_u_ntdefines) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metaphor/monal/lcode/codeconst.py":146
 * commondict = {
 *     ftntdefines : "ntdefines",  #12
 *     ftmathplus : "mathplus",  # = 13             # <<<<<<<<<<<<<<
 *     ftreadfile : "readfile",  # = 14
 *     ftnetraltype : "nttype",  # = 15
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftmathplus); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_18, __pyx_t_1, __pyx_n_u_mathplus) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metaphor/monal/lcode/codeconst.py":147
 *     ftntdefines : "ntdefines",  #12
 *     ftmathplus : "mathplus",  # = 13
 *     ftreadfile : "readfile",  # = 14             # <<<<<<<<<<<<<<
 *     ftnetraltype : "nttype",  # = 15
 *     ftnttoolbox : "nttoolbox",  # = 16
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftreadfile); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_18, __pyx_t_1, __pyx_n_u_readfile) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metaphor/monal/lcode/codeconst.py":148
 *     ftmathplus : "mathplus",  # = 13
 *     ftreadfile : "readfile",  # = 14
 *     ftnetraltype : "nttype",  # = 15             # <<<<<<<<<<<<<<
 *     ftnttoolbox : "nttoolbox",  # = 16
 *     ftdlmutil : "dlmutil",  # = 17
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftnetraltype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_18, __pyx_t_1, __pyx_n_u_nttype) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metaphor/monal/lcode/codeconst.py":149
 *     ftreadfile : "readfile",  # = 14
 *     ftnetraltype : "nttype",  # = 15
 *     ftnttoolbox : "nttoolbox",  # = 16             # <<<<<<<<<<<<<<
 *     ftdlmutil : "dlmutil",  # = 17
 *     ftdlmuse : "dlmuse",  # = 18
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftnttoolbox); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_18, __pyx_t_1, __pyx_n_u_nttoolbox) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metaphor/monal/lcode/codeconst.py":150
 *     ftnetraltype : "nttype",  # = 15
 *     ftnttoolbox : "nttoolbox",  # = 16
 *     ftdlmutil : "dlmutil",  # = 17             # <<<<<<<<<<<<<<
 *     ftdlmuse : "dlmuse",  # = 18
 *     ftgenerictrain : "generictrain",  # = 19
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftdlmutil); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_18, __pyx_t_1, __pyx_n_u_dlmutil) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metaphor/monal/lcode/codeconst.py":151
 *     ftnttoolbox : "nttoolbox",  # = 16
 *     ftdlmutil : "dlmutil",  # = 17
 *     ftdlmuse : "dlmuse",  # = 18             # <<<<<<<<<<<<<<
 *     ftgenerictrain : "generictrain",  # = 19
 *     ftdlmdriver : "dlmdriver",  # = 20   #ftDemoTrain
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftdlmuse); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_18, __pyx_t_1, __pyx_n_u_dlmuse) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metaphor/monal/lcode/codeconst.py":152
 *     ftdlmutil : "dlmutil",  # = 17
 *     ftdlmuse : "dlmuse",  # = 18
 *     ftgenerictrain : "generictrain",  # = 19             # <<<<<<<<<<<<<<
 *     ftdlmdriver : "dlmdriver",  # = 20   #ftDemoTrain
 *     ftdlmdriverhelp : "dlmdriverhelp",  # = 21
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftgenerictrain); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_18, __pyx_t_1, __pyx_n_u_generictrain) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metaphor/monal/lcode/codeconst.py":153
 *     ftdlmuse : "dlmuse",  # = 18
 *     ftgenerictrain : "generictrain",  # = 19
 *     ftdlmdriver : "dlmdriver",  # = 20   #ftDemoTrain             # <<<<<<<<<<<<<<
 *     ftdlmdriverhelp : "dlmdriverhelp",  # = 21
 *     ftMake : "Make",
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftdlmdriver); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_18, __pyx_t_1, __pyx_n_u_dlmdriver) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metaphor/monal/lcode/codeconst.py":154
 *     ftgenerictrain : "generictrain",  # = 19
 *     ftdlmdriver : "dlmdriver",  # = 20   #ftDemoTrain
 *     ftdlmdriverhelp : "dlmdriverhelp",  # = 21             # <<<<<<<<<<<<<<
 *     ftMake : "Make",
 *     ftsetupcfg : "setup.cfg"}
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftdlmdriverhelp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_18, __pyx_t_1, __pyx_n_u_dlmdriverhelp) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metaphor/monal/lcode/codeconst.py":155
 *     ftdlmdriver : "dlmdriver",  # = 20   #ftDemoTrain
 *     ftdlmdriverhelp : "dlmdriverhelp",  # = 21
 *     ftMake : "Make",             # <<<<<<<<<<<<<<
 *     ftsetupcfg : "setup.cfg"}
 * 
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftMake); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_18, __pyx_t_1, __pyx_n_u_Make) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "metaphor/monal/lcode/codeconst.py":156
 *     ftdlmdriverhelp : "dlmdriverhelp",  # = 21
 *     ftMake : "Make",
 *     ftsetupcfg : "setup.cfg"}             # <<<<<<<<<<<<<<
 * 
 * #commonfiles = commondict.keys()
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftsetupcfg); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_18, __pyx_t_1, __pyx_kp_u_setup_cfg) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_commondict, __pyx_t_18) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

  /* "metaphor/monal/lcode/codeconst.py":160
 * #commonfiles = commondict.keys()
 * 
 * ftWithHeader = [ftTransfert, ftGradient, ftTransfertLeverage, ftDLM, ftmodeltrain,             # <<<<<<<<<<<<<<
 *                 ftntdefines, ftmathplus, ftreadfile, ftnetraltype, ftnttoolbox,
 *                 ftdlmutil, ftdlmuse, ftgenerictrain, ftdlmdriver, ftBase, ftbasetrain,
 */
  __pyx_t_18 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftTransfert); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftGradient); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftTransfertLeverage); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftDLM); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftmodeltrain); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "metaphor/monal/lcode/codeconst.py":161
 * 
 * ftWithHeader = [ftTransfert, ftGradient, ftTransfertLeverage, ftDLM, ftmodeltrain,
 *                 ftntdefines, ftmathplus, ftreadfile, ftnetraltype, ftnttoolbox,             # <<<<<<<<<<<<<<
 *                 ftdlmutil, ftdlmuse, ftgenerictrain, ftdlmdriver, ftBase, ftbasetrain,
 *                 fttrain, ftbaseuse, fttraindll, ftusedll, ftsingletraindll,
 */
  __pyx_t_11 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftntdefines); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftmathplus); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftreadfile); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_14 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftnetraltype); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftnttoolbox); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);

  /* "metaphor/monal/lcode/codeconst.py":162
 * ftWithHeader = [ftTransfert, ftGradient, ftTransfertLeverage, ftDLM, ftmodeltrain,
 *                 ftntdefines, ftmathplus, ftreadfile, ftnetraltype, ftnttoolbox,
 *                 ftdlmutil, ftdlmuse, ftgenerictrain, ftdlmdriver, ftBase, ftbasetrain,             # <<<<<<<<<<<<<<
 *                 fttrain, ftbaseuse, fttraindll, ftusedll, ftsingletraindll,
 *                 ftsinglebasetrain]
 */
  __pyx_t_16 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftdlmutil); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftdlmuse); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftgenerictrain); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftdlmdriver); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftBase); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftbasetrain); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "metaphor/monal/lcode/codeconst.py":163
 *                 ftntdefines, ftmathplus, ftreadfile, ftnetraltype, ftnttoolbox,
 *                 ftdlmutil, ftdlmuse, ftgenerictrain, ftdlmdriver, ftBase, ftbasetrain,
 *                 fttrain, ftbaseuse, fttraindll, ftusedll, ftsingletraindll,             # <<<<<<<<<<<<<<
 *                 ftsinglebasetrain]
 * 
 */
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_fttrain); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftbaseuse); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_19 = __Pyx_GetModuleGlobalName(__pyx_n_s_fttraindll); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftusedll); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_21 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftsingletraindll); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);

  /* "metaphor/monal/lcode/codeconst.py":164
 *                 ftdlmutil, ftdlmuse, ftgenerictrain, ftdlmdriver, ftBase, ftbasetrain,
 *                 fttrain, ftbaseuse, fttraindll, ftusedll, ftsingletraindll,
 *                 ftsinglebasetrain]             # <<<<<<<<<<<<<<
 * 
 * ftwithCext = ftWithHeader
 */
  __pyx_t_22 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftsinglebasetrain); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);

  /* "metaphor/monal/lcode/codeconst.py":160
 * #commonfiles = commondict.keys()
 * 
 * ftWithHeader = [ftTransfert, ftGradient, ftTransfertLeverage, ftDLM, ftmodeltrain,             # <<<<<<<<<<<<<<
 *                 ftntdefines, ftmathplus, ftreadfile, ftnetraltype, ftnttoolbox,
 *                 ftdlmutil, ftdlmuse, ftgenerictrain, ftdlmdriver, ftBase, ftbasetrain,
 */
  __pyx_t_23 = PyList_New(22); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_18);
  PyList_SET_ITEM(__pyx_t_23, 0, __pyx_t_18);
  __Pyx_GIVEREF(__pyx_t_1);
  PyList_SET_ITEM(__pyx_t_23, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_8);
  PyList_SET_ITEM(__pyx_t_23, 2, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyList_SET_ITEM(__pyx_t_23, 3, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyList_SET_ITEM(__pyx_t_23, 4, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyList_SET_ITEM(__pyx_t_23, 5, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyList_SET_ITEM(__pyx_t_23, 6, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyList_SET_ITEM(__pyx_t_23, 7, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyList_SET_ITEM(__pyx_t_23, 8, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_15);
  PyList_SET_ITEM(__pyx_t_23, 9, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_16);
  PyList_SET_ITEM(__pyx_t_23, 10, __pyx_t_16);
  __Pyx_GIVEREF(__pyx_t_17);
  PyList_SET_ITEM(__pyx_t_23, 11, __pyx_t_17);
  __Pyx_GIVEREF(__pyx_t_7);
  PyList_SET_ITEM(__pyx_t_23, 12, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_23, 13, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyList_SET_ITEM(__pyx_t_23, 14, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_23, 15, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_23, 16, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_23, 17, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_19);
  PyList_SET_ITEM(__pyx_t_23, 18, __pyx_t_19);
  __Pyx_GIVEREF(__pyx_t_20);
  PyList_SET_ITEM(__pyx_t_23, 19, __pyx_t_20);
  __Pyx_GIVEREF(__pyx_t_21);
  PyList_SET_ITEM(__pyx_t_23, 20, __pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_22);
  PyList_SET_ITEM(__pyx_t_23, 21, __pyx_t_22);
  __pyx_t_18 = 0;
  __pyx_t_1 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_t_15 = 0;
  __pyx_t_16 = 0;
  __pyx_t_17 = 0;
  __pyx_t_7 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_2 = 0;
  __pyx_t_19 = 0;
  __pyx_t_20 = 0;
  __pyx_t_21 = 0;
  __pyx_t_22 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftWithHeader, __pyx_t_23) < 0) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "metaphor/monal/lcode/codeconst.py":166
 *                 ftsinglebasetrain]
 * 
 * ftwithCext = ftWithHeader             # <<<<<<<<<<<<<<
 * 
 * ftOnlyHeader = [ftntdefines]
 */
  __pyx_t_23 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftWithHeader); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftwithCext, __pyx_t_23) < 0) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "metaphor/monal/lcode/codeconst.py":168
 * ftwithCext = ftWithHeader
 * 
 * ftOnlyHeader = [ftntdefines]             # <<<<<<<<<<<<<<
 * ftSpecialSub = [ftbasetrain]
 * (vtWeight, vtMin, vtMax, vtStdDev) = list(range(4))
 */
  __pyx_t_23 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftntdefines); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_22 = PyList_New(1); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_GIVEREF(__pyx_t_23);
  PyList_SET_ITEM(__pyx_t_22, 0, __pyx_t_23);
  __pyx_t_23 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftOnlyHeader, __pyx_t_22) < 0) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "metaphor/monal/lcode/codeconst.py":169
 * 
 * ftOnlyHeader = [ftntdefines]
 * ftSpecialSub = [ftbasetrain]             # <<<<<<<<<<<<<<
 * (vtWeight, vtMin, vtMax, vtStdDev) = list(range(4))
 * (nkStat, nkDyn) = list(range(2))  #, nkKoh
 */
  __pyx_t_22 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftbasetrain); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_23 = PyList_New(1); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_22);
  PyList_SET_ITEM(__pyx_t_23, 0, __pyx_t_22);
  __pyx_t_22 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ftSpecialSub, __pyx_t_23) < 0) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "metaphor/monal/lcode/codeconst.py":170
 * ftOnlyHeader = [ftntdefines]
 * ftSpecialSub = [ftbasetrain]
 * (vtWeight, vtMin, vtMax, vtStdDev) = list(range(4))             # <<<<<<<<<<<<<<
 * (nkStat, nkDyn) = list(range(2))  #, nkKoh
 * 
 */
  __pyx_t_23 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_22 = PySequence_List(__pyx_t_23); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  if (1) {
    PyObject* sequence = __pyx_t_22;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 4)) {
      if (size > 4) __Pyx_RaiseTooManyValuesError(4);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 170, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_23 = PyList_GET_ITEM(sequence, 0); 
    __pyx_t_21 = PyList_GET_ITEM(sequence, 1); 
    __pyx_t_20 = PyList_GET_ITEM(sequence, 2); 
    __pyx_t_19 = PyList_GET_ITEM(sequence, 3); 
    __Pyx_INCREF(__pyx_t_23);
    __Pyx_INCREF(__pyx_t_21);
    __Pyx_INCREF(__pyx_t_20);
    __Pyx_INCREF(__pyx_t_19);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[4] = {&__pyx_t_23,&__pyx_t_21,&__pyx_t_20,&__pyx_t_19};
      for (i=0; i < 4; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 170, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  }
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vtWeight, __pyx_t_23) < 0) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vtMin, __pyx_t_21) < 0) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vtMax, __pyx_t_20) < 0) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vtStdDev, __pyx_t_19) < 0) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "metaphor/monal/lcode/codeconst.py":171
 * ftSpecialSub = [ftbasetrain]
 * (vtWeight, vtMin, vtMax, vtStdDev) = list(range(4))
 * (nkStat, nkDyn) = list(range(2))  #, nkKoh             # <<<<<<<<<<<<<<
 * 
 * codename = {
 */
  __pyx_t_22 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_19 = PySequence_List(__pyx_t_22); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  if (1) {
    PyObject* sequence = __pyx_t_19;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 171, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_22 = PyList_GET_ITEM(sequence, 0); 
    __pyx_t_20 = PyList_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_22);
    __Pyx_INCREF(__pyx_t_20);
    #else
    __pyx_t_22 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_22);
    __pyx_t_20 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_20);
    #endif
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  }
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nkStat, __pyx_t_22) < 0) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nkDyn, __pyx_t_20) < 0) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":174
 * 
 * codename = {
 *             ftTransfert: "Transfert",             # <<<<<<<<<<<<<<
 *             ftGradient: "Gradient",
 *             ftTransfertLeverage: "TransfertLeverage",
 */
  __pyx_t_19 = __Pyx_PyDict_NewPresized(29); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftTransfert); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_Transfert) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":175
 * codename = {
 *             ftTransfert: "Transfert",
 *             ftGradient: "Gradient",             # <<<<<<<<<<<<<<
 *             ftTransfertLeverage: "TransfertLeverage",
 *             ftDLM: "DLM",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftGradient); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_Gradient) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":176
 *             ftTransfert: "Transfert",
 *             ftGradient: "Gradient",
 *             ftTransfertLeverage: "TransfertLeverage",             # <<<<<<<<<<<<<<
 *             ftDLM: "DLM",
 *             ftMakeDLM: "MakeDLM",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftTransfertLeverage); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_TransfertLeverage) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":177
 *             ftGradient: "Gradient",
 *             ftTransfertLeverage: "TransfertLeverage",
 *             ftDLM: "DLM",             # <<<<<<<<<<<<<<
 *             ftMakeDLM: "MakeDLM",
 *             ftDesc: "Desc",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftDLM); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_DLM) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":178
 *             ftTransfertLeverage: "TransfertLeverage",
 *             ftDLM: "DLM",
 *             ftMakeDLM: "MakeDLM",             # <<<<<<<<<<<<<<
 *             ftDesc: "Desc",
 *             ftmodeltrain: "modeltrain",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftMakeDLM); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_MakeDLM) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":179
 *             ftDLM: "DLM",
 *             ftMakeDLM: "MakeDLM",
 *             ftDesc: "Desc",             # <<<<<<<<<<<<<<
 *             ftmodeltrain: "modeltrain",
 *             #ftPyInterface: "PyInterfaceC",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftDesc); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_Desc) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":180
 *             ftMakeDLM: "MakeDLM",
 *             ftDesc: "Desc",
 *             ftmodeltrain: "modeltrain",             # <<<<<<<<<<<<<<
 *             #ftPyInterface: "PyInterfaceC",
 *             ftSetupPy: "SetupPy",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftmodeltrain); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_modeltrain) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":182
 *             ftmodeltrain: "modeltrain",
 *             #ftPyInterface: "PyInterfaceC",
 *             ftSetupPy: "SetupPy",             # <<<<<<<<<<<<<<
 *             ftSetupMulti: "PyInterfaceCMulti",
 *             #ftPyInterfaceMulti: "",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftSetupPy); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_SetupPy) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":183
 *             #ftPyInterface: "PyInterfaceC",
 *             ftSetupPy: "SetupPy",
 *             ftSetupMulti: "PyInterfaceCMulti",             # <<<<<<<<<<<<<<
 *             #ftPyInterfaceMulti: "",
 *             ftntdefines: "ntdefines",  #12
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftSetupMulti); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_PyInterfaceCMulti) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":185
 *             ftSetupMulti: "PyInterfaceCMulti",
 *             #ftPyInterfaceMulti: "",
 *             ftntdefines: "ntdefines",  #12             # <<<<<<<<<<<<<<
 *             ftmathplus: "mathplus",  # = 13
 *             ftreadfile: "readfile",  # = 14
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftntdefines); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_ntdefines) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":186
 *             #ftPyInterfaceMulti: "",
 *             ftntdefines: "ntdefines",  #12
 *             ftmathplus: "mathplus",  # = 13             # <<<<<<<<<<<<<<
 *             ftreadfile: "readfile",  # = 14
 *             ftnetraltype: "nttype",  # = 15
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftmathplus); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_mathplus) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":187
 *             ftntdefines: "ntdefines",  #12
 *             ftmathplus: "mathplus",  # = 13
 *             ftreadfile: "readfile",  # = 14             # <<<<<<<<<<<<<<
 *             ftnetraltype: "nttype",  # = 15
 *             ftnttoolbox: "nttoolbox",  # = 16
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftreadfile); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_readfile) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":188
 *             ftmathplus: "mathplus",  # = 13
 *             ftreadfile: "readfile",  # = 14
 *             ftnetraltype: "nttype",  # = 15             # <<<<<<<<<<<<<<
 *             ftnttoolbox: "nttoolbox",  # = 16
 *             ftdlmutil: "dlmutil",  # = 17
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftnetraltype); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_nttype) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":189
 *             ftreadfile: "readfile",  # = 14
 *             ftnetraltype: "nttype",  # = 15
 *             ftnttoolbox: "nttoolbox",  # = 16             # <<<<<<<<<<<<<<
 *             ftdlmutil: "dlmutil",  # = 17
 *             ftdlmuse: "dlmuse",  # = 18
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftnttoolbox); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_nttoolbox) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":190
 *             ftnetraltype: "nttype",  # = 15
 *             ftnttoolbox: "nttoolbox",  # = 16
 *             ftdlmutil: "dlmutil",  # = 17             # <<<<<<<<<<<<<<
 *             ftdlmuse: "dlmuse",  # = 18
 *             ftgenerictrain: "generictrain",  # = 19
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftdlmutil); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_dlmutil) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":191
 *             ftnttoolbox: "nttoolbox",  # = 16
 *             ftdlmutil: "dlmutil",  # = 17
 *             ftdlmuse: "dlmuse",  # = 18             # <<<<<<<<<<<<<<
 *             ftgenerictrain: "generictrain",  # = 19
 *             ftdlmdriver: "dlmdriver",  # = 20   #ftDemoTrain
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftdlmuse); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_dlmuse) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":192
 *             ftdlmutil: "dlmutil",  # = 17
 *             ftdlmuse: "dlmuse",  # = 18
 *             ftgenerictrain: "generictrain",  # = 19             # <<<<<<<<<<<<<<
 *             ftdlmdriver: "dlmdriver",  # = 20   #ftDemoTrain
 *             ftdlmdriverhelp: "dlmdriverhelp",  # = 21
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftgenerictrain); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_generictrain) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":193
 *             ftdlmuse: "dlmuse",  # = 18
 *             ftgenerictrain: "generictrain",  # = 19
 *             ftdlmdriver: "dlmdriver",  # = 20   #ftDemoTrain             # <<<<<<<<<<<<<<
 *             ftdlmdriverhelp: "dlmdriverhelp",  # = 21
 *             ftMake: "Make",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftdlmdriver); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_dlmdriver) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":194
 *             ftgenerictrain: "generictrain",  # = 19
 *             ftdlmdriver: "dlmdriver",  # = 20   #ftDemoTrain
 *             ftdlmdriverhelp: "dlmdriverhelp",  # = 21             # <<<<<<<<<<<<<<
 *             ftMake: "Make",
 *             ftsetupcfg: "Setupcfg",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftdlmdriverhelp); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_dlmdriverhelp) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":195
 *             ftdlmdriver: "dlmdriver",  # = 20   #ftDemoTrain
 *             ftdlmdriverhelp: "dlmdriverhelp",  # = 21
 *             ftMake: "Make",             # <<<<<<<<<<<<<<
 *             ftsetupcfg: "Setupcfg",
 *             ftBase: "base",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftMake); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_Make) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":196
 *             ftdlmdriverhelp: "dlmdriverhelp",  # = 21
 *             ftMake: "Make",
 *             ftsetupcfg: "Setupcfg",             # <<<<<<<<<<<<<<
 *             ftBase: "base",
 *             ftbasetrain: "trainbase",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftsetupcfg); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_Setupcfg) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":197
 *             ftMake: "Make",
 *             ftsetupcfg: "Setupcfg",
 *             ftBase: "base",             # <<<<<<<<<<<<<<
 *             ftbasetrain: "trainbase",
 *             ftsinglebasetrain: "singletrainbase",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftBase); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_base) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":198
 *             ftsetupcfg: "Setupcfg",
 *             ftBase: "base",
 *             ftbasetrain: "trainbase",             # <<<<<<<<<<<<<<
 *             ftsinglebasetrain: "singletrainbase",
 *             ftsingletraindll: 'singleinterfacetraindll',
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftbasetrain); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_trainbase) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":199
 *             ftBase: "base",
 *             ftbasetrain: "trainbase",
 *             ftsinglebasetrain: "singletrainbase",             # <<<<<<<<<<<<<<
 *             ftsingletraindll: 'singleinterfacetraindll',
 *             fttrain: "train",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftsinglebasetrain); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_singletrainbase) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":200
 *             ftbasetrain: "trainbase",
 *             ftsinglebasetrain: "singletrainbase",
 *             ftsingletraindll: 'singleinterfacetraindll',             # <<<<<<<<<<<<<<
 *             fttrain: "train",
 *             ftbaseuse: "usebase",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftsingletraindll); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_singleinterfacetraindll) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":201
 *             ftsinglebasetrain: "singletrainbase",
 *             ftsingletraindll: 'singleinterfacetraindll',
 *             fttrain: "train",             # <<<<<<<<<<<<<<
 *             ftbaseuse: "usebase",
 *             fttraindll: "interfacetraindll",
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_fttrain); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_train) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":202
 *             ftsingletraindll: 'singleinterfacetraindll',
 *             fttrain: "train",
 *             ftbaseuse: "usebase",             # <<<<<<<<<<<<<<
 *             fttraindll: "interfacetraindll",
 *             ftusedll: "interfaceusedll"}
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftbaseuse); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_usebase) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":203
 *             fttrain: "train",
 *             ftbaseuse: "usebase",
 *             fttraindll: "interfacetraindll",             # <<<<<<<<<<<<<<
 *             ftusedll: "interfaceusedll"}
 * 
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_fttraindll); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_interfacetraindll) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

  /* "metaphor/monal/lcode/codeconst.py":204
 *             ftbaseuse: "usebase",
 *             fttraindll: "interfacetraindll",
 *             ftusedll: "interfaceusedll"}             # <<<<<<<<<<<<<<
 * 
 * PyCodeList0 = [ftSetupPy, ftTransfert, ftGradient, ftTransfertLeverage, ftsetupcfg, ftnetraltype, ftnttoolbox, ftmathplus]  #ftPyInterface,
 */
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftusedll); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  if (PyDict_SetItem(__pyx_t_19, __pyx_t_20, __pyx_n_u_interfaceusedll) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_codename, __pyx_t_19) < 0) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;

  /* "metaphor/monal/lcode/codeconst.py":206
 *             ftusedll: "interfaceusedll"}
 * 
 * PyCodeList0 = [ftSetupPy, ftTransfert, ftGradient, ftTransfertLeverage, ftsetupcfg, ftnetraltype, ftnttoolbox, ftmathplus]  #ftPyInterface,             # <<<<<<<<<<<<<<
 * PyCodeList = [ftBase, ftbaseuse]
 * #PyCodeList = [ ftbaseuse]
 */
  __pyx_t_19 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftSetupPy); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_20 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftTransfert); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_20);
  __pyx_t_22 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftGradient); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_21 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftTransfertLeverage); if (unlikely(!__pyx_t_21)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_21);
  __pyx_t_23 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftsetupcfg); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftnetraltype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftnttoolbox); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftmathplus); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyList_New(8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_19);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_19);
  __Pyx_GIVEREF(__pyx_t_20);
  PyList_SET_ITEM(__pyx_t_4, 1, __pyx_t_20);
  __Pyx_GIVEREF(__pyx_t_22);
  PyList_SET_ITEM(__pyx_t_4, 2, __pyx_t_22);
  __Pyx_GIVEREF(__pyx_t_21);
  PyList_SET_ITEM(__pyx_t_4, 3, __pyx_t_21);
  __Pyx_GIVEREF(__pyx_t_23);
  PyList_SET_ITEM(__pyx_t_4, 4, __pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_4, 5, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_4, 6, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_4, 7, __pyx_t_5);
  __pyx_t_19 = 0;
  __pyx_t_20 = 0;
  __pyx_t_22 = 0;
  __pyx_t_21 = 0;
  __pyx_t_23 = 0;
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PyCodeList0, __pyx_t_4) < 0) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "metaphor/monal/lcode/codeconst.py":207
 * 
 * PyCodeList0 = [ftSetupPy, ftTransfert, ftGradient, ftTransfertLeverage, ftsetupcfg, ftnetraltype, ftnttoolbox, ftmathplus]  #ftPyInterface,
 * PyCodeList = [ftBase, ftbaseuse]             # <<<<<<<<<<<<<<
 * #PyCodeList = [ ftbaseuse]
 * PyCodeList2 = [ftbaseuse, ftusedll]
 */
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftBase); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftbaseuse); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyList_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_4);
  PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PyCodeList, __pyx_t_6) < 0) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "metaphor/monal/lcode/codeconst.py":209
 * PyCodeList = [ftBase, ftbaseuse]
 * #PyCodeList = [ ftbaseuse]
 * PyCodeList2 = [ftbaseuse, ftusedll]             # <<<<<<<<<<<<<<
 * #PyCodeList2 = [ftBase, ftbaseuse, ftusedll]
 * PyCodeListShort = [ftBase]
 */
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftbaseuse); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftusedll); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_4, 1, __pyx_t_5);
  __pyx_t_6 = 0;
  __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PyCodeList2, __pyx_t_4) < 0) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "metaphor/monal/lcode/codeconst.py":211
 * PyCodeList2 = [ftbaseuse, ftusedll]
 * #PyCodeList2 = [ftBase, ftbaseuse, ftusedll]
 * PyCodeListShort = [ftBase]             # <<<<<<<<<<<<<<
 * PyCodeListFull = [ftSetupPy, ftBase, ftsetupcfg]
 * PyCodeMulti = [ftbasetrain, fttraindll]
 */
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftBase); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyList_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyList_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
  __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PyCodeListShort, __pyx_t_5) < 0) __PYX_ERR(0, 211, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "metaphor/monal/lcode/codeconst.py":212
 * #PyCodeList2 = [ftBase, ftbaseuse, ftusedll]
 * PyCodeListShort = [ftBase]
 * PyCodeListFull = [ftSetupPy, ftBase, ftsetupcfg]             # <<<<<<<<<<<<<<
 * PyCodeMulti = [ftbasetrain, fttraindll]
 * PyCodeMultiFull = [ftbasetrain, fttraindll, fttrain, ftSetupMulti]
 */
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftSetupPy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftBase); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftsetupcfg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = PyList_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_t_6);
  __pyx_t_5 = 0;
  __pyx_t_4 = 0;
  __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PyCodeListFull, __pyx_t_2) < 0) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "metaphor/monal/lcode/codeconst.py":213
 * PyCodeListShort = [ftBase]
 * PyCodeListFull = [ftSetupPy, ftBase, ftsetupcfg]
 * PyCodeMulti = [ftbasetrain, fttraindll]             # <<<<<<<<<<<<<<
 * PyCodeMultiFull = [ftbasetrain, fttraindll, fttrain, ftSetupMulti]
 * PyCodeSingleTrain = [ftsinglebasetrain, ftsingletraindll]
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftbasetrain); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_fttraindll); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_4, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
  __pyx_t_2 = 0;
  __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PyCodeMulti, __pyx_t_4) < 0) __PYX_ERR(0, 213, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "metaphor/monal/lcode/codeconst.py":214
 * PyCodeListFull = [ftSetupPy, ftBase, ftsetupcfg]
 * PyCodeMulti = [ftbasetrain, fttraindll]
 * PyCodeMultiFull = [ftbasetrain, fttraindll, fttrain, ftSetupMulti]             # <<<<<<<<<<<<<<
 * PyCodeSingleTrain = [ftsinglebasetrain, ftsingletraindll]
 * 
 */
  __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftbasetrain); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_fttraindll); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_fttrain); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftSetupMulti); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_23 = PyList_New(4); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_4);
  PyList_SET_ITEM(__pyx_t_23, 0, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_6);
  PyList_SET_ITEM(__pyx_t_23, 1, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyList_SET_ITEM(__pyx_t_23, 2, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_23, 3, __pyx_t_5);
  __pyx_t_4 = 0;
  __pyx_t_6 = 0;
  __pyx_t_2 = 0;
  __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PyCodeMultiFull, __pyx_t_23) < 0) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "metaphor/monal/lcode/codeconst.py":215
 * PyCodeMulti = [ftbasetrain, fttraindll]
 * PyCodeMultiFull = [ftbasetrain, fttraindll, fttrain, ftSetupMulti]
 * PyCodeSingleTrain = [ftsinglebasetrain, ftsingletraindll]             # <<<<<<<<<<<<<<
 * 
 * (nncNode, nncName) = list(range(2))
 */
  __pyx_t_23 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftsinglebasetrain); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_ftsingletraindll); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_23);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_t_23);
  __Pyx_GIVEREF(__pyx_t_5);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
  __pyx_t_23 = 0;
  __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PyCodeSingleTrain, __pyx_t_2) < 0) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "metaphor/monal/lcode/codeconst.py":217
 * PyCodeSingleTrain = [ftsinglebasetrain, ftsingletraindll]
 * 
 * (nncNode, nncName) = list(range(2))             # <<<<<<<<<<<<<<
 * 
 * (ncNone, ncSigma, ncPi) = list(range(3))  #, ncSigma2, ncSigmaPlus
 */
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PySequence_List(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (1) {
    PyObject* sequence = __pyx_t_5;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 217, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
    __pyx_t_23 = PyList_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_23);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_23 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    #endif
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nncNode, __pyx_t_2) < 0) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nncName, __pyx_t_23) < 0) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;

  /* "metaphor/monal/lcode/codeconst.py":219
 * (nncNode, nncName) = list(range(2))
 * 
 * (ncNone, ncSigma, ncPi) = list(range(3))  #, ncSigma2, ncSigmaPlus             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_23 = PySequence_List(__pyx_t_5); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (1) {
    PyObject* sequence = __pyx_t_23;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 219, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
    __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_6);
    #else
    __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 219, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  }
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ncNone, __pyx_t_5) < 0) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ncSigma, __pyx_t_2) < 0) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ncPi, __pyx_t_6) < 0) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "metaphor/monal/lcode/codeconst.py":223
 * 
 * (_GetLicenseNb, _GetUserName, _GetComputerName, _GetFileName,
 *     _GetStudent95, _GetUseCount) = list(range(6))             # <<<<<<<<<<<<<<
 * 
 * (ctUnknown, ctC, ctPascal, ctFortran77, ctFortran90, ctVBasic,
 */
  __pyx_t_23 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_23);
  __pyx_t_6 = PySequence_List(__pyx_t_23); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  if (1) {
    PyObject* sequence = __pyx_t_6;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 6)) {
      if (size > 6) __Pyx_RaiseTooManyValuesError(6);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 222, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_23 = PyList_GET_ITEM(sequence, 0); 
    __pyx_t_2 = PyList_GET_ITEM(sequence, 1); 
    __pyx_t_5 = PyList_GET_ITEM(sequence, 2); 
    __pyx_t_4 = PyList_GET_ITEM(sequence, 3); 
    __pyx_t_21 = PyList_GET_ITEM(sequence, 4); 
    __pyx_t_22 = PyList_GET_ITEM(sequence, 5); 
    __Pyx_INCREF(__pyx_t_23);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_21);
    __Pyx_INCREF(__pyx_t_22);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[6] = {&__pyx_t_23,&__pyx_t_2,&__pyx_t_5,&__pyx_t_4,&__pyx_t_21,&__pyx_t_22};
      for (i=0; i < 6; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 222, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }

  /* "metaphor/monal/lcode/codeconst.py":222
 * 
 * 
 * (_GetLicenseNb, _GetUserName, _GetComputerName, _GetFileName,             # <<<<<<<<<<<<<<
 *     _GetStudent95, _GetUseCount) = list(range(6))
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GetLicenseNb, __pyx_t_23) < 0) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GetUserName, __pyx_t_2) < 0) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GetComputerName, __pyx_t_5) < 0) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GetFileName, __pyx_t_4) < 0) __PYX_ERR(0, 222, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GetStudent95, __pyx_t_21) < 0) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GetUseCount, __pyx_t_22) < 0) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "metaphor/monal/lcode/codeconst.py":226
 * 
 * (ctUnknown, ctC, ctPascal, ctFortran77, ctFortran90, ctVBasic,
 *     ctXML, ctExcel, ctPython, ctCython) = list(range(10))             # <<<<<<<<<<<<<<
 * 
 * cs_Common = 1
 */
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_22 = PySequence_List(__pyx_t_6); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (1) {
    PyObject* sequence = __pyx_t_22;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 10)) {
      if (size > 10) __Pyx_RaiseTooManyValuesError(10);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 225, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
    __pyx_t_21 = PyList_GET_ITEM(sequence, 1); 
    __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
    __pyx_t_5 = PyList_GET_ITEM(sequence, 3); 
    __pyx_t_2 = PyList_GET_ITEM(sequence, 4); 
    __pyx_t_23 = PyList_GET_ITEM(sequence, 5); 
    __pyx_t_20 = PyList_GET_ITEM(sequence, 6); 
    __pyx_t_19 = PyList_GET_ITEM(sequence, 7); 
    __pyx_t_3 = PyList_GET_ITEM(sequence, 8); 
    __pyx_t_7 = PyList_GET_ITEM(sequence, 9); 
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_21);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_23);
    __Pyx_INCREF(__pyx_t_20);
    __Pyx_INCREF(__pyx_t_19);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_7);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[10] = {&__pyx_t_6,&__pyx_t_21,&__pyx_t_4,&__pyx_t_5,&__pyx_t_2,&__pyx_t_23,&__pyx_t_20,&__pyx_t_19,&__pyx_t_3,&__pyx_t_7};
      for (i=0; i < 10; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 225, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
  }

  /* "metaphor/monal/lcode/codeconst.py":225
 *     _GetStudent95, _GetUseCount) = list(range(6))
 * 
 * (ctUnknown, ctC, ctPascal, ctFortran77, ctFortran90, ctVBasic,             # <<<<<<<<<<<<<<
 *     ctXML, ctExcel, ctPython, ctCython) = list(range(10))
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ctUnknown, __pyx_t_6) < 0) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ctC, __pyx_t_21) < 0) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_21); __pyx_t_21 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ctPascal, __pyx_t_4) < 0) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ctFortran77, __pyx_t_5) < 0) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ctFortran90, __pyx_t_2) < 0) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ctVBasic, __pyx_t_23) < 0) __PYX_ERR(0, 225, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ctXML, __pyx_t_20) < 0) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ctExcel, __pyx_t_19) < 0) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ctPython, __pyx_t_3) < 0) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ctCython, __pyx_t_7) < 0) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "metaphor/monal/lcode/codeconst.py":228
 *     ctXML, ctExcel, ctPython, ctCython) = list(range(10))
 * 
 * cs_Common = 1             # <<<<<<<<<<<<<<
 * cs_Model = 2
 * cs_All = cs_Common or cs_Model
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cs_Common, __pyx_int_1) < 0) __PYX_ERR(0, 228, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":229
 * 
 * cs_Common = 1
 * cs_Model = 2             # <<<<<<<<<<<<<<
 * cs_All = cs_Common or cs_Model
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cs_Model, __pyx_int_2) < 0) __PYX_ERR(0, 229, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":230
 * cs_Common = 1
 * cs_Model = 2
 * cs_All = cs_Common or cs_Model             # <<<<<<<<<<<<<<
 * 
 * #Chaines indpendantes du langage
 */
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_cs_Common); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_24 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_24 < 0)) __PYX_ERR(0, 230, __pyx_L1_error)
  if (!__pyx_t_24) {
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_7);
    __pyx_t_22 = __pyx_t_7;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L2_bool_binop_done;
  }
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_cs_Model); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_INCREF(__pyx_t_7);
  __pyx_t_22 = __pyx_t_7;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_L2_bool_binop_done:;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cs_All, __pyx_t_22) < 0) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "metaphor/monal/lcode/codeconst.py":233
 * 
 * #Chaines indpendantes du langage
 * DefaultDecimalSeparator = "."             # <<<<<<<<<<<<<<
 * nl = "\n"
 * nlf = ";\n"
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DefaultDecimalSeparator, __pyx_kp_u__10) < 0) __PYX_ERR(0, 233, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":234
 * #Chaines indpendantes du langage
 * DefaultDecimalSeparator = "."
 * nl = "\n"             # <<<<<<<<<<<<<<
 * nlf = ";\n"
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nl, __pyx_kp_u__11) < 0) __PYX_ERR(0, 234, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":235
 * DefaultDecimalSeparator = "."
 * nl = "\n"
 * nlf = ";\n"             # <<<<<<<<<<<<<<
 * 
 * st_InDev = 'Under construction'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_nlf, __pyx_kp_u__12) < 0) __PYX_ERR(0, 235, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":237
 * nlf = ";\n"
 * 
 * st_InDev = 'Under construction'             # <<<<<<<<<<<<<<
 * 
 * StAdvert = ['This code has been automatically created the by the package monal.lcode',
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_st_InDev, __pyx_kp_u_Under_construction) < 0) __PYX_ERR(0, 237, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":239
 * st_InDev = 'Under construction'
 * 
 * StAdvert = ['This code has been automatically created the by the package monal.lcode',             # <<<<<<<<<<<<<<
 *     'This code can be freely used without any royalties with the',
 *     'following conditions :',
 */
  __pyx_t_22 = PyList_New(8); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_INCREF(__pyx_kp_u_This_code_has_been_automatically);
  __Pyx_GIVEREF(__pyx_kp_u_This_code_has_been_automatically);
  PyList_SET_ITEM(__pyx_t_22, 0, __pyx_kp_u_This_code_has_been_automatically);
  __Pyx_INCREF(__pyx_kp_u_This_code_can_be_freely_used_wit);
  __Pyx_GIVEREF(__pyx_kp_u_This_code_can_be_freely_used_wit);
  PyList_SET_ITEM(__pyx_t_22, 1, __pyx_kp_u_This_code_can_be_freely_used_wit);
  __Pyx_INCREF(__pyx_kp_u_following_conditions);
  __Pyx_GIVEREF(__pyx_kp_u_following_conditions);
  PyList_SET_ITEM(__pyx_t_22, 2, __pyx_kp_u_following_conditions);
  __Pyx_INCREF(__pyx_kp_u_Netral_is_not_responsible_of_any);
  __Pyx_GIVEREF(__pyx_kp_u_Netral_is_not_responsible_of_any);
  PyList_SET_ITEM(__pyx_t_22, 3, __pyx_kp_u_Netral_is_not_responsible_of_any);
  __Pyx_INCREF(__pyx_kp_u_code_The_transmission_of_this_co);
  __Pyx_GIVEREF(__pyx_kp_u_code_The_transmission_of_this_co);
  PyList_SET_ITEM(__pyx_t_22, 4, __pyx_kp_u_code_The_transmission_of_this_co);
  __Pyx_INCREF(__pyx_kp_u_organisation_is_free_provided_th);
  __Pyx_GIVEREF(__pyx_kp_u_organisation_is_free_provided_th);
  PyList_SET_ITEM(__pyx_t_22, 5, __pyx_kp_u_organisation_is_free_provided_th);
  __Pyx_INCREF(__pyx_kp_u_is_clearly_mentionned);
  __Pyx_GIVEREF(__pyx_kp_u_is_clearly_mentionned);
  PyList_SET_ITEM(__pyx_t_22, 6, __pyx_kp_u_is_clearly_mentionned);
  __Pyx_INCREF(__pyx_kp_u__13);
  __Pyx_GIVEREF(__pyx_kp_u__13);
  PyList_SET_ITEM(__pyx_t_22, 7, __pyx_kp_u__13);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StAdvert, __pyx_t_22) < 0) __PYX_ERR(0, 239, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "metaphor/monal/lcode/codeconst.py":247
 *     'is clearly mentionned.', '']
 * 
 * StAdvertAdv = ['Ce code a t cre automatiquement le paquet monal.lcode',             # <<<<<<<<<<<<<<
 *     'Ce code est destin uniquement au test, et ne peut en aucun cas ',
 *     'tre utilis de faon permanente sur une installation industrielle.',
 */
  __pyx_t_22 = PyList_New(5); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __Pyx_INCREF(__pyx_kp_u_Ce_code_a_t_cre_automatiquement);
  __Pyx_GIVEREF(__pyx_kp_u_Ce_code_a_t_cre_automatiquement);
  PyList_SET_ITEM(__pyx_t_22, 0, __pyx_kp_u_Ce_code_a_t_cre_automatiquement);
  __Pyx_INCREF(__pyx_kp_u_Ce_code_est_destin_uniquement_au);
  __Pyx_GIVEREF(__pyx_kp_u_Ce_code_est_destin_uniquement_au);
  PyList_SET_ITEM(__pyx_t_22, 1, __pyx_kp_u_Ce_code_est_destin_uniquement_au);
  __Pyx_INCREF(__pyx_kp_u_tre_utilis_de_faon_permanente_su);
  __Pyx_GIVEREF(__pyx_kp_u_tre_utilis_de_faon_permanente_su);
  PyList_SET_ITEM(__pyx_t_22, 2, __pyx_kp_u_tre_utilis_de_faon_permanente_su);
  __Pyx_INCREF(__pyx_kp_u_NETRAL_ne_peut_pas_tre_tenu_pour);
  __Pyx_GIVEREF(__pyx_kp_u_NETRAL_ne_peut_pas_tre_tenu_pour);
  PyList_SET_ITEM(__pyx_t_22, 3, __pyx_kp_u_NETRAL_ne_peut_pas_tre_tenu_pour);
  __Pyx_INCREF(__pyx_kp_u_lutilisation_de_ce_code);
  __Pyx_GIVEREF(__pyx_kp_u_lutilisation_de_ce_code);
  PyList_SET_ITEM(__pyx_t_22, 4, __pyx_kp_u_lutilisation_de_ce_code);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StAdvertAdv, __pyx_t_22) < 0) __PYX_ERR(0, 247, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;

  /* "metaphor/monal/lcode/codeconst.py":253
 *     'l''utilisation de ce code.']
 * 
 * stAdvertDate = 'Date     : %s'             # <<<<<<<<<<<<<<
 * stAdvertTime = 'Time     : %s'
 * stAdvertUser = 'User     : %s'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stAdvertDate, __pyx_kp_u_Date_s) < 0) __PYX_ERR(0, 253, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":254
 * 
 * stAdvertDate = 'Date     : %s'
 * stAdvertTime = 'Time     : %s'             # <<<<<<<<<<<<<<
 * stAdvertUser = 'User     : %s'
 * stAdvertMachine = 'Computer : %s'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stAdvertTime, __pyx_kp_u_Time_s) < 0) __PYX_ERR(0, 254, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":255
 * stAdvertDate = 'Date     : %s'
 * stAdvertTime = 'Time     : %s'
 * stAdvertUser = 'User     : %s'             # <<<<<<<<<<<<<<
 * stAdvertMachine = 'Computer : %s'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stAdvertUser, __pyx_kp_u_User_s) < 0) __PYX_ERR(0, 255, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":256
 * stAdvertTime = 'Time     : %s'
 * stAdvertUser = 'User     : %s'
 * stAdvertMachine = 'Computer : %s'             # <<<<<<<<<<<<<<
 * 
 * #Chaine constantes pour C
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stAdvertMachine, __pyx_kp_u_Computer_s) < 0) __PYX_ERR(0, 256, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":259
 * 
 * #Chaine constantes pour C
 * Eq = ' = '             # <<<<<<<<<<<<<<
 * MEq = ' *= '
 * PEq = ' += '
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Eq, __pyx_kp_u__14) < 0) __PYX_ERR(0, 259, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":260
 * #Chaine constantes pour C
 * Eq = ' = '
 * MEq = ' *= '             # <<<<<<<<<<<<<<
 * PEq = ' += '
 * DEq = ' /= '
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MEq, __pyx_kp_u__15) < 0) __PYX_ERR(0, 260, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":261
 * Eq = ' = '
 * MEq = ' *= '
 * PEq = ' += '             # <<<<<<<<<<<<<<
 * DEq = ' /= '
 * StDefineInputs = '#define %sINPUTS %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PEq, __pyx_kp_u__16) < 0) __PYX_ERR(0, 261, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":262
 * MEq = ' *= '
 * PEq = ' += '
 * DEq = ' /= '             # <<<<<<<<<<<<<<
 * StDefineInputs = '#define %sINPUTS %d'
 * StDefineOutputs = '#define %sOUTPUTS %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DEq, __pyx_kp_u__17) < 0) __PYX_ERR(0, 262, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":263
 * PEq = ' += '
 * DEq = ' /= '
 * StDefineInputs = '#define %sINPUTS %d'             # <<<<<<<<<<<<<<
 * StDefineOutputs = '#define %sOUTPUTS %d'
 * StDefineOrder = '#define %sORDER %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StDefineInputs, __pyx_kp_u_define_sINPUTS_d) < 0) __PYX_ERR(0, 263, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":264
 * DEq = ' /= '
 * StDefineInputs = '#define %sINPUTS %d'
 * StDefineOutputs = '#define %sOUTPUTS %d'             # <<<<<<<<<<<<<<
 * StDefineOrder = '#define %sORDER %d'
 * stDefineNodes = '#define %sNODES %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StDefineOutputs, __pyx_kp_u_define_sOUTPUTS_d) < 0) __PYX_ERR(0, 264, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":265
 * StDefineInputs = '#define %sINPUTS %d'
 * StDefineOutputs = '#define %sOUTPUTS %d'
 * StDefineOrder = '#define %sORDER %d'             # <<<<<<<<<<<<<<
 * stDefineNodes = '#define %sNODES %d'
 * stDefineSynapses = '#define %sSYNAPSES %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StDefineOrder, __pyx_kp_u_define_sORDER_d) < 0) __PYX_ERR(0, 265, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":266
 * StDefineOutputs = '#define %sOUTPUTS %d'
 * StDefineOrder = '#define %sORDER %d'
 * stDefineNodes = '#define %sNODES %d'             # <<<<<<<<<<<<<<
 * stDefineSynapses = '#define %sSYNAPSES %d'
 * stDefineMatrixes = '#define %sMATRIXES %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stDefineNodes, __pyx_kp_u_define_sNODES_d) < 0) __PYX_ERR(0, 266, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":267
 * StDefineOrder = '#define %sORDER %d'
 * stDefineNodes = '#define %sNODES %d'
 * stDefineSynapses = '#define %sSYNAPSES %d'             # <<<<<<<<<<<<<<
 * stDefineMatrixes = '#define %sMATRIXES %d'
 * StdefineStud95 = '#define %sSTUDENT95 %g'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stDefineSynapses, __pyx_kp_u_define_sSYNAPSES_d) < 0) __PYX_ERR(0, 267, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":268
 * stDefineNodes = '#define %sNODES %d'
 * stDefineSynapses = '#define %sSYNAPSES %d'
 * stDefineMatrixes = '#define %sMATRIXES %d'             # <<<<<<<<<<<<<<
 * StdefineStud95 = '#define %sSTUDENT95 %g'
 * StdefineStdDev = '#define %sSTDTRDEV %g'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stDefineMatrixes, __pyx_kp_u_define_sMATRIXES_d) < 0) __PYX_ERR(0, 268, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":269
 * stDefineSynapses = '#define %sSYNAPSES %d'
 * stDefineMatrixes = '#define %sMATRIXES %d'
 * StdefineStud95 = '#define %sSTUDENT95 %g'             # <<<<<<<<<<<<<<
 * StdefineStdDev = '#define %sSTDTRDEV %g'
 * StDefineConstants = '#define %sCONSTANTS %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StdefineStud95, __pyx_kp_u_define_sSTUDENT95_g) < 0) __PYX_ERR(0, 269, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":270
 * stDefineMatrixes = '#define %sMATRIXES %d'
 * StdefineStud95 = '#define %sSTUDENT95 %g'
 * StdefineStdDev = '#define %sSTDTRDEV %g'             # <<<<<<<<<<<<<<
 * StDefineConstants = '#define %sCONSTANTS %d'
 * stDefineRetros = '#define %sMEMNEURO %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StdefineStdDev, __pyx_kp_u_define_sSTDTRDEV_g) < 0) __PYX_ERR(0, 270, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":271
 * StdefineStud95 = '#define %sSTUDENT95 %g'
 * StdefineStdDev = '#define %sSTDTRDEV %g'
 * StDefineConstants = '#define %sCONSTANTS %d'             # <<<<<<<<<<<<<<
 * stDefineRetros = '#define %sMEMNEURO %d'
 * stDefineLinks = '#define %sMEMSYN %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StDefineConstants, __pyx_kp_u_define_sCONSTANTS_d) < 0) __PYX_ERR(0, 271, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":272
 * StdefineStdDev = '#define %sSTDTRDEV %g'
 * StDefineConstants = '#define %sCONSTANTS %d'
 * stDefineRetros = '#define %sMEMNEURO %d'             # <<<<<<<<<<<<<<
 * stDefineLinks = '#define %sMEMSYN %d'
 * stDefineDimn = '#define %sDIM%d %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stDefineRetros, __pyx_kp_u_define_sMEMNEURO_d) < 0) __PYX_ERR(0, 272, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":273
 * StDefineConstants = '#define %sCONSTANTS %d'
 * stDefineRetros = '#define %sMEMNEURO %d'
 * stDefineLinks = '#define %sMEMSYN %d'             # <<<<<<<<<<<<<<
 * stDefineDimn = '#define %sDIM%d %d'
 * stDefineDimension = '#define %sDIMENSION %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stDefineLinks, __pyx_kp_u_define_sMEMSYN_d) < 0) __PYX_ERR(0, 273, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":274
 * stDefineRetros = '#define %sMEMNEURO %d'
 * stDefineLinks = '#define %sMEMSYN %d'
 * stDefineDimn = '#define %sDIM%d %d'             # <<<<<<<<<<<<<<
 * stDefineDimension = '#define %sDIMENSION %d'
 * StIfNDefTransfer = '#ifndef %sTRANSFER_H'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stDefineDimn, __pyx_kp_u_define_sDIM_d_d) < 0) __PYX_ERR(0, 274, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":275
 * stDefineLinks = '#define %sMEMSYN %d'
 * stDefineDimn = '#define %sDIM%d %d'
 * stDefineDimension = '#define %sDIMENSION %d'             # <<<<<<<<<<<<<<
 * StIfNDefTransfer = '#ifndef %sTRANSFER_H'
 * StDefineTransfer = '#define %sTRANSFER_H'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stDefineDimension, __pyx_kp_u_define_sDIMENSION_d) < 0) __PYX_ERR(0, 275, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":276
 * stDefineDimn = '#define %sDIM%d %d'
 * stDefineDimension = '#define %sDIMENSION %d'
 * StIfNDefTransfer = '#ifndef %sTRANSFER_H'             # <<<<<<<<<<<<<<
 * StDefineTransfer = '#define %sTRANSFER_H'
 * StIfNDefTransferGr = '#ifndef %sTRANSFERGR_H'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StIfNDefTransfer, __pyx_kp_u_ifndef_sTRANSFER_H) < 0) __PYX_ERR(0, 276, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":277
 * stDefineDimension = '#define %sDIMENSION %d'
 * StIfNDefTransfer = '#ifndef %sTRANSFER_H'
 * StDefineTransfer = '#define %sTRANSFER_H'             # <<<<<<<<<<<<<<
 * StIfNDefTransferGr = '#ifndef %sTRANSFERGR_H'
 * StDefineTransferGr = '#define %sTRANSFERGR_H'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StDefineTransfer, __pyx_kp_u_define_sTRANSFER_H) < 0) __PYX_ERR(0, 277, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":278
 * StIfNDefTransfer = '#ifndef %sTRANSFER_H'
 * StDefineTransfer = '#define %sTRANSFER_H'
 * StIfNDefTransferGr = '#ifndef %sTRANSFERGR_H'             # <<<<<<<<<<<<<<
 * StDefineTransferGr = '#define %sTRANSFERGR_H'
 * StIfNDefTransferLev = '#ifndef %sTRANSFERLEV_H'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StIfNDefTransferGr, __pyx_kp_u_ifndef_sTRANSFERGR_H) < 0) __PYX_ERR(0, 278, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":279
 * StDefineTransfer = '#define %sTRANSFER_H'
 * StIfNDefTransferGr = '#ifndef %sTRANSFERGR_H'
 * StDefineTransferGr = '#define %sTRANSFERGR_H'             # <<<<<<<<<<<<<<
 * StIfNDefTransferLev = '#ifndef %sTRANSFERLEV_H'
 * StDefineTransferLev = '#define %sTRANSFERLEV_H'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StDefineTransferGr, __pyx_kp_u_define_sTRANSFERGR_H) < 0) __PYX_ERR(0, 279, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":280
 * StIfNDefTransferGr = '#ifndef %sTRANSFERGR_H'
 * StDefineTransferGr = '#define %sTRANSFERGR_H'
 * StIfNDefTransferLev = '#ifndef %sTRANSFERLEV_H'             # <<<<<<<<<<<<<<
 * StDefineTransferLev = '#define %sTRANSFERLEV_H'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StIfNDefTransferLev, __pyx_kp_u_ifndef_sTRANSFERLEV_H) < 0) __PYX_ERR(0, 280, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":281
 * StDefineTransferGr = '#define %sTRANSFERGR_H'
 * StIfNDefTransferLev = '#ifndef %sTRANSFERLEV_H'
 * StDefineTransferLev = '#define %sTRANSFERLEV_H'             # <<<<<<<<<<<<<<
 * 
 * StMatrixCompute = '%smatrix[mindex(i,j)] * gradient[j]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StDefineTransferLev, __pyx_kp_u_define_sTRANSFERLEV_H) < 0) __PYX_ERR(0, 281, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":283
 * StDefineTransferLev = '#define %sTRANSFERLEV_H'
 * 
 * StMatrixCompute = '%smatrix[mindex(i,j)] * gradient[j]'             # <<<<<<<<<<<<<<
 * StForAllOutputs = 'for (i=1; i < %sOUTPUTS; i++)'
 * StForAllSynapses = 'for (i=0; i < %sSYNAPSES; i++)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StMatrixCompute, __pyx_kp_u_smatrix_mindex_i_j_gradient_j) < 0) __PYX_ERR(0, 283, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":284
 * 
 * StMatrixCompute = '%smatrix[mindex(i,j)] * gradient[j]'
 * StForAllOutputs = 'for (i=1; i < %sOUTPUTS; i++)'             # <<<<<<<<<<<<<<
 * StForAllSynapses = 'for (i=0; i < %sSYNAPSES; i++)'
 * StForAllSynapsesj = 'for (j=0; j < %sSYNAPSES; j++)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllOutputs, __pyx_kp_u_for_i_1_i_sOUTPUTS_i) < 0) __PYX_ERR(0, 284, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":285
 * StMatrixCompute = '%smatrix[mindex(i,j)] * gradient[j]'
 * StForAllOutputs = 'for (i=1; i < %sOUTPUTS; i++)'
 * StForAllSynapses = 'for (i=0; i < %sSYNAPSES; i++)'             # <<<<<<<<<<<<<<
 * StForAllSynapsesj = 'for (j=0; j < %sSYNAPSES; j++)'
 * StForAllInitGradient = 'for (i=0; i < %sSYNAPSES; gradient[i++] = 0)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllSynapses, __pyx_kp_u_for_i_0_i_sSYNAPSES_i) < 0) __PYX_ERR(0, 285, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":286
 * StForAllOutputs = 'for (i=1; i < %sOUTPUTS; i++)'
 * StForAllSynapses = 'for (i=0; i < %sSYNAPSES; i++)'
 * StForAllSynapsesj = 'for (j=0; j < %sSYNAPSES; j++)'             # <<<<<<<<<<<<<<
 * StForAllInitGradient = 'for (i=0; i < %sSYNAPSES; gradient[i++] = 0)'
 * StForAllInitIniBack = 'for (i=0; i < %sOUTPUTS; iniback[i++] = 0)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllSynapsesj, __pyx_kp_u_for_j_0_j_sSYNAPSES_j) < 0) __PYX_ERR(0, 286, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":287
 * StForAllSynapses = 'for (i=0; i < %sSYNAPSES; i++)'
 * StForAllSynapsesj = 'for (j=0; j < %sSYNAPSES; j++)'
 * StForAllInitGradient = 'for (i=0; i < %sSYNAPSES; gradient[i++] = 0)'             # <<<<<<<<<<<<<<
 * StForAllInitIniBack = 'for (i=0; i < %sOUTPUTS; iniback[i++] = 0)'
 * StForAllInitLeverage = 'for (i=0; i < %sOUTPUTS; leverage[i++] = 0)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllInitGradient, __pyx_kp_u_for_i_0_i_sSYNAPSES_gradient_i_0) < 0) __PYX_ERR(0, 287, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":288
 * StForAllSynapsesj = 'for (j=0; j < %sSYNAPSES; j++)'
 * StForAllInitGradient = 'for (i=0; i < %sSYNAPSES; gradient[i++] = 0)'
 * StForAllInitIniBack = 'for (i=0; i < %sOUTPUTS; iniback[i++] = 0)'             # <<<<<<<<<<<<<<
 * StForAllInitLeverage = 'for (i=0; i < %sOUTPUTS; leverage[i++] = 0)'
 * StForAllInitMemSyn = 'for (i=0; i < %sMEMSYN; memsyn[i++] = 1)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllInitIniBack, __pyx_kp_u_for_i_0_i_sOUTPUTS_iniback_i_0) < 0) __PYX_ERR(0, 288, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":289
 * StForAllInitGradient = 'for (i=0; i < %sSYNAPSES; gradient[i++] = 0)'
 * StForAllInitIniBack = 'for (i=0; i < %sOUTPUTS; iniback[i++] = 0)'
 * StForAllInitLeverage = 'for (i=0; i < %sOUTPUTS; leverage[i++] = 0)'             # <<<<<<<<<<<<<<
 * StForAllInitMemSyn = 'for (i=0; i < %sMEMSYN; memsyn[i++] = 1)'
 * StMatIniS = 'matrix[%sMATRIXES]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllInitLeverage, __pyx_kp_u_for_i_0_i_sOUTPUTS_leverage_i_0) < 0) __PYX_ERR(0, 289, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":290
 * StForAllInitIniBack = 'for (i=0; i < %sOUTPUTS; iniback[i++] = 0)'
 * StForAllInitLeverage = 'for (i=0; i < %sOUTPUTS; leverage[i++] = 0)'
 * StForAllInitMemSyn = 'for (i=0; i < %sMEMSYN; memsyn[i++] = 1)'             # <<<<<<<<<<<<<<
 * StMatIniS = 'matrix[%sMATRIXES]'
 * StWeightIni = 'weights[%d]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllInitMemSyn, __pyx_kp_u_for_i_0_i_sMEMSYN_memsyn_i_1) < 0) __PYX_ERR(0, 290, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":291
 * StForAllInitLeverage = 'for (i=0; i < %sOUTPUTS; leverage[i++] = 0)'
 * StForAllInitMemSyn = 'for (i=0; i < %sMEMSYN; memsyn[i++] = 1)'
 * StMatIniS = 'matrix[%sMATRIXES]'             # <<<<<<<<<<<<<<
 * StWeightIni = 'weights[%d]'
 * StWeightRedirIni = '(*weights[weightsredir[%d]])'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StMatIniS, __pyx_kp_u_matrix_sMATRIXES) < 0) __PYX_ERR(0, 291, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":292
 * StForAllInitMemSyn = 'for (i=0; i < %sMEMSYN; memsyn[i++] = 1)'
 * StMatIniS = 'matrix[%sMATRIXES]'
 * StWeightIni = 'weights[%d]'             # <<<<<<<<<<<<<<
 * StWeightRedirIni = '(*weights[weightsredir[%d]])'
 * StMatrixIni = 'matrix[%d]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StWeightIni, __pyx_kp_u_weights_d) < 0) __PYX_ERR(0, 292, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":293
 * StMatIniS = 'matrix[%sMATRIXES]'
 * StWeightIni = 'weights[%d]'
 * StWeightRedirIni = '(*weights[weightsredir[%d]])'             # <<<<<<<<<<<<<<
 * StMatrixIni = 'matrix[%d]'
 * StWeightIniS = 'weights[%sSYNAPSES]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StWeightRedirIni, __pyx_kp_u_weights_weightsredir_d) < 0) __PYX_ERR(0, 293, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":294
 * StWeightIni = 'weights[%d]'
 * StWeightRedirIni = '(*weights[weightsredir[%d]])'
 * StMatrixIni = 'matrix[%d]'             # <<<<<<<<<<<<<<
 * StWeightIniS = 'weights[%sSYNAPSES]'
 * StMinIniS = 'min[%sINPUTS + %sOUTPUTS]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StMatrixIni, __pyx_kp_u_matrix_d) < 0) __PYX_ERR(0, 294, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":295
 * StWeightRedirIni = '(*weights[weightsredir[%d]])'
 * StMatrixIni = 'matrix[%d]'
 * StWeightIniS = 'weights[%sSYNAPSES]'             # <<<<<<<<<<<<<<
 * StMinIniS = 'min[%sINPUTS + %sOUTPUTS]'
 * StMaxIniS = 'max[%sINPUTS + %sOUTPUTS]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StWeightIniS, __pyx_kp_u_weights_sSYNAPSES) < 0) __PYX_ERR(0, 295, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":296
 * StMatrixIni = 'matrix[%d]'
 * StWeightIniS = 'weights[%sSYNAPSES]'
 * StMinIniS = 'min[%sINPUTS + %sOUTPUTS]'             # <<<<<<<<<<<<<<
 * StMaxIniS = 'max[%sINPUTS + %sOUTPUTS]'
 * StStdDEvInis = 'stddev[%sOUTPUTS]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StMinIniS, __pyx_kp_u_min_sINPUTS_sOUTPUTS) < 0) __PYX_ERR(0, 296, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":297
 * StWeightIniS = 'weights[%sSYNAPSES]'
 * StMinIniS = 'min[%sINPUTS + %sOUTPUTS]'
 * StMaxIniS = 'max[%sINPUTS + %sOUTPUTS]'             # <<<<<<<<<<<<<<
 * StStdDEvInis = 'stddev[%sOUTPUTS]'
 * stNodedS = 'double node[%sNODES]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StMaxIniS, __pyx_kp_u_max_sINPUTS_sOUTPUTS) < 0) __PYX_ERR(0, 297, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":298
 * StMinIniS = 'min[%sINPUTS + %sOUTPUTS]'
 * StMaxIniS = 'max[%sINPUTS + %sOUTPUTS]'
 * StStdDEvInis = 'stddev[%sOUTPUTS]'             # <<<<<<<<<<<<<<
 * stNodedS = 'double node[%sNODES]'
 * stNodedSK = 'double node[%sOUTPUTS]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StStdDEvInis, __pyx_kp_u_stddev_sOUTPUTS) < 0) __PYX_ERR(0, 298, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":299
 * StMaxIniS = 'max[%sINPUTS + %sOUTPUTS]'
 * StStdDEvInis = 'stddev[%sOUTPUTS]'
 * stNodedS = 'double node[%sNODES]'             # <<<<<<<<<<<<<<
 * stNodedSK = 'double node[%sOUTPUTS]'
 * stNodedS2 = 'double node[%sNODES][2]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stNodedS, __pyx_kp_u_double_node_sNODES) < 0) __PYX_ERR(0, 299, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":300
 * StStdDEvInis = 'stddev[%sOUTPUTS]'
 * stNodedS = 'double node[%sNODES]'
 * stNodedSK = 'double node[%sOUTPUTS]'             # <<<<<<<<<<<<<<
 * stNodedS2 = 'double node[%sNODES][2]'
 * stNodedS3 = 'double node[%sNODES][3]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stNodedSK, __pyx_kp_u_double_node_sOUTPUTS) < 0) __PYX_ERR(0, 300, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":301
 * stNodedS = 'double node[%sNODES]'
 * stNodedSK = 'double node[%sOUTPUTS]'
 * stNodedS2 = 'double node[%sNODES][2]'             # <<<<<<<<<<<<<<
 * stNodedS3 = 'double node[%sNODES][3]'
 * stMemSyndS = 'double memsyn[%sMEMSYN]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stNodedS2, __pyx_kp_u_double_node_sNODES_2) < 0) __PYX_ERR(0, 301, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":302
 * stNodedSK = 'double node[%sOUTPUTS]'
 * stNodedS2 = 'double node[%sNODES][2]'
 * stNodedS3 = 'double node[%sNODES][3]'             # <<<<<<<<<<<<<<
 * stMemSyndS = 'double memsyn[%sMEMSYN]'
 * stGradientdS = 'double gradient[%sSYNAPSES]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stNodedS3, __pyx_kp_u_double_node_sNODES_3) < 0) __PYX_ERR(0, 302, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":303
 * stNodedS2 = 'double node[%sNODES][2]'
 * stNodedS3 = 'double node[%sNODES][3]'
 * stMemSyndS = 'double memsyn[%sMEMSYN]'             # <<<<<<<<<<<<<<
 * stGradientdS = 'double gradient[%sSYNAPSES]'
 * StIniBackdS = 'double iniback[%sOUTPUTS]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stMemSyndS, __pyx_kp_u_double_memsyn_sMEMSYN) < 0) __PYX_ERR(0, 303, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":304
 * stNodedS3 = 'double node[%sNODES][3]'
 * stMemSyndS = 'double memsyn[%sMEMSYN]'
 * stGradientdS = 'double gradient[%sSYNAPSES]'             # <<<<<<<<<<<<<<
 * StIniBackdS = 'double iniback[%sOUTPUTS]'
 * StInputNameDecl = 'static char *names[%sINPUTS] = {'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stGradientdS, __pyx_kp_u_double_gradient_sSYNAPSES) < 0) __PYX_ERR(0, 304, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":305
 * stMemSyndS = 'double memsyn[%sMEMSYN]'
 * stGradientdS = 'double gradient[%sSYNAPSES]'
 * StIniBackdS = 'double iniback[%sOUTPUTS]'             # <<<<<<<<<<<<<<
 * StInputNameDecl = 'static char *names[%sINPUTS] = {'
 * StOutputNameDecl = 'static char *names[%sOUTPUTS] = {'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StIniBackdS, __pyx_kp_u_double_iniback_sOUTPUTS) < 0) __PYX_ERR(0, 305, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":306
 * stGradientdS = 'double gradient[%sSYNAPSES]'
 * StIniBackdS = 'double iniback[%sOUTPUTS]'
 * StInputNameDecl = 'static char *names[%sINPUTS] = {'             # <<<<<<<<<<<<<<
 * StOutputNameDecl = 'static char *names[%sOUTPUTS] = {'
 * StDefineMIndex = '#define mindex(i,j) (i>j)?(j+i*(i+1)/2):(i+j*(j+1)/2)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StInputNameDecl, __pyx_kp_u_static_char_names_sINPUTS) < 0) __PYX_ERR(0, 306, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":307
 * StIniBackdS = 'double iniback[%sOUTPUTS]'
 * StInputNameDecl = 'static char *names[%sINPUTS] = {'
 * StOutputNameDecl = 'static char *names[%sOUTPUTS] = {'             # <<<<<<<<<<<<<<
 * StDefineMIndex = '#define mindex(i,j) (i>j)?(j+i*(i+1)/2):(i+j*(j+1)/2)'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StOutputNameDecl, __pyx_kp_u_static_char_names_sOUTPUTS) < 0) __PYX_ERR(0, 307, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":308
 * StInputNameDecl = 'static char *names[%sINPUTS] = {'
 * StOutputNameDecl = 'static char *names[%sOUTPUTS] = {'
 * StDefineMIndex = '#define mindex(i,j) (i>j)?(j+i*(i+1)/2):(i+j*(j+1)/2)'             # <<<<<<<<<<<<<<
 * 
 * #StTransferKohProto = 'int %stransfer(double*, double*, int*)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StDefineMIndex, __pyx_kp_u_define_mindex_i_j_i_j_j_i_i_1_2) < 0) __PYX_ERR(0, 308, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":312
 * #StTransferKohProto = 'int %stransfer(double*, double*, int*)'
 * #StTransferKoh = 'int %stransferK(double *weights, double *inputs, int *coord)'
 * StTransferStatProto = 'void %stransfer(double*, double*)'             # <<<<<<<<<<<<<<
 * StTransferStatProto2 = 'void %stransferlev(double*, double*, double*, double*, double*)'
 * StTransferStat = 'void %stransfer(double *inputs, double *outputs)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferStatProto, __pyx_kp_u_void_stransfer_double_double) < 0) __PYX_ERR(0, 312, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":313
 * #StTransferKoh = 'int %stransferK(double *weights, double *inputs, int *coord)'
 * StTransferStatProto = 'void %stransfer(double*, double*)'
 * StTransferStatProto2 = 'void %stransferlev(double*, double*, double*, double*, double*)'             # <<<<<<<<<<<<<<
 * StTransferStat = 'void %stransfer(double *inputs, double *outputs)'
 * StTransferStat2 = 'void %stransferlev(double *weights, double *zmat, double *inputs, double *outputs, double *leverage)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferStatProto2, __pyx_kp_u_void_stransferlev_double_double) < 0) __PYX_ERR(0, 313, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":314
 * StTransferStatProto = 'void %stransfer(double*, double*)'
 * StTransferStatProto2 = 'void %stransferlev(double*, double*, double*, double*, double*)'
 * StTransferStat = 'void %stransfer(double *inputs, double *outputs)'             # <<<<<<<<<<<<<<
 * StTransferStat2 = 'void %stransferlev(double *weights, double *zmat, double *inputs, double *outputs, double *leverage)'
 * StTransGradStat = 'void %stransfergradient(double *inputs, double *outputs, double *gradient, double *iniback)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferStat, __pyx_kp_u_void_stransfer_double_inputs_dou) < 0) __PYX_ERR(0, 314, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":315
 * StTransferStatProto2 = 'void %stransferlev(double*, double*, double*, double*, double*)'
 * StTransferStat = 'void %stransfer(double *inputs, double *outputs)'
 * StTransferStat2 = 'void %stransferlev(double *weights, double *zmat, double *inputs, double *outputs, double *leverage)'             # <<<<<<<<<<<<<<
 * StTransGradStat = 'void %stransfergradient(double *inputs, double *outputs, double *gradient, double *iniback)'
 * StTransGradStatProto = 'void %stransfergradient(double*, double*, double*, double*)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferStat2, __pyx_kp_u_void_stransferlev_double_weights) < 0) __PYX_ERR(0, 315, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":316
 * StTransferStat = 'void %stransfer(double *inputs, double *outputs)'
 * StTransferStat2 = 'void %stransferlev(double *weights, double *zmat, double *inputs, double *outputs, double *leverage)'
 * StTransGradStat = 'void %stransfergradient(double *inputs, double *outputs, double *gradient, double *iniback)'             # <<<<<<<<<<<<<<
 * StTransGradStatProto = 'void %stransfergradient(double*, double*, double*, double*)'
 * StTransGradStatCall = '%stransfergradient(inputs, outputs, gradient, iniback)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransGradStat, __pyx_kp_u_void_stransfergradient_double_in) < 0) __PYX_ERR(0, 316, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":317
 * StTransferStat2 = 'void %stransferlev(double *weights, double *zmat, double *inputs, double *outputs, double *leverage)'
 * StTransGradStat = 'void %stransfergradient(double *inputs, double *outputs, double *gradient, double *iniback)'
 * StTransGradStatProto = 'void %stransfergradient(double*, double*, double*, double*)'             # <<<<<<<<<<<<<<
 * StTransGradStatCall = '%stransfergradient(inputs, outputs, gradient, iniback)'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransGradStatProto, __pyx_kp_u_void_stransfergradient_double_do) < 0) __PYX_ERR(0, 317, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":318
 * StTransGradStat = 'void %stransfergradient(double *inputs, double *outputs, double *gradient, double *iniback)'
 * StTransGradStatProto = 'void %stransfergradient(double*, double*, double*, double*)'
 * StTransGradStatCall = '%stransfergradient(inputs, outputs, gradient, iniback)'             # <<<<<<<<<<<<<<
 * 
 * StTransferDynProto = 'void %stransfer(double*, double*, double*)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransGradStatCall, __pyx_kp_u_stransfergradient_inputs_output) < 0) __PYX_ERR(0, 318, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":320
 * StTransGradStatCall = '%stransfergradient(inputs, outputs, gradient, iniback)'
 * 
 * StTransferDynProto = 'void %stransfer(double*, double*, double*)'             # <<<<<<<<<<<<<<
 * StTransferDynProto2 = 'void %stransferlev(double*, double*, double*, double*)'
 * StTransferDyn = 'void %stransfer(double *inputs, double *outputs, double *states)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferDynProto, __pyx_kp_u_void_stransfer_double_double_dou) < 0) __PYX_ERR(0, 320, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":321
 * 
 * StTransferDynProto = 'void %stransfer(double*, double*, double*)'
 * StTransferDynProto2 = 'void %stransferlev(double*, double*, double*, double*)'             # <<<<<<<<<<<<<<
 * StTransferDyn = 'void %stransfer(double *inputs, double *outputs, double *states)'
 * StTransferDyn2 = 'void %stransferlev(double *inputs, double *outputs, double *states, double *leverage)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferDynProto2, __pyx_kp_u_void_stransferlev_double_double_2) < 0) __PYX_ERR(0, 321, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":322
 * StTransferDynProto = 'void %stransfer(double*, double*, double*)'
 * StTransferDynProto2 = 'void %stransferlev(double*, double*, double*, double*)'
 * StTransferDyn = 'void %stransfer(double *inputs, double *outputs, double *states)'             # <<<<<<<<<<<<<<
 * StTransferDyn2 = 'void %stransferlev(double *inputs, double *outputs, double *states, double *leverage)'
 * StTransGradDyn = 'void %stransfergradient(double *inputs, double *outputs, double *states, double *gradient, double *iniback)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferDyn, __pyx_kp_u_void_stransfer_double_inputs_dou_2) < 0) __PYX_ERR(0, 322, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":323
 * StTransferDynProto2 = 'void %stransferlev(double*, double*, double*, double*)'
 * StTransferDyn = 'void %stransfer(double *inputs, double *outputs, double *states)'
 * StTransferDyn2 = 'void %stransferlev(double *inputs, double *outputs, double *states, double *leverage)'             # <<<<<<<<<<<<<<
 * StTransGradDyn = 'void %stransfergradient(double *inputs, double *outputs, double *states, double *gradient, double *iniback)'
 * StTransGradDynProto = 'void %stransfergradient(double*, double*, double*, double*, double*)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferDyn2, __pyx_kp_u_void_stransferlev_double_inputs) < 0) __PYX_ERR(0, 323, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":324
 * StTransferDyn = 'void %stransfer(double *inputs, double *outputs, double *states)'
 * StTransferDyn2 = 'void %stransferlev(double *inputs, double *outputs, double *states, double *leverage)'
 * StTransGradDyn = 'void %stransfergradient(double *inputs, double *outputs, double *states, double *gradient, double *iniback)'             # <<<<<<<<<<<<<<
 * StTransGradDynProto = 'void %stransfergradient(double*, double*, double*, double*, double*)'
 * StTransGradDynCall = '%stransfergradient(inputs, outputs, states, gradient, iniback)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransGradDyn, __pyx_kp_u_void_stransfergradient_double_in_2) < 0) __PYX_ERR(0, 324, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":325
 * StTransferDyn2 = 'void %stransferlev(double *inputs, double *outputs, double *states, double *leverage)'
 * StTransGradDyn = 'void %stransfergradient(double *inputs, double *outputs, double *states, double *gradient, double *iniback)'
 * StTransGradDynProto = 'void %stransfergradient(double*, double*, double*, double*, double*)'             # <<<<<<<<<<<<<<
 * StTransGradDynCall = '%stransfergradient(inputs, outputs, states, gradient, iniback)'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransGradDynProto, __pyx_kp_u_void_stransfergradient_double_do_2) < 0) __PYX_ERR(0, 325, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":326
 * StTransGradDyn = 'void %stransfergradient(double *inputs, double *outputs, double *states, double *gradient, double *iniback)'
 * StTransGradDynProto = 'void %stransfergradient(double*, double*, double*, double*, double*)'
 * StTransGradDynCall = '%stransfergradient(inputs, outputs, states, gradient, iniback)'             # <<<<<<<<<<<<<<
 * 
 * StInputNameProto = 'char *%sinputname(int)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransGradDynCall, __pyx_kp_u_stransfergradient_inputs_output_2) < 0) __PYX_ERR(0, 326, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":328
 * StTransGradDynCall = '%stransfergradient(inputs, outputs, states, gradient, iniback)'
 * 
 * StInputNameProto = 'char *%sinputname(int)'             # <<<<<<<<<<<<<<
 * StOutputNameProto = 'char *%soutputname(int)'
 * StInputName = 'char *%sinputname(int index)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StInputNameProto, __pyx_kp_u_char_sinputname_int) < 0) __PYX_ERR(0, 328, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":329
 * 
 * StInputNameProto = 'char *%sinputname(int)'
 * StOutputNameProto = 'char *%soutputname(int)'             # <<<<<<<<<<<<<<
 * StInputName = 'char *%sinputname(int index)'
 * StOutputName = 'char *%soutputname(int index)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StOutputNameProto, __pyx_kp_u_char_soutputname_int) < 0) __PYX_ERR(0, 329, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":330
 * StInputNameProto = 'char *%sinputname(int)'
 * StOutputNameProto = 'char *%soutputname(int)'
 * StInputName = 'char *%sinputname(int index)'             # <<<<<<<<<<<<<<
 * StOutputName = 'char *%soutputname(int index)'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StInputName, __pyx_kp_u_char_sinputname_int_index) < 0) __PYX_ERR(0, 330, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":331
 * StOutputNameProto = 'char *%soutputname(int)'
 * StInputName = 'char *%sinputname(int index)'
 * StOutputName = 'char *%soutputname(int index)'             # <<<<<<<<<<<<<<
 * 
 * stAlert = 'Please check name of included header file'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StOutputName, __pyx_kp_u_char_soutputname_int_index) < 0) __PYX_ERR(0, 331, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":333
 * StOutputName = 'char *%soutputname(int index)'
 * 
 * stAlert = 'Please check name of included header file'             # <<<<<<<<<<<<<<
 * stIncludeFile = '#include "%s"'
 * StReturnInputName = 'return (index < 0 || index >= %sINPUTS) ? NULL : names[index]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stAlert, __pyx_kp_u_Please_check_name_of_included_he) < 0) __PYX_ERR(0, 333, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":334
 * 
 * stAlert = 'Please check name of included header file'
 * stIncludeFile = '#include "%s"'             # <<<<<<<<<<<<<<
 * StReturnInputName = 'return (index < 0 || index >= %sINPUTS) ? NULL : names[index]'
 * StReturnOutputName = 'return (index < 0 || index >= %sOUTPUTS) ? NULL : names[index]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stIncludeFile, __pyx_kp_u_include_s) < 0) __PYX_ERR(0, 334, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":335
 * stAlert = 'Please check name of included header file'
 * stIncludeFile = '#include "%s"'
 * StReturnInputName = 'return (index < 0 || index >= %sINPUTS) ? NULL : names[index]'             # <<<<<<<<<<<<<<
 * StReturnOutputName = 'return (index < 0 || index >= %sOUTPUTS) ? NULL : names[index]'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StReturnInputName, __pyx_kp_u_return_index_0_index_sINPUTS_NUL) < 0) __PYX_ERR(0, 335, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":336
 * stIncludeFile = '#include "%s"'
 * StReturnInputName = 'return (index < 0 || index >= %sINPUTS) ? NULL : names[index]'
 * StReturnOutputName = 'return (index < 0 || index >= %sOUTPUTS) ? NULL : names[index]'             # <<<<<<<<<<<<<<
 * 
 * #Chaine constantes pour Pascal
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StReturnOutputName, __pyx_kp_u_return_index_0_index_sOUTPUTS_NU) < 0) __PYX_ERR(0, 336, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":339
 * 
 * #Chaine constantes pour Pascal
 * stIncludeFileP = '{$I %s}'             # <<<<<<<<<<<<<<
 * StConstInputs = '%sINPUTS= %d'
 * StConstOutputs = '%sOUTPUTS = %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stIncludeFileP, __pyx_kp_u_I_s) < 0) __PYX_ERR(0, 339, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":340
 * #Chaine constantes pour Pascal
 * stIncludeFileP = '{$I %s}'
 * StConstInputs = '%sINPUTS= %d'             # <<<<<<<<<<<<<<
 * StConstOutputs = '%sOUTPUTS = %d'
 * StConstOrder = '%sORDER = %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StConstInputs, __pyx_kp_u_sINPUTS_d) < 0) __PYX_ERR(0, 340, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":341
 * stIncludeFileP = '{$I %s}'
 * StConstInputs = '%sINPUTS= %d'
 * StConstOutputs = '%sOUTPUTS = %d'             # <<<<<<<<<<<<<<
 * StConstOrder = '%sORDER = %d'
 * stConstNodes = '%sNODES = %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StConstOutputs, __pyx_kp_u_sOUTPUTS_d) < 0) __PYX_ERR(0, 341, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":342
 * StConstInputs = '%sINPUTS= %d'
 * StConstOutputs = '%sOUTPUTS = %d'
 * StConstOrder = '%sORDER = %d'             # <<<<<<<<<<<<<<
 * stConstNodes = '%sNODES = %d'
 * stConstSynapses = '%sSYNAPSES= %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StConstOrder, __pyx_kp_u_sORDER_d) < 0) __PYX_ERR(0, 342, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":343
 * StConstOutputs = '%sOUTPUTS = %d'
 * StConstOrder = '%sORDER = %d'
 * stConstNodes = '%sNODES = %d'             # <<<<<<<<<<<<<<
 * stConstSynapses = '%sSYNAPSES= %d'
 * StConstConstants = '%sCONSTANTS = %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stConstNodes, __pyx_kp_u_sNODES_d) < 0) __PYX_ERR(0, 343, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":344
 * StConstOrder = '%sORDER = %d'
 * stConstNodes = '%sNODES = %d'
 * stConstSynapses = '%sSYNAPSES= %d'             # <<<<<<<<<<<<<<
 * StConstConstants = '%sCONSTANTS = %d'
 * StConstDimension = '%sDIMENSION = %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stConstSynapses, __pyx_kp_u_sSYNAPSES_d) < 0) __PYX_ERR(0, 344, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":345
 * stConstNodes = '%sNODES = %d'
 * stConstSynapses = '%sSYNAPSES= %d'
 * StConstConstants = '%sCONSTANTS = %d'             # <<<<<<<<<<<<<<
 * StConstDimension = '%sDIMENSION = %d'
 * StConstDimn = '%sDIM%d= %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StConstConstants, __pyx_kp_u_sCONSTANTS_d) < 0) __PYX_ERR(0, 345, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":346
 * stConstSynapses = '%sSYNAPSES= %d'
 * StConstConstants = '%sCONSTANTS = %d'
 * StConstDimension = '%sDIMENSION = %d'             # <<<<<<<<<<<<<<
 * StConstDimn = '%sDIM%d= %d'
 * stConstMatrix = '%sMATRIXES= %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StConstDimension, __pyx_kp_u_sDIMENSION_d) < 0) __PYX_ERR(0, 346, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":347
 * StConstConstants = '%sCONSTANTS = %d'
 * StConstDimension = '%sDIMENSION = %d'
 * StConstDimn = '%sDIM%d= %d'             # <<<<<<<<<<<<<<
 * stConstMatrix = '%sMATRIXES= %d'
 * stConstLinks = '%sMEMSYN= %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StConstDimn, __pyx_kp_u_sDIM_d_d) < 0) __PYX_ERR(0, 347, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":348
 * StConstDimension = '%sDIMENSION = %d'
 * StConstDimn = '%sDIM%d= %d'
 * stConstMatrix = '%sMATRIXES= %d'             # <<<<<<<<<<<<<<
 * stConstLinks = '%sMEMSYN= %d'
 * StConstStud95 = '%sSTUDENT95 = %g'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stConstMatrix, __pyx_kp_u_sMATRIXES_d) < 0) __PYX_ERR(0, 348, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":349
 * StConstDimn = '%sDIM%d= %d'
 * stConstMatrix = '%sMATRIXES= %d'
 * stConstLinks = '%sMEMSYN= %d'             # <<<<<<<<<<<<<<
 * StConstStud95 = '%sSTUDENT95 = %g'
 * StConstStdDev = '%sSTDTRDEV[%d] = %g'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stConstLinks, __pyx_kp_u_sMEMSYN_d) < 0) __PYX_ERR(0, 349, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":350
 * stConstMatrix = '%sMATRIXES= %d'
 * stConstLinks = '%sMEMSYN= %d'
 * StConstStud95 = '%sSTUDENT95 = %g'             # <<<<<<<<<<<<<<
 * StConstStdDev = '%sSTDTRDEV[%d] = %g'
 * StConstStdDev1 = '%sSTDTRDEV = %g'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StConstStud95, __pyx_kp_u_sSTUDENT95_g) < 0) __PYX_ERR(0, 350, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":351
 * stConstLinks = '%sMEMSYN= %d'
 * StConstStud95 = '%sSTUDENT95 = %g'
 * StConstStdDev = '%sSTDTRDEV[%d] = %g'             # <<<<<<<<<<<<<<
 * StConstStdDev1 = '%sSTDTRDEV = %g'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StConstStdDev, __pyx_kp_u_sSTDTRDEV_d_g) < 0) __PYX_ERR(0, 351, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":352
 * StConstStud95 = '%sSTUDENT95 = %g'
 * StConstStdDev = '%sSTDTRDEV[%d] = %g'
 * StConstStdDev1 = '%sSTDTRDEV = %g'             # <<<<<<<<<<<<<<
 * 
 * StWeightIniSP = 'weights: %s'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StConstStdDev1, __pyx_kp_u_sSTDTRDEV_g) < 0) __PYX_ERR(0, 352, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":354
 * StConstStdDev1 = '%sSTDTRDEV = %g'
 * 
 * StWeightIniSP = 'weights: %s'             # <<<<<<<<<<<<<<
 * StMinIniSP = 'min: %s'
 * StMaxIniSP = 'max: %s'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StWeightIniSP, __pyx_kp_u_weights_s) < 0) __PYX_ERR(0, 354, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":355
 * 
 * StWeightIniSP = 'weights: %s'
 * StMinIniSP = 'min: %s'             # <<<<<<<<<<<<<<
 * StMaxIniSP = 'max: %s'
 * StStdDEvIniP = 'StdDev: %s'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StMinIniSP, __pyx_kp_u_min_s) < 0) __PYX_ERR(0, 355, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":356
 * StWeightIniSP = 'weights: %s'
 * StMinIniSP = 'min: %s'
 * StMaxIniSP = 'max: %s'             # <<<<<<<<<<<<<<
 * StStdDEvIniP = 'StdDev: %s'
 * StMatIniSP = 'matrix: %s'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StMaxIniSP, __pyx_kp_u_max_s) < 0) __PYX_ERR(0, 356, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":357
 * StMinIniSP = 'min: %s'
 * StMaxIniSP = 'max: %s'
 * StStdDEvIniP = 'StdDev: %s'             # <<<<<<<<<<<<<<
 * StMatIniSP = 'matrix: %s'
 * stPotentialdSP = 'potential: Double'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StStdDEvIniP, __pyx_kp_u_StdDev_s) < 0) __PYX_ERR(0, 357, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":358
 * StMaxIniSP = 'max: %s'
 * StStdDEvIniP = 'StdDev: %s'
 * StMatIniSP = 'matrix: %s'             # <<<<<<<<<<<<<<
 * stPotentialdSP = 'potential: Double'
 * stXPotentialdSP = 'X, potential: Double'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StMatIniSP, __pyx_kp_u_matrix_s) < 0) __PYX_ERR(0, 358, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":359
 * StStdDEvIniP = 'StdDev: %s'
 * StMatIniSP = 'matrix: %s'
 * stPotentialdSP = 'potential: Double'             # <<<<<<<<<<<<<<
 * stXPotentialdSP = 'X, potential: Double'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stPotentialdSP, __pyx_kp_u_potential_Double) < 0) __PYX_ERR(0, 359, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":360
 * StMatIniSP = 'matrix: %s'
 * stPotentialdSP = 'potential: Double'
 * stXPotentialdSP = 'X, potential: Double'             # <<<<<<<<<<<<<<
 * 
 * StInputNameDeclP = 'names: Array[0..%sINPUTS-1] of String = ('
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stXPotentialdSP, __pyx_kp_u_X_potential_Double) < 0) __PYX_ERR(0, 360, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":362
 * stXPotentialdSP = 'X, potential: Double'
 * 
 * StInputNameDeclP = 'names: Array[0..%sINPUTS-1] of String = ('             # <<<<<<<<<<<<<<
 * StOutputNameDeclP = 'names: Array[0..%sOUTPUTS-1] of String = ('
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StInputNameDeclP, __pyx_kp_u_names_Array_0_sINPUTS_1_of_Strin) < 0) __PYX_ERR(0, 362, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":363
 * 
 * StInputNameDeclP = 'names: Array[0..%sINPUTS-1] of String = ('
 * StOutputNameDeclP = 'names: Array[0..%sOUTPUTS-1] of String = ('             # <<<<<<<<<<<<<<
 * 
 * stTypeSynapse = 'T%sSynapse'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StOutputNameDeclP, __pyx_kp_u_names_Array_0_sOUTPUTS_1_of_Stri) < 0) __PYX_ERR(0, 363, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":365
 * StOutputNameDeclP = 'names: Array[0..%sOUTPUTS-1] of String = ('
 * 
 * stTypeSynapse = 'T%sSynapse'             # <<<<<<<<<<<<<<
 * stTypeMinMax = 'T%sInOut'
 * #stDimKohP = 'T%sDimension'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stTypeSynapse, __pyx_kp_u_T_sSynapse) < 0) __PYX_ERR(0, 365, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":366
 * 
 * stTypeSynapse = 'T%sSynapse'
 * stTypeMinMax = 'T%sInOut'             # <<<<<<<<<<<<<<
 * #stDimKohP = 'T%sDimension'
 * stTypeOutput = 'T%sOutput'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stTypeMinMax, __pyx_kp_u_T_sInOut) < 0) __PYX_ERR(0, 366, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":368
 * stTypeMinMax = 'T%sInOut'
 * #stDimKohP = 'T%sDimension'
 * stTypeOutput = 'T%sOutput'             # <<<<<<<<<<<<<<
 * stTypeRetroP = 'T%sMemsyn'
 * stTypeMatrixes = 'T%sMatrix'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stTypeOutput, __pyx_kp_u_T_sOutput) < 0) __PYX_ERR(0, 368, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":369
 * #stDimKohP = 'T%sDimension'
 * stTypeOutput = 'T%sOutput'
 * stTypeRetroP = 'T%sMemsyn'             # <<<<<<<<<<<<<<
 * stTypeMatrixes = 'T%sMatrix'
 * stTypeNodes = 'T%sNode'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stTypeRetroP, __pyx_kp_u_T_sMemsyn) < 0) __PYX_ERR(0, 369, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":370
 * stTypeOutput = 'T%sOutput'
 * stTypeRetroP = 'T%sMemsyn'
 * stTypeMatrixes = 'T%sMatrix'             # <<<<<<<<<<<<<<
 * stTypeNodes = 'T%sNode'
 * stGradientdSP = 'gradient: T%sSynapse'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stTypeMatrixes, __pyx_kp_u_T_sMatrix) < 0) __PYX_ERR(0, 370, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":371
 * stTypeRetroP = 'T%sMemsyn'
 * stTypeMatrixes = 'T%sMatrix'
 * stTypeNodes = 'T%sNode'             # <<<<<<<<<<<<<<
 * stGradientdSP = 'gradient: T%sSynapse'
 * StIniBackdSP = 'iniback: T%sOutput'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stTypeNodes, __pyx_kp_u_T_sNode) < 0) __PYX_ERR(0, 371, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":372
 * stTypeMatrixes = 'T%sMatrix'
 * stTypeNodes = 'T%sNode'
 * stGradientdSP = 'gradient: T%sSynapse'             # <<<<<<<<<<<<<<
 * StIniBackdSP = 'iniback: T%sOutput'
 * stNodedSP = 'node: T%sNode'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stGradientdSP, __pyx_kp_u_gradient_T_sSynapse) < 0) __PYX_ERR(0, 372, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":373
 * stTypeNodes = 'T%sNode'
 * stGradientdSP = 'gradient: T%sSynapse'
 * StIniBackdSP = 'iniback: T%sOutput'             # <<<<<<<<<<<<<<
 * stNodedSP = 'node: T%sNode'
 * stNodedSPK = 'node: T%sOutput'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StIniBackdSP, __pyx_kp_u_iniback_T_sOutput) < 0) __PYX_ERR(0, 373, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":374
 * stGradientdSP = 'gradient: T%sSynapse'
 * StIniBackdSP = 'iniback: T%sOutput'
 * stNodedSP = 'node: T%sNode'             # <<<<<<<<<<<<<<
 * stNodedSPK = 'node: T%sOutput'
 * stNodedSP2 = 'node: T%sNode2'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stNodedSP, __pyx_kp_u_node_T_sNode) < 0) __PYX_ERR(0, 374, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":375
 * StIniBackdSP = 'iniback: T%sOutput'
 * stNodedSP = 'node: T%sNode'
 * stNodedSPK = 'node: T%sOutput'             # <<<<<<<<<<<<<<
 * stNodedSP2 = 'node: T%sNode2'
 * stNodedSP3 = 'node: T%sNode3'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stNodedSPK, __pyx_kp_u_node_T_sOutput) < 0) __PYX_ERR(0, 375, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":376
 * stNodedSP = 'node: T%sNode'
 * stNodedSPK = 'node: T%sOutput'
 * stNodedSP2 = 'node: T%sNode2'             # <<<<<<<<<<<<<<
 * stNodedSP3 = 'node: T%sNode3'
 * stMemSyndSP = 'memsyn: T%sMemsyn'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stNodedSP2, __pyx_kp_u_node_T_sNode2) < 0) __PYX_ERR(0, 376, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":377
 * stNodedSPK = 'node: T%sOutput'
 * stNodedSP2 = 'node: T%sNode2'
 * stNodedSP3 = 'node: T%sNode3'             # <<<<<<<<<<<<<<
 * stMemSyndSP = 'memsyn: T%sMemsyn'
 * stArrayOfDoubleDef = 'ADouble = Array [0..65535] Of Double'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stNodedSP3, __pyx_kp_u_node_T_sNode3) < 0) __PYX_ERR(0, 377, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":378
 * stNodedSP2 = 'node: T%sNode2'
 * stNodedSP3 = 'node: T%sNode3'
 * stMemSyndSP = 'memsyn: T%sMemsyn'             # <<<<<<<<<<<<<<
 * stArrayOfDoubleDef = 'ADouble = Array [0..65535] Of Double'
 * stPArrayOfDoubleDef = 'PADouble = ^ADouble'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stMemSyndSP, __pyx_kp_u_memsyn_T_sMemsyn) < 0) __PYX_ERR(0, 378, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":379
 * stNodedSP3 = 'node: T%sNode3'
 * stMemSyndSP = 'memsyn: T%sMemsyn'
 * stArrayOfDoubleDef = 'ADouble = Array [0..65535] Of Double'             # <<<<<<<<<<<<<<
 * stPArrayOfDoubleDef = 'PADouble = ^ADouble'
 * StArrayOutputsP = 'Array[0..%sOUTPUTS-1] of double'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stArrayOfDoubleDef, __pyx_kp_u_ADouble_Array_0_65535_Of_Double) < 0) __PYX_ERR(0, 379, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":380
 * stMemSyndSP = 'memsyn: T%sMemsyn'
 * stArrayOfDoubleDef = 'ADouble = Array [0..65535] Of Double'
 * stPArrayOfDoubleDef = 'PADouble = ^ADouble'             # <<<<<<<<<<<<<<
 * StArrayOutputsP = 'Array[0..%sOUTPUTS-1] of double'
 * StArraySynapsesP = 'Array[0..%sSYNAPSES-1] of double'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stPArrayOfDoubleDef, __pyx_kp_u_PADouble_ADouble) < 0) __PYX_ERR(0, 380, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":381
 * stArrayOfDoubleDef = 'ADouble = Array [0..65535] Of Double'
 * stPArrayOfDoubleDef = 'PADouble = ^ADouble'
 * StArrayOutputsP = 'Array[0..%sOUTPUTS-1] of double'             # <<<<<<<<<<<<<<
 * StArraySynapsesP = 'Array[0..%sSYNAPSES-1] of double'
 * stArrayMatrixesP = 'Array[0..%sMATRIXES-1] of double'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StArrayOutputsP, __pyx_kp_u_Array_0_sOUTPUTS_1_of_double) < 0) __PYX_ERR(0, 381, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":382
 * stPArrayOfDoubleDef = 'PADouble = ^ADouble'
 * StArrayOutputsP = 'Array[0..%sOUTPUTS-1] of double'
 * StArraySynapsesP = 'Array[0..%sSYNAPSES-1] of double'             # <<<<<<<<<<<<<<
 * stArrayMatrixesP = 'Array[0..%sMATRIXES-1] of double'
 * stArrayRetroP = 'Array[0..%sMEMSYN-1] of double'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StArraySynapsesP, __pyx_kp_u_Array_0_sSYNAPSES_1_of_double) < 0) __PYX_ERR(0, 382, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":383
 * StArrayOutputsP = 'Array[0..%sOUTPUTS-1] of double'
 * StArraySynapsesP = 'Array[0..%sSYNAPSES-1] of double'
 * stArrayMatrixesP = 'Array[0..%sMATRIXES-1] of double'             # <<<<<<<<<<<<<<
 * stArrayRetroP = 'Array[0..%sMEMSYN-1] of double'
 * StArrayNodesP = 'Array[0..%sNODES-1] of double'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stArrayMatrixesP, __pyx_kp_u_Array_0_sMATRIXES_1_of_double) < 0) __PYX_ERR(0, 383, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":384
 * StArraySynapsesP = 'Array[0..%sSYNAPSES-1] of double'
 * stArrayMatrixesP = 'Array[0..%sMATRIXES-1] of double'
 * stArrayRetroP = 'Array[0..%sMEMSYN-1] of double'             # <<<<<<<<<<<<<<
 * StArrayNodesP = 'Array[0..%sNODES-1] of double'
 * StArrayInOutP = 'Array[0..%sINPUTS + %sOUTPUTS-1] of double'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stArrayRetroP, __pyx_kp_u_Array_0_sMEMSYN_1_of_double) < 0) __PYX_ERR(0, 384, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":385
 * stArrayMatrixesP = 'Array[0..%sMATRIXES-1] of double'
 * stArrayRetroP = 'Array[0..%sMEMSYN-1] of double'
 * StArrayNodesP = 'Array[0..%sNODES-1] of double'             # <<<<<<<<<<<<<<
 * StArrayInOutP = 'Array[0..%sINPUTS + %sOUTPUTS-1] of double'
 * #StArrayKohDimP = 'Array[0..%sDIMENSION-1]'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StArrayNodesP, __pyx_kp_u_Array_0_sNODES_1_of_double) < 0) __PYX_ERR(0, 385, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":386
 * stArrayRetroP = 'Array[0..%sMEMSYN-1] of double'
 * StArrayNodesP = 'Array[0..%sNODES-1] of double'
 * StArrayInOutP = 'Array[0..%sINPUTS + %sOUTPUTS-1] of double'             # <<<<<<<<<<<<<<
 * #StArrayKohDimP = 'Array[0..%sDIMENSION-1]'
 * StArrayNodes2P = 'Array[0..%sNODES-1, 0..1] of double'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StArrayInOutP, __pyx_kp_u_Array_0_sINPUTS_sOUTPUTS_1_of_do) < 0) __PYX_ERR(0, 386, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":388
 * StArrayInOutP = 'Array[0..%sINPUTS + %sOUTPUTS-1] of double'
 * #StArrayKohDimP = 'Array[0..%sDIMENSION-1]'
 * StArrayNodes2P = 'Array[0..%sNODES-1, 0..1] of double'             # <<<<<<<<<<<<<<
 * StArrayNodes3P = 'Array[0..%sNODES-1, 0..2] of double'
 * StTransferStatPP = 'PROCEDURE %stransfer(inputs, outputs: PADouble)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StArrayNodes2P, __pyx_kp_u_Array_0_sNODES_1_0_1_of_double) < 0) __PYX_ERR(0, 388, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":389
 * #StArrayKohDimP = 'Array[0..%sDIMENSION-1]'
 * StArrayNodes2P = 'Array[0..%sNODES-1, 0..1] of double'
 * StArrayNodes3P = 'Array[0..%sNODES-1, 0..2] of double'             # <<<<<<<<<<<<<<
 * StTransferStatPP = 'PROCEDURE %stransfer(inputs, outputs: PADouble)'
 * #StTransferKohP = 'FUNCTION %stransferK(inputs: PADouble; var Coord: T%sDimension): Integer'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StArrayNodes3P, __pyx_kp_u_Array_0_sNODES_1_0_2_of_double) < 0) __PYX_ERR(0, 389, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":390
 * StArrayNodes2P = 'Array[0..%sNODES-1, 0..1] of double'
 * StArrayNodes3P = 'Array[0..%sNODES-1, 0..2] of double'
 * StTransferStatPP = 'PROCEDURE %stransfer(inputs, outputs: PADouble)'             # <<<<<<<<<<<<<<
 * #StTransferKohP = 'FUNCTION %stransferK(inputs: PADouble; var Coord: T%sDimension): Integer'
 * StTransferStatPP2 = 'PROCEDURE %stransferlev(inputs, outputs, leverage: PADouble)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferStatPP, __pyx_kp_u_PROCEDURE_stransfer_inputs_outpu) < 0) __PYX_ERR(0, 390, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":392
 * StTransferStatPP = 'PROCEDURE %stransfer(inputs, outputs: PADouble)'
 * #StTransferKohP = 'FUNCTION %stransferK(inputs: PADouble; var Coord: T%sDimension): Integer'
 * StTransferStatPP2 = 'PROCEDURE %stransferlev(inputs, outputs, leverage: PADouble)'             # <<<<<<<<<<<<<<
 * StTransGradStatPP = 'PROCEDURE %stransfergradient(inputs, outputs, gradient, iniback: PADouble)'
 * StTransGradStatCallP = '%stransfergradient(inputs, outputs, @gradient, @iniback)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferStatPP2, __pyx_kp_u_PROCEDURE_stransferlev_inputs_ou) < 0) __PYX_ERR(0, 392, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":393
 * #StTransferKohP = 'FUNCTION %stransferK(inputs: PADouble; var Coord: T%sDimension): Integer'
 * StTransferStatPP2 = 'PROCEDURE %stransferlev(inputs, outputs, leverage: PADouble)'
 * StTransGradStatPP = 'PROCEDURE %stransfergradient(inputs, outputs, gradient, iniback: PADouble)'             # <<<<<<<<<<<<<<
 * StTransGradStatCallP = '%stransfergradient(inputs, outputs, @gradient, @iniback)'
 * StTransferDynPP = 'PROCEDURE %stransfer(inputs, outputs, states: PADouble)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransGradStatPP, __pyx_kp_u_PROCEDURE_stransfergradient_inpu) < 0) __PYX_ERR(0, 393, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":394
 * StTransferStatPP2 = 'PROCEDURE %stransferlev(inputs, outputs, leverage: PADouble)'
 * StTransGradStatPP = 'PROCEDURE %stransfergradient(inputs, outputs, gradient, iniback: PADouble)'
 * StTransGradStatCallP = '%stransfergradient(inputs, outputs, @gradient, @iniback)'             # <<<<<<<<<<<<<<
 * StTransferDynPP = 'PROCEDURE %stransfer(inputs, outputs, states: PADouble)'
 * StTransferDynPP2 = 'PROCEDURE %stranserlev(inputs, outputs, states, leverage: PADouble)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransGradStatCallP, __pyx_kp_u_stransfergradient_inputs_output_3) < 0) __PYX_ERR(0, 394, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":395
 * StTransGradStatPP = 'PROCEDURE %stransfergradient(inputs, outputs, gradient, iniback: PADouble)'
 * StTransGradStatCallP = '%stransfergradient(inputs, outputs, @gradient, @iniback)'
 * StTransferDynPP = 'PROCEDURE %stransfer(inputs, outputs, states: PADouble)'             # <<<<<<<<<<<<<<
 * StTransferDynPP2 = 'PROCEDURE %stranserlev(inputs, outputs, states, leverage: PADouble)'
 * StTransGradDynPP = 'PROCEDURE %stransfergradient(inputs, outputs, states, gradient, iniback: PADouble)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferDynPP, __pyx_kp_u_PROCEDURE_stransfer_inputs_outpu_2) < 0) __PYX_ERR(0, 395, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":396
 * StTransGradStatCallP = '%stransfergradient(inputs, outputs, @gradient, @iniback)'
 * StTransferDynPP = 'PROCEDURE %stransfer(inputs, outputs, states: PADouble)'
 * StTransferDynPP2 = 'PROCEDURE %stranserlev(inputs, outputs, states, leverage: PADouble)'             # <<<<<<<<<<<<<<
 * StTransGradDynPP = 'PROCEDURE %stransfergradient(inputs, outputs, states, gradient, iniback: PADouble)'
 * StTransGradDynCallP = '%stransfergradient(inputs, outputs, states, @gradient, @iniback)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferDynPP2, __pyx_kp_u_PROCEDURE_stranserlev_inputs_out) < 0) __PYX_ERR(0, 396, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":397
 * StTransferDynPP = 'PROCEDURE %stransfer(inputs, outputs, states: PADouble)'
 * StTransferDynPP2 = 'PROCEDURE %stranserlev(inputs, outputs, states, leverage: PADouble)'
 * StTransGradDynPP = 'PROCEDURE %stransfergradient(inputs, outputs, states, gradient, iniback: PADouble)'             # <<<<<<<<<<<<<<
 * StTransGradDynCallP = '%stransfergradient(inputs, outputs, states, @gradient, @iniback)'
 * StInputNameP = 'FUNCTION %sinputname(index: integer): String'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransGradDynPP, __pyx_kp_u_PROCEDURE_stransfergradient_inpu_2) < 0) __PYX_ERR(0, 397, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":398
 * StTransferDynPP2 = 'PROCEDURE %stranserlev(inputs, outputs, states, leverage: PADouble)'
 * StTransGradDynPP = 'PROCEDURE %stransfergradient(inputs, outputs, states, gradient, iniback: PADouble)'
 * StTransGradDynCallP = '%stransfergradient(inputs, outputs, states, @gradient, @iniback)'             # <<<<<<<<<<<<<<
 * StInputNameP = 'FUNCTION %sinputname(index: integer): String'
 * StOutputNameP = 'FUNCTION %soutputname(index: integer): String'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransGradDynCallP, __pyx_kp_u_stransfergradient_inputs_output_4) < 0) __PYX_ERR(0, 398, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":399
 * StTransGradDynPP = 'PROCEDURE %stransfergradient(inputs, outputs, states, gradient, iniback: PADouble)'
 * StTransGradDynCallP = '%stransfergradient(inputs, outputs, states, @gradient, @iniback)'
 * StInputNameP = 'FUNCTION %sinputname(index: integer): String'             # <<<<<<<<<<<<<<
 * StOutputNameP = 'FUNCTION %soutputname(index: integer): String'
 * StReturnInputNameP1 = \
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StInputNameP, __pyx_kp_u_FUNCTION_sinputname_index_intege) < 0) __PYX_ERR(0, 399, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":400
 * StTransGradDynCallP = '%stransfergradient(inputs, outputs, states, @gradient, @iniback)'
 * StInputNameP = 'FUNCTION %sinputname(index: integer): String'
 * StOutputNameP = 'FUNCTION %soutputname(index: integer): String'             # <<<<<<<<<<<<<<
 * StReturnInputNameP1 = \
 *     'if (index < 0) or (index >= %sINPUTS) then Result := ''''';
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StOutputNameP, __pyx_kp_u_FUNCTION_soutputname_index_integ) < 0) __PYX_ERR(0, 400, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":402
 * StOutputNameP = 'FUNCTION %soutputname(index: integer): String'
 * StReturnInputNameP1 = \
 *     'if (index < 0) or (index >= %sINPUTS) then Result := ''''';             # <<<<<<<<<<<<<<
 *   StReturnInputNameP2 = 'else Result := names[index]';
 *   StReturnOutputNameP1 =
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StReturnInputNameP1, __pyx_kp_u_if_index_0_or_index_sINPUTS_then) < 0) __PYX_ERR(0, 401, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":406
 *   StReturnOutputNameP1 =
 *     'if (index < 0) or (index >= %sOUTPUTS) then Result := '''''
 * StReturnOutputNameP2 = 'else Result := names[index]'             # <<<<<<<<<<<<<<
 * StMatrixComputeP = '%smatrix[k] * gradient[j]'
 * StForAllOutputsP = 'for i := 1 to %sOUTPUTS-1 do'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StReturnOutputNameP2, __pyx_kp_u_else_Result_names_index) < 0) __PYX_ERR(0, 406, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":407
 *     'if (index < 0) or (index >= %sOUTPUTS) then Result := '''''
 * StReturnOutputNameP2 = 'else Result := names[index]'
 * StMatrixComputeP = '%smatrix[k] * gradient[j]'             # <<<<<<<<<<<<<<
 * StForAllOutputsP = 'for i := 1 to %sOUTPUTS-1 do'
 * StForAllInitMemSynP = 'for i := 0 to %sMEMSYN-1 do memsyn[i] := 1'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StMatrixComputeP, __pyx_kp_u_smatrix_k_gradient_j) < 0) __PYX_ERR(0, 407, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":408
 * StReturnOutputNameP2 = 'else Result := names[index]'
 * StMatrixComputeP = '%smatrix[k] * gradient[j]'
 * StForAllOutputsP = 'for i := 1 to %sOUTPUTS-1 do'             # <<<<<<<<<<<<<<
 * StForAllInitMemSynP = 'for i := 0 to %sMEMSYN-1 do memsyn[i] := 1'
 * StForAllInitGradientP = 'for i := 0 to %sSYNAPSES-1 do gradient[i] := 0'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllOutputsP, __pyx_kp_u_for_i_1_to_sOUTPUTS_1_do) < 0) __PYX_ERR(0, 408, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":409
 * StMatrixComputeP = '%smatrix[k] * gradient[j]'
 * StForAllOutputsP = 'for i := 1 to %sOUTPUTS-1 do'
 * StForAllInitMemSynP = 'for i := 0 to %sMEMSYN-1 do memsyn[i] := 1'             # <<<<<<<<<<<<<<
 * StForAllInitGradientP = 'for i := 0 to %sSYNAPSES-1 do gradient[i] := 0'
 * StForAllInitIniBackP = 'for i := 0 to %sOUTPUTS-1 do iniback[i] := 0'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllInitMemSynP, __pyx_kp_u_for_i_0_to_sMEMSYN_1_do_memsyn_i) < 0) __PYX_ERR(0, 409, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":410
 * StForAllOutputsP = 'for i := 1 to %sOUTPUTS-1 do'
 * StForAllInitMemSynP = 'for i := 0 to %sMEMSYN-1 do memsyn[i] := 1'
 * StForAllInitGradientP = 'for i := 0 to %sSYNAPSES-1 do gradient[i] := 0'             # <<<<<<<<<<<<<<
 * StForAllInitIniBackP = 'for i := 0 to %sOUTPUTS-1 do iniback[i] := 0'
 * StForAllInitLeverageP = 'for i := 0 to %sOUTPUTS-1 do Leverage[i] := 0'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllInitGradientP, __pyx_kp_u_for_i_0_to_sSYNAPSES_1_do_gradie) < 0) __PYX_ERR(0, 410, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":411
 * StForAllInitMemSynP = 'for i := 0 to %sMEMSYN-1 do memsyn[i] := 1'
 * StForAllInitGradientP = 'for i := 0 to %sSYNAPSES-1 do gradient[i] := 0'
 * StForAllInitIniBackP = 'for i := 0 to %sOUTPUTS-1 do iniback[i] := 0'             # <<<<<<<<<<<<<<
 * StForAllInitLeverageP = 'for i := 0 to %sOUTPUTS-1 do Leverage[i] := 0'
 * StForAllSynapsesP = 'for i := 0 to %sSYNAPSES-1 do'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllInitIniBackP, __pyx_kp_u_for_i_0_to_sOUTPUTS_1_do_iniback) < 0) __PYX_ERR(0, 411, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":412
 * StForAllInitGradientP = 'for i := 0 to %sSYNAPSES-1 do gradient[i] := 0'
 * StForAllInitIniBackP = 'for i := 0 to %sOUTPUTS-1 do iniback[i] := 0'
 * StForAllInitLeverageP = 'for i := 0 to %sOUTPUTS-1 do Leverage[i] := 0'             # <<<<<<<<<<<<<<
 * StForAllSynapsesP = 'for i := 0 to %sSYNAPSES-1 do'
 * StForAllSynapsesjP = 'for j := 0 to %sSYNAPSES-1 do'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllInitLeverageP, __pyx_kp_u_for_i_0_to_sOUTPUTS_1_do_Leverag) < 0) __PYX_ERR(0, 412, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":413
 * StForAllInitIniBackP = 'for i := 0 to %sOUTPUTS-1 do iniback[i] := 0'
 * StForAllInitLeverageP = 'for i := 0 to %sOUTPUTS-1 do Leverage[i] := 0'
 * StForAllSynapsesP = 'for i := 0 to %sSYNAPSES-1 do'             # <<<<<<<<<<<<<<
 * StForAllSynapsesjP = 'for j := 0 to %sSYNAPSES-1 do'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllSynapsesP, __pyx_kp_u_for_i_0_to_sSYNAPSES_1_do) < 0) __PYX_ERR(0, 413, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":414
 * StForAllInitLeverageP = 'for i := 0 to %sOUTPUTS-1 do Leverage[i] := 0'
 * StForAllSynapsesP = 'for i := 0 to %sSYNAPSES-1 do'
 * StForAllSynapsesjP = 'for j := 0 to %sSYNAPSES-1 do'             # <<<<<<<<<<<<<<
 * 
 * #Chaines pour Fortran
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllSynapsesjP, __pyx_kp_u_for_j_0_to_sSYNAPSES_1_do) < 0) __PYX_ERR(0, 414, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":417
 * 
 * #Chaines pour Fortran
 * StWeightIniF = 'WEIGHTS'             # <<<<<<<<<<<<<<
 * StMaxIniSF = 'MAX'
 * StMinIniSF = 'MIN'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StWeightIniF, __pyx_n_u_WEIGHTS) < 0) __PYX_ERR(0, 417, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":418
 * #Chaines pour Fortran
 * StWeightIniF = 'WEIGHTS'
 * StMaxIniSF = 'MAX'             # <<<<<<<<<<<<<<
 * StMinIniSF = 'MIN'
 * StStdDEvIniF = 'STDTRDEV'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StMaxIniSF, __pyx_n_u_MAX) < 0) __PYX_ERR(0, 418, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":419
 * StWeightIniF = 'WEIGHTS'
 * StMaxIniSF = 'MAX'
 * StMinIniSF = 'MIN'             # <<<<<<<<<<<<<<
 * StStdDEvIniF = 'STDTRDEV'
 * StMatrixIniF = 'MATRIX'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StMinIniSF, __pyx_n_u_MIN) < 0) __PYX_ERR(0, 419, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":420
 * StMaxIniSF = 'MAX'
 * StMinIniSF = 'MIN'
 * StStdDEvIniF = 'STDTRDEV'             # <<<<<<<<<<<<<<
 * StMatrixIniF = 'MATRIX'
 * StTransferStatF = 'SUBROUTINE %sTRANSFER (INPUTS, OUTPUTS)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StStdDEvIniF, __pyx_n_u_STDTRDEV) < 0) __PYX_ERR(0, 420, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":421
 * StMinIniSF = 'MIN'
 * StStdDEvIniF = 'STDTRDEV'
 * StMatrixIniF = 'MATRIX'             # <<<<<<<<<<<<<<
 * StTransferStatF = 'SUBROUTINE %sTRANSFER (INPUTS, OUTPUTS)'
 * StTransferStatF2 = 'SUBROUTINE %sTRANSFERLEV (INPUTS, OUTPUTS, LEVERAGE)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StMatrixIniF, __pyx_n_u_MATRIX) < 0) __PYX_ERR(0, 421, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":422
 * StStdDEvIniF = 'STDTRDEV'
 * StMatrixIniF = 'MATRIX'
 * StTransferStatF = 'SUBROUTINE %sTRANSFER (INPUTS, OUTPUTS)'             # <<<<<<<<<<<<<<
 * StTransferStatF2 = 'SUBROUTINE %sTRANSFERLEV (INPUTS, OUTPUTS, LEVERAGE)'
 * StTransGradStatF = 'SUBROUTINE %sTRANSFERGRADIENT (INPUTS, OUTPUTS, GRADIENT, INIBACK)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferStatF, __pyx_kp_u_SUBROUTINE_sTRANSFER_INPUTS_OUTP) < 0) __PYX_ERR(0, 422, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":423
 * StMatrixIniF = 'MATRIX'
 * StTransferStatF = 'SUBROUTINE %sTRANSFER (INPUTS, OUTPUTS)'
 * StTransferStatF2 = 'SUBROUTINE %sTRANSFERLEV (INPUTS, OUTPUTS, LEVERAGE)'             # <<<<<<<<<<<<<<
 * StTransGradStatF = 'SUBROUTINE %sTRANSFERGRADIENT (INPUTS, OUTPUTS, GRADIENT, INIBACK)'
 * StTransGradStatCallF = 'CALL %sTRANSFERGRADIENT (INPUTS, OUTPUTS, GRADIENT, INIBACK)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferStatF2, __pyx_kp_u_SUBROUTINE_sTRANSFERLEV_INPUTS_O) < 0) __PYX_ERR(0, 423, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":424
 * StTransferStatF = 'SUBROUTINE %sTRANSFER (INPUTS, OUTPUTS)'
 * StTransferStatF2 = 'SUBROUTINE %sTRANSFERLEV (INPUTS, OUTPUTS, LEVERAGE)'
 * StTransGradStatF = 'SUBROUTINE %sTRANSFERGRADIENT (INPUTS, OUTPUTS, GRADIENT, INIBACK)'             # <<<<<<<<<<<<<<
 * StTransGradStatCallF = 'CALL %sTRANSFERGRADIENT (INPUTS, OUTPUTS, GRADIENT, INIBACK)'
 * StTransferDynF = 'SUBROUTINE %sTRANSFER (INPUTS, OUTPUTS, STATES)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransGradStatF, __pyx_kp_u_SUBROUTINE_sTRANSFERGRADIENT_INP) < 0) __PYX_ERR(0, 424, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":425
 * StTransferStatF2 = 'SUBROUTINE %sTRANSFERLEV (INPUTS, OUTPUTS, LEVERAGE)'
 * StTransGradStatF = 'SUBROUTINE %sTRANSFERGRADIENT (INPUTS, OUTPUTS, GRADIENT, INIBACK)'
 * StTransGradStatCallF = 'CALL %sTRANSFERGRADIENT (INPUTS, OUTPUTS, GRADIENT, INIBACK)'             # <<<<<<<<<<<<<<
 * StTransferDynF = 'SUBROUTINE %sTRANSFER (INPUTS, OUTPUTS, STATES)'
 * StTransferDynF2 = 'SUBROUTINE %sTRANSFERLEV (INPUTS, OUTPUTS, STATES, LEVERAGE)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransGradStatCallF, __pyx_kp_u_CALL_sTRANSFERGRADIENT_INPUTS_OU) < 0) __PYX_ERR(0, 425, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":426
 * StTransGradStatF = 'SUBROUTINE %sTRANSFERGRADIENT (INPUTS, OUTPUTS, GRADIENT, INIBACK)'
 * StTransGradStatCallF = 'CALL %sTRANSFERGRADIENT (INPUTS, OUTPUTS, GRADIENT, INIBACK)'
 * StTransferDynF = 'SUBROUTINE %sTRANSFER (INPUTS, OUTPUTS, STATES)'             # <<<<<<<<<<<<<<
 * StTransferDynF2 = 'SUBROUTINE %sTRANSFERLEV (INPUTS, OUTPUTS, STATES, LEVERAGE)'
 * StTransGradDynF = 'SUBROUTINE %sTRANSFERGRADIENT (INPUTS, OUTPUTS, STATES, GRADIENT, XINIBACK)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferDynF, __pyx_kp_u_SUBROUTINE_sTRANSFER_INPUTS_OUTP_2) < 0) __PYX_ERR(0, 426, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":427
 * StTransGradStatCallF = 'CALL %sTRANSFERGRADIENT (INPUTS, OUTPUTS, GRADIENT, INIBACK)'
 * StTransferDynF = 'SUBROUTINE %sTRANSFER (INPUTS, OUTPUTS, STATES)'
 * StTransferDynF2 = 'SUBROUTINE %sTRANSFERLEV (INPUTS, OUTPUTS, STATES, LEVERAGE)'             # <<<<<<<<<<<<<<
 * StTransGradDynF = 'SUBROUTINE %sTRANSFERGRADIENT (INPUTS, OUTPUTS, STATES, GRADIENT, XINIBACK)'
 * StTransGradDynCallF = '%sTRANSFERGRADIENT (INPUTS, OUTPUTS, STATES, GRADIENT, INIBACK)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransferDynF2, __pyx_kp_u_SUBROUTINE_sTRANSFERLEV_INPUTS_O_2) < 0) __PYX_ERR(0, 427, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":428
 * StTransferDynF = 'SUBROUTINE %sTRANSFER (INPUTS, OUTPUTS, STATES)'
 * StTransferDynF2 = 'SUBROUTINE %sTRANSFERLEV (INPUTS, OUTPUTS, STATES, LEVERAGE)'
 * StTransGradDynF = 'SUBROUTINE %sTRANSFERGRADIENT (INPUTS, OUTPUTS, STATES, GRADIENT, XINIBACK)'             # <<<<<<<<<<<<<<
 * StTransGradDynCallF = '%sTRANSFERGRADIENT (INPUTS, OUTPUTS, STATES, GRADIENT, INIBACK)'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransGradDynF, __pyx_kp_u_SUBROUTINE_sTRANSFERGRADIENT_INP_2) < 0) __PYX_ERR(0, 428, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":429
 * StTransferDynF2 = 'SUBROUTINE %sTRANSFERLEV (INPUTS, OUTPUTS, STATES, LEVERAGE)'
 * StTransGradDynF = 'SUBROUTINE %sTRANSFERGRADIENT (INPUTS, OUTPUTS, STATES, GRADIENT, XINIBACK)'
 * StTransGradDynCallF = '%sTRANSFERGRADIENT (INPUTS, OUTPUTS, STATES, GRADIENT, INIBACK)'             # <<<<<<<<<<<<<<
 * 
 * StIntParam = 'INTEGER, PARAMETER :: '
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StTransGradDynCallF, __pyx_kp_u_sTRANSFERGRADIENT_INPUTS_OUTPUT) < 0) __PYX_ERR(0, 429, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":431
 * StTransGradDynCallF = '%sTRANSFERGRADIENT (INPUTS, OUTPUTS, STATES, GRADIENT, INIBACK)'
 * 
 * StIntParam = 'INTEGER, PARAMETER :: '             # <<<<<<<<<<<<<<
 * StConstInputsF = '%sINPUTS= %d'
 * StConstOutputsF = '%sOUTPUTS = %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StIntParam, __pyx_kp_u_INTEGER_PARAMETER) < 0) __PYX_ERR(0, 431, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":432
 * 
 * StIntParam = 'INTEGER, PARAMETER :: '
 * StConstInputsF = '%sINPUTS= %d'             # <<<<<<<<<<<<<<
 * StConstOutputsF = '%sOUTPUTS = %d'
 * StConstOrderF = '%sORDER = %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StConstInputsF, __pyx_kp_u_sINPUTS_d) < 0) __PYX_ERR(0, 432, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":433
 * StIntParam = 'INTEGER, PARAMETER :: '
 * StConstInputsF = '%sINPUTS= %d'
 * StConstOutputsF = '%sOUTPUTS = %d'             # <<<<<<<<<<<<<<
 * StConstOrderF = '%sORDER = %d'
 * stConstNodesF = '%sNODES = %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StConstOutputsF, __pyx_kp_u_sOUTPUTS_d) < 0) __PYX_ERR(0, 433, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":434
 * StConstInputsF = '%sINPUTS= %d'
 * StConstOutputsF = '%sOUTPUTS = %d'
 * StConstOrderF = '%sORDER = %d'             # <<<<<<<<<<<<<<
 * stConstNodesF = '%sNODES = %d'
 * stConstSynapsesF = '%sSYNAPSES= %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StConstOrderF, __pyx_kp_u_sORDER_d) < 0) __PYX_ERR(0, 434, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":435
 * StConstOutputsF = '%sOUTPUTS = %d'
 * StConstOrderF = '%sORDER = %d'
 * stConstNodesF = '%sNODES = %d'             # <<<<<<<<<<<<<<
 * stConstSynapsesF = '%sSYNAPSES= %d'
 * StConstConstantsF = '%sCONSTANTS = %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stConstNodesF, __pyx_kp_u_sNODES_d) < 0) __PYX_ERR(0, 435, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":436
 * StConstOrderF = '%sORDER = %d'
 * stConstNodesF = '%sNODES = %d'
 * stConstSynapsesF = '%sSYNAPSES= %d'             # <<<<<<<<<<<<<<
 * StConstConstantsF = '%sCONSTANTS = %d'
 * stConstMatrixF = '%sMATRIXES= %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stConstSynapsesF, __pyx_kp_u_sSYNAPSES_d) < 0) __PYX_ERR(0, 436, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":437
 * stConstNodesF = '%sNODES = %d'
 * stConstSynapsesF = '%sSYNAPSES= %d'
 * StConstConstantsF = '%sCONSTANTS = %d'             # <<<<<<<<<<<<<<
 * stConstMatrixF = '%sMATRIXES= %d'
 * stConstLinksF = '%sMEMSYN= %d'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StConstConstantsF, __pyx_kp_u_sCONSTANTS_d) < 0) __PYX_ERR(0, 437, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":438
 * stConstSynapsesF = '%sSYNAPSES= %d'
 * StConstConstantsF = '%sCONSTANTS = %d'
 * stConstMatrixF = '%sMATRIXES= %d'             # <<<<<<<<<<<<<<
 * stConstLinksF = '%sMEMSYN= %d'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stConstMatrixF, __pyx_kp_u_sMATRIXES_d) < 0) __PYX_ERR(0, 438, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":439
 * StConstConstantsF = '%sCONSTANTS = %d'
 * stConstMatrixF = '%sMATRIXES= %d'
 * stConstLinksF = '%sMEMSYN= %d'             # <<<<<<<<<<<<<<
 * 
 * stPotentialdSF77 = 'REAL*8 POTENTIAL'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stConstLinksF, __pyx_kp_u_sMEMSYN_d) < 0) __PYX_ERR(0, 439, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":441
 * stConstLinksF = '%sMEMSYN= %d'
 * 
 * stPotentialdSF77 = 'REAL*8 POTENTIAL'             # <<<<<<<<<<<<<<
 * stXdSF77 = 'REAL*8 X'
 * stXPotentialdSF77 = 'REAL*8 X, POTENTIAL'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stPotentialdSF77, __pyx_kp_u_REAL_8_POTENTIAL) < 0) __PYX_ERR(0, 441, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":442
 * 
 * stPotentialdSF77 = 'REAL*8 POTENTIAL'
 * stXdSF77 = 'REAL*8 X'             # <<<<<<<<<<<<<<
 * stXPotentialdSF77 = 'REAL*8 X, POTENTIAL'
 * StArraySynapsesF77 = 'DIMENSION REAL*8 %s (%sSYNAPSES)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stXdSF77, __pyx_kp_u_REAL_8_X) < 0) __PYX_ERR(0, 442, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":443
 * stPotentialdSF77 = 'REAL*8 POTENTIAL'
 * stXdSF77 = 'REAL*8 X'
 * stXPotentialdSF77 = 'REAL*8 X, POTENTIAL'             # <<<<<<<<<<<<<<
 * StArraySynapsesF77 = 'DIMENSION REAL*8 %s (%sSYNAPSES)'
 * StArrayMatrixF77 = 'DIMENSION REAL*8 %s (%sMATRIXES)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stXPotentialdSF77, __pyx_kp_u_REAL_8_X_POTENTIAL) < 0) __PYX_ERR(0, 443, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":444
 * stXdSF77 = 'REAL*8 X'
 * stXPotentialdSF77 = 'REAL*8 X, POTENTIAL'
 * StArraySynapsesF77 = 'DIMENSION REAL*8 %s (%sSYNAPSES)'             # <<<<<<<<<<<<<<
 * StArrayMatrixF77 = 'DIMENSION REAL*8 %s (%sMATRIXES)'
 * stStatesDeclF77 = 'DIMENSION REAL*8 STATES (%sORDER)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StArraySynapsesF77, __pyx_kp_u_DIMENSION_REAL_8_s_sSYNAPSES) < 0) __PYX_ERR(0, 444, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":445
 * stXPotentialdSF77 = 'REAL*8 X, POTENTIAL'
 * StArraySynapsesF77 = 'DIMENSION REAL*8 %s (%sSYNAPSES)'
 * StArrayMatrixF77 = 'DIMENSION REAL*8 %s (%sMATRIXES)'             # <<<<<<<<<<<<<<
 * stStatesDeclF77 = 'DIMENSION REAL*8 STATES (%sORDER)'
 * stNodedSF77 = 'DIMENSION REAL*8 NODE (%sNODES)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StArrayMatrixF77, __pyx_kp_u_DIMENSION_REAL_8_s_sMATRIXES) < 0) __PYX_ERR(0, 445, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":446
 * StArraySynapsesF77 = 'DIMENSION REAL*8 %s (%sSYNAPSES)'
 * StArrayMatrixF77 = 'DIMENSION REAL*8 %s (%sMATRIXES)'
 * stStatesDeclF77 = 'DIMENSION REAL*8 STATES (%sORDER)'             # <<<<<<<<<<<<<<
 * stNodedSF77 = 'DIMENSION REAL*8 NODE (%sNODES)'
 * stNodedS2F77 = 'DIMENSION REAL*8 NODE (2, %sNODES)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stStatesDeclF77, __pyx_kp_u_DIMENSION_REAL_8_STATES_sORDER) < 0) __PYX_ERR(0, 446, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":447
 * StArrayMatrixF77 = 'DIMENSION REAL*8 %s (%sMATRIXES)'
 * stStatesDeclF77 = 'DIMENSION REAL*8 STATES (%sORDER)'
 * stNodedSF77 = 'DIMENSION REAL*8 NODE (%sNODES)'             # <<<<<<<<<<<<<<
 * stNodedS2F77 = 'DIMENSION REAL*8 NODE (2, %sNODES)'
 * stNodedS3F77 = 'DIMENSION REAL*8 NODE (3, %sNODES)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stNodedSF77, __pyx_kp_u_DIMENSION_REAL_8_NODE_sNODES) < 0) __PYX_ERR(0, 447, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":448
 * stStatesDeclF77 = 'DIMENSION REAL*8 STATES (%sORDER)'
 * stNodedSF77 = 'DIMENSION REAL*8 NODE (%sNODES)'
 * stNodedS2F77 = 'DIMENSION REAL*8 NODE (2, %sNODES)'             # <<<<<<<<<<<<<<
 * stNodedS3F77 = 'DIMENSION REAL*8 NODE (3, %sNODES)'
 * stMemSyndSF77 = 'DIMENSION REAL*8 MEMSYN (%sMEMSYN)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stNodedS2F77, __pyx_kp_u_DIMENSION_REAL_8_NODE_2_sNODES) < 0) __PYX_ERR(0, 448, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":449
 * stNodedSF77 = 'DIMENSION REAL*8 NODE (%sNODES)'
 * stNodedS2F77 = 'DIMENSION REAL*8 NODE (2, %sNODES)'
 * stNodedS3F77 = 'DIMENSION REAL*8 NODE (3, %sNODES)'             # <<<<<<<<<<<<<<
 * stMemSyndSF77 = 'DIMENSION REAL*8 MEMSYN (%sMEMSYN)'
 * StInputsDeclF77 = 'DIMENSION REAL*8 INPUTS (%sINPUTS)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stNodedS3F77, __pyx_kp_u_DIMENSION_REAL_8_NODE_3_sNODES) < 0) __PYX_ERR(0, 449, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":450
 * stNodedS2F77 = 'DIMENSION REAL*8 NODE (2, %sNODES)'
 * stNodedS3F77 = 'DIMENSION REAL*8 NODE (3, %sNODES)'
 * stMemSyndSF77 = 'DIMENSION REAL*8 MEMSYN (%sMEMSYN)'             # <<<<<<<<<<<<<<
 * StInputsDeclF77 = 'DIMENSION REAL*8 INPUTS (%sINPUTS)'
 * StOutPutsDeclF77 = 'DIMENSION REAL*8 OUTPUTS (%sOUTPUTS)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stMemSyndSF77, __pyx_kp_u_DIMENSION_REAL_8_MEMSYN_sMEMSYN) < 0) __PYX_ERR(0, 450, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":451
 * stNodedS3F77 = 'DIMENSION REAL*8 NODE (3, %sNODES)'
 * stMemSyndSF77 = 'DIMENSION REAL*8 MEMSYN (%sMEMSYN)'
 * StInputsDeclF77 = 'DIMENSION REAL*8 INPUTS (%sINPUTS)'             # <<<<<<<<<<<<<<
 * StOutPutsDeclF77 = 'DIMENSION REAL*8 OUTPUTS (%sOUTPUTS)'
 * stGradientDeclF77 = 'DIMENSION REAL*8 GRADIENT (%sSYNAPSES)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StInputsDeclF77, __pyx_kp_u_DIMENSION_REAL_8_INPUTS_sINPUTS) < 0) __PYX_ERR(0, 451, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":452
 * stMemSyndSF77 = 'DIMENSION REAL*8 MEMSYN (%sMEMSYN)'
 * StInputsDeclF77 = 'DIMENSION REAL*8 INPUTS (%sINPUTS)'
 * StOutPutsDeclF77 = 'DIMENSION REAL*8 OUTPUTS (%sOUTPUTS)'             # <<<<<<<<<<<<<<
 * stGradientDeclF77 = 'DIMENSION REAL*8 GRADIENT (%sSYNAPSES)'
 * stInibackDeclF77 = 'DIMENSION REAL*8 INIBACK (%sOUTPUTS)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StOutPutsDeclF77, __pyx_kp_u_DIMENSION_REAL_8_OUTPUTS_sOUTPUT) < 0) __PYX_ERR(0, 452, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":453
 * StInputsDeclF77 = 'DIMENSION REAL*8 INPUTS (%sINPUTS)'
 * StOutPutsDeclF77 = 'DIMENSION REAL*8 OUTPUTS (%sOUTPUTS)'
 * stGradientDeclF77 = 'DIMENSION REAL*8 GRADIENT (%sSYNAPSES)'             # <<<<<<<<<<<<<<
 * stInibackDeclF77 = 'DIMENSION REAL*8 INIBACK (%sOUTPUTS)'
 * stLeverageDeclF77 = 'DIMENSION REAL*8 LEVERAGE (%sOUTPUTS)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stGradientDeclF77, __pyx_kp_u_DIMENSION_REAL_8_GRADIENT_sSYNAP) < 0) __PYX_ERR(0, 453, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":454
 * StOutPutsDeclF77 = 'DIMENSION REAL*8 OUTPUTS (%sOUTPUTS)'
 * stGradientDeclF77 = 'DIMENSION REAL*8 GRADIENT (%sSYNAPSES)'
 * stInibackDeclF77 = 'DIMENSION REAL*8 INIBACK (%sOUTPUTS)'             # <<<<<<<<<<<<<<
 * stLeverageDeclF77 = 'DIMENSION REAL*8 LEVERAGE (%sOUTPUTS)'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stInibackDeclF77, __pyx_kp_u_DIMENSION_REAL_8_INIBACK_sOUTPUT) < 0) __PYX_ERR(0, 454, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":455
 * stGradientDeclF77 = 'DIMENSION REAL*8 GRADIENT (%sSYNAPSES)'
 * stInibackDeclF77 = 'DIMENSION REAL*8 INIBACK (%sOUTPUTS)'
 * stLeverageDeclF77 = 'DIMENSION REAL*8 LEVERAGE (%sOUTPUTS)'             # <<<<<<<<<<<<<<
 * 
 * stPotentialdSF90 = 'REAL*8 POTENTIAL'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stLeverageDeclF77, __pyx_kp_u_DIMENSION_REAL_8_LEVERAGE_sOUTPU) < 0) __PYX_ERR(0, 455, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":457
 * stLeverageDeclF77 = 'DIMENSION REAL*8 LEVERAGE (%sOUTPUTS)'
 * 
 * stPotentialdSF90 = 'REAL*8 POTENTIAL'             # <<<<<<<<<<<<<<
 * stXdSF90 = 'REAL*8 X'
 * stXPotentialdSF90 = 'REAL*8 X, POTENTIAL'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stPotentialdSF90, __pyx_kp_u_REAL_8_POTENTIAL) < 0) __PYX_ERR(0, 457, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":458
 * 
 * stPotentialdSF90 = 'REAL*8 POTENTIAL'
 * stXdSF90 = 'REAL*8 X'             # <<<<<<<<<<<<<<
 * stXPotentialdSF90 = 'REAL*8 X, POTENTIAL'
 * StArraySynapsesF90 = 'REAL*8, DIMENSION(%sSYNAPSES) :: %s'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stXdSF90, __pyx_kp_u_REAL_8_X) < 0) __PYX_ERR(0, 458, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":459
 * stPotentialdSF90 = 'REAL*8 POTENTIAL'
 * stXdSF90 = 'REAL*8 X'
 * stXPotentialdSF90 = 'REAL*8 X, POTENTIAL'             # <<<<<<<<<<<<<<
 * StArraySynapsesF90 = 'REAL*8, DIMENSION(%sSYNAPSES) :: %s'
 * StArrayMatrixF90 = 'REAL*8, DIMENSION(%sMATRIXES) :: %s'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stXPotentialdSF90, __pyx_kp_u_REAL_8_X_POTENTIAL) < 0) __PYX_ERR(0, 459, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":460
 * stXdSF90 = 'REAL*8 X'
 * stXPotentialdSF90 = 'REAL*8 X, POTENTIAL'
 * StArraySynapsesF90 = 'REAL*8, DIMENSION(%sSYNAPSES) :: %s'             # <<<<<<<<<<<<<<
 * StArrayMatrixF90 = 'REAL*8, DIMENSION(%sMATRIXES) :: %s'
 * stNodedSF90 = 'REAL*8, DIMENSION(%sNODES):: NODE'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StArraySynapsesF90, __pyx_kp_u_REAL_8_DIMENSION_sSYNAPSES_s) < 0) __PYX_ERR(0, 460, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":461
 * stXPotentialdSF90 = 'REAL*8 X, POTENTIAL'
 * StArraySynapsesF90 = 'REAL*8, DIMENSION(%sSYNAPSES) :: %s'
 * StArrayMatrixF90 = 'REAL*8, DIMENSION(%sMATRIXES) :: %s'             # <<<<<<<<<<<<<<
 * stNodedSF90 = 'REAL*8, DIMENSION(%sNODES):: NODE'
 * stNodedS2F90 = 'REAL*8, DIMENSION(2, %sNODES) :: NODE'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StArrayMatrixF90, __pyx_kp_u_REAL_8_DIMENSION_sMATRIXES_s) < 0) __PYX_ERR(0, 461, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":462
 * StArraySynapsesF90 = 'REAL*8, DIMENSION(%sSYNAPSES) :: %s'
 * StArrayMatrixF90 = 'REAL*8, DIMENSION(%sMATRIXES) :: %s'
 * stNodedSF90 = 'REAL*8, DIMENSION(%sNODES):: NODE'             # <<<<<<<<<<<<<<
 * stNodedS2F90 = 'REAL*8, DIMENSION(2, %sNODES) :: NODE'
 * stNodedS3F90 = 'REAL*8, DIMENSION(3, %sNODES) :: NODE'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stNodedSF90, __pyx_kp_u_REAL_8_DIMENSION_sNODES_NODE) < 0) __PYX_ERR(0, 462, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":463
 * StArrayMatrixF90 = 'REAL*8, DIMENSION(%sMATRIXES) :: %s'
 * stNodedSF90 = 'REAL*8, DIMENSION(%sNODES):: NODE'
 * stNodedS2F90 = 'REAL*8, DIMENSION(2, %sNODES) :: NODE'             # <<<<<<<<<<<<<<
 * stNodedS3F90 = 'REAL*8, DIMENSION(3, %sNODES) :: NODE'
 * stStatesDeclF90 = 'REAL*8, DIMENSION(%sORDER):: STATES'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stNodedS2F90, __pyx_kp_u_REAL_8_DIMENSION_2_sNODES_NODE) < 0) __PYX_ERR(0, 463, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":464
 * stNodedSF90 = 'REAL*8, DIMENSION(%sNODES):: NODE'
 * stNodedS2F90 = 'REAL*8, DIMENSION(2, %sNODES) :: NODE'
 * stNodedS3F90 = 'REAL*8, DIMENSION(3, %sNODES) :: NODE'             # <<<<<<<<<<<<<<
 * stStatesDeclF90 = 'REAL*8, DIMENSION(%sORDER):: STATES'
 * stMemSyndSF90 = 'REAL*8, DIMENSION(%sMEMSYN)   :: MEMSYN=(/(1.,i=1,%sMEMSYN)/)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stNodedS3F90, __pyx_kp_u_REAL_8_DIMENSION_3_sNODES_NODE) < 0) __PYX_ERR(0, 464, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":465
 * stNodedS2F90 = 'REAL*8, DIMENSION(2, %sNODES) :: NODE'
 * stNodedS3F90 = 'REAL*8, DIMENSION(3, %sNODES) :: NODE'
 * stStatesDeclF90 = 'REAL*8, DIMENSION(%sORDER):: STATES'             # <<<<<<<<<<<<<<
 * stMemSyndSF90 = 'REAL*8, DIMENSION(%sMEMSYN)   :: MEMSYN=(/(1.,i=1,%sMEMSYN)/)'
 * StInputsDeclF90 = 'REAL*8, DIMENSION(%sINPUTS) :: INPUTS'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stStatesDeclF90, __pyx_kp_u_REAL_8_DIMENSION_sORDER_STATES) < 0) __PYX_ERR(0, 465, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":466
 * stNodedS3F90 = 'REAL*8, DIMENSION(3, %sNODES) :: NODE'
 * stStatesDeclF90 = 'REAL*8, DIMENSION(%sORDER):: STATES'
 * stMemSyndSF90 = 'REAL*8, DIMENSION(%sMEMSYN)   :: MEMSYN=(/(1.,i=1,%sMEMSYN)/)'             # <<<<<<<<<<<<<<
 * StInputsDeclF90 = 'REAL*8, DIMENSION(%sINPUTS) :: INPUTS'
 * StOutPutsDeclF90 = 'REAL*8, DIMENSION(%sOUTPUTS):: OUTPUTS'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stMemSyndSF90, __pyx_kp_u_REAL_8_DIMENSION_sMEMSYN_MEMSYN) < 0) __PYX_ERR(0, 466, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":467
 * stStatesDeclF90 = 'REAL*8, DIMENSION(%sORDER):: STATES'
 * stMemSyndSF90 = 'REAL*8, DIMENSION(%sMEMSYN)   :: MEMSYN=(/(1.,i=1,%sMEMSYN)/)'
 * StInputsDeclF90 = 'REAL*8, DIMENSION(%sINPUTS) :: INPUTS'             # <<<<<<<<<<<<<<
 * StOutPutsDeclF90 = 'REAL*8, DIMENSION(%sOUTPUTS):: OUTPUTS'
 * stGradientDeclF90 = 'REAL*8, DIMENSION(%sSYNAPSES) :: GRADIENT=(/(0.,i=1,%sSYNAPSES)/)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StInputsDeclF90, __pyx_kp_u_REAL_8_DIMENSION_sINPUTS_INPUTS) < 0) __PYX_ERR(0, 467, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":468
 * stMemSyndSF90 = 'REAL*8, DIMENSION(%sMEMSYN)   :: MEMSYN=(/(1.,i=1,%sMEMSYN)/)'
 * StInputsDeclF90 = 'REAL*8, DIMENSION(%sINPUTS) :: INPUTS'
 * StOutPutsDeclF90 = 'REAL*8, DIMENSION(%sOUTPUTS):: OUTPUTS'             # <<<<<<<<<<<<<<
 * stGradientDeclF90 = 'REAL*8, DIMENSION(%sSYNAPSES) :: GRADIENT=(/(0.,i=1,%sSYNAPSES)/)'
 * stGradientDeclF90S = 'REAL*8, DIMENSION(%sSYNAPSES) :: GRADIENT'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StOutPutsDeclF90, __pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_OUTPUT) < 0) __PYX_ERR(0, 468, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":469
 * StInputsDeclF90 = 'REAL*8, DIMENSION(%sINPUTS) :: INPUTS'
 * StOutPutsDeclF90 = 'REAL*8, DIMENSION(%sOUTPUTS):: OUTPUTS'
 * stGradientDeclF90 = 'REAL*8, DIMENSION(%sSYNAPSES) :: GRADIENT=(/(0.,i=1,%sSYNAPSES)/)'             # <<<<<<<<<<<<<<
 * stGradientDeclF90S = 'REAL*8, DIMENSION(%sSYNAPSES) :: GRADIENT'
 * stInibackDeclF90 = 'REAL*8, DIMENSION(%sOUTPUTS):: INIBACK=(/(0.,i=1,%sOUTPUTS)/)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stGradientDeclF90, __pyx_kp_u_REAL_8_DIMENSION_sSYNAPSES_GRADI) < 0) __PYX_ERR(0, 469, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":470
 * StOutPutsDeclF90 = 'REAL*8, DIMENSION(%sOUTPUTS):: OUTPUTS'
 * stGradientDeclF90 = 'REAL*8, DIMENSION(%sSYNAPSES) :: GRADIENT=(/(0.,i=1,%sSYNAPSES)/)'
 * stGradientDeclF90S = 'REAL*8, DIMENSION(%sSYNAPSES) :: GRADIENT'             # <<<<<<<<<<<<<<
 * stInibackDeclF90 = 'REAL*8, DIMENSION(%sOUTPUTS):: INIBACK=(/(0.,i=1,%sOUTPUTS)/)'
 * stInibackDeclF90S = 'REAL*8, DIMENSION(%sOUTPUTS):: INIBACK'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stGradientDeclF90S, __pyx_kp_u_REAL_8_DIMENSION_sSYNAPSES_GRADI_2) < 0) __PYX_ERR(0, 470, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":471
 * stGradientDeclF90 = 'REAL*8, DIMENSION(%sSYNAPSES) :: GRADIENT=(/(0.,i=1,%sSYNAPSES)/)'
 * stGradientDeclF90S = 'REAL*8, DIMENSION(%sSYNAPSES) :: GRADIENT'
 * stInibackDeclF90 = 'REAL*8, DIMENSION(%sOUTPUTS):: INIBACK=(/(0.,i=1,%sOUTPUTS)/)'             # <<<<<<<<<<<<<<
 * stInibackDeclF90S = 'REAL*8, DIMENSION(%sOUTPUTS):: INIBACK'
 * stLeverageDeclF90 = 'REAL*8, DIMENSION(%sOUTPUTS):: LEVERAGE=(/(0.,i=1,%sOUTPUTS)/)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stInibackDeclF90, __pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_INIBAC) < 0) __PYX_ERR(0, 471, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":472
 * stGradientDeclF90S = 'REAL*8, DIMENSION(%sSYNAPSES) :: GRADIENT'
 * stInibackDeclF90 = 'REAL*8, DIMENSION(%sOUTPUTS):: INIBACK=(/(0.,i=1,%sOUTPUTS)/)'
 * stInibackDeclF90S = 'REAL*8, DIMENSION(%sOUTPUTS):: INIBACK'             # <<<<<<<<<<<<<<
 * stLeverageDeclF90 = 'REAL*8, DIMENSION(%sOUTPUTS):: LEVERAGE=(/(0.,i=1,%sOUTPUTS)/)'
 * stLeverageDeclF90S = 'REAL*8, DIMENSION(%sOUTPUTS):: LEVERAGE'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stInibackDeclF90S, __pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_INIBAC_2) < 0) __PYX_ERR(0, 472, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":473
 * stInibackDeclF90 = 'REAL*8, DIMENSION(%sOUTPUTS):: INIBACK=(/(0.,i=1,%sOUTPUTS)/)'
 * stInibackDeclF90S = 'REAL*8, DIMENSION(%sOUTPUTS):: INIBACK'
 * stLeverageDeclF90 = 'REAL*8, DIMENSION(%sOUTPUTS):: LEVERAGE=(/(0.,i=1,%sOUTPUTS)/)'             # <<<<<<<<<<<<<<
 * stLeverageDeclF90S = 'REAL*8, DIMENSION(%sOUTPUTS):: LEVERAGE'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stLeverageDeclF90, __pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_LEVERA) < 0) __PYX_ERR(0, 473, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":474
 * stInibackDeclF90S = 'REAL*8, DIMENSION(%sOUTPUTS):: INIBACK'
 * stLeverageDeclF90 = 'REAL*8, DIMENSION(%sOUTPUTS):: LEVERAGE=(/(0.,i=1,%sOUTPUTS)/)'
 * stLeverageDeclF90S = 'REAL*8, DIMENSION(%sOUTPUTS):: LEVERAGE'             # <<<<<<<<<<<<<<
 * 
 * StInputNameFU = '%sINPUTNAME'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_stLeverageDeclF90S, __pyx_kp_u_REAL_8_DIMENSION_sOUTPUTS_LEVERA_2) < 0) __PYX_ERR(0, 474, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":476
 * stLeverageDeclF90S = 'REAL*8, DIMENSION(%sOUTPUTS):: LEVERAGE'
 * 
 * StInputNameFU = '%sINPUTNAME'             # <<<<<<<<<<<<<<
 * StOutputNameFU = '%sOUTPUTNAME'
 * StInputNameF = 'FUNCTION %sINPUTNAME(INDEX)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StInputNameFU, __pyx_kp_u_sINPUTNAME) < 0) __PYX_ERR(0, 476, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":477
 * 
 * StInputNameFU = '%sINPUTNAME'
 * StOutputNameFU = '%sOUTPUTNAME'             # <<<<<<<<<<<<<<
 * StInputNameF = 'FUNCTION %sINPUTNAME(INDEX)'
 * StOutputNameF = 'FUNCTION %sOUTPUTNAME(INDEX)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StOutputNameFU, __pyx_kp_u_sOUTPUTNAME) < 0) __PYX_ERR(0, 477, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":478
 * StInputNameFU = '%sINPUTNAME'
 * StOutputNameFU = '%sOUTPUTNAME'
 * StInputNameF = 'FUNCTION %sINPUTNAME(INDEX)'             # <<<<<<<<<<<<<<
 * StOutputNameF = 'FUNCTION %sOUTPUTNAME(INDEX)'
 * StInputNameDeclF77 = 'CHARACTER*32 names(%sINPUTS)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StInputNameF, __pyx_kp_u_FUNCTION_sINPUTNAME_INDEX) < 0) __PYX_ERR(0, 478, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":479
 * StOutputNameFU = '%sOUTPUTNAME'
 * StInputNameF = 'FUNCTION %sINPUTNAME(INDEX)'
 * StOutputNameF = 'FUNCTION %sOUTPUTNAME(INDEX)'             # <<<<<<<<<<<<<<
 * StInputNameDeclF77 = 'CHARACTER*32 names(%sINPUTS)'
 * StOutputNameDeclF77 = 'CHARACTER*32 names(%sOUTPUTS)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StOutputNameF, __pyx_kp_u_FUNCTION_sOUTPUTNAME_INDEX) < 0) __PYX_ERR(0, 479, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":480
 * StInputNameF = 'FUNCTION %sINPUTNAME(INDEX)'
 * StOutputNameF = 'FUNCTION %sOUTPUTNAME(INDEX)'
 * StInputNameDeclF77 = 'CHARACTER*32 names(%sINPUTS)'             # <<<<<<<<<<<<<<
 * StOutputNameDeclF77 = 'CHARACTER*32 names(%sOUTPUTS)'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StInputNameDeclF77, __pyx_kp_u_CHARACTER_32_names_sINPUTS) < 0) __PYX_ERR(0, 480, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":481
 * StOutputNameF = 'FUNCTION %sOUTPUTNAME(INDEX)'
 * StInputNameDeclF77 = 'CHARACTER*32 names(%sINPUTS)'
 * StOutputNameDeclF77 = 'CHARACTER*32 names(%sOUTPUTS)'             # <<<<<<<<<<<<<<
 * 
 * StInputNameDeclF90 = 'CHARACTER(LEN=32), DIMENSION(%sINPUTS) :: names'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StOutputNameDeclF77, __pyx_kp_u_CHARACTER_32_names_sOUTPUTS) < 0) __PYX_ERR(0, 481, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":483
 * StOutputNameDeclF77 = 'CHARACTER*32 names(%sOUTPUTS)'
 * 
 * StInputNameDeclF90 = 'CHARACTER(LEN=32), DIMENSION(%sINPUTS) :: names'             # <<<<<<<<<<<<<<
 * StOutputNameDeclF90 = 'CHARACTER(LEN=32), DIMENSION(%sOUTPUTS) :: names'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StInputNameDeclF90, __pyx_kp_u_CHARACTER_LEN_32_DIMENSION_sINPU) < 0) __PYX_ERR(0, 483, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":484
 * 
 * StInputNameDeclF90 = 'CHARACTER(LEN=32), DIMENSION(%sINPUTS) :: names'
 * StOutputNameDeclF90 = 'CHARACTER(LEN=32), DIMENSION(%sOUTPUTS) :: names'             # <<<<<<<<<<<<<<
 * 
 * StDataMemSynF77 = 'DATA MEMSYN /%sMEMSYN * 1./'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StOutputNameDeclF90, __pyx_kp_u_CHARACTER_LEN_32_DIMENSION_sOUTP) < 0) __PYX_ERR(0, 484, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":486
 * StOutputNameDeclF90 = 'CHARACTER(LEN=32), DIMENSION(%sOUTPUTS) :: names'
 * 
 * StDataMemSynF77 = 'DATA MEMSYN /%sMEMSYN * 1./'             # <<<<<<<<<<<<<<
 * StDataGrIniLevF77 = 'DATA GRADIENT, INIBACK, LEVERAGE /'
 * StForAllSynapsesF77 = 'DO 140 I = 1, %sSYNAPSES'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StDataMemSynF77, __pyx_kp_u_DATA_MEMSYN_sMEMSYN_1) < 0) __PYX_ERR(0, 486, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":487
 * 
 * StDataMemSynF77 = 'DATA MEMSYN /%sMEMSYN * 1./'
 * StDataGrIniLevF77 = 'DATA GRADIENT, INIBACK, LEVERAGE /'             # <<<<<<<<<<<<<<
 * StForAllSynapsesF77 = 'DO 140 I = 1, %sSYNAPSES'
 * StForAllSynapsesjF77 = 'DO 150 J = 1, %sSYNAPSES'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StDataGrIniLevF77, __pyx_kp_u_DATA_GRADIENT_INIBACK_LEVERAGE) < 0) __PYX_ERR(0, 487, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":488
 * StDataMemSynF77 = 'DATA MEMSYN /%sMEMSYN * 1./'
 * StDataGrIniLevF77 = 'DATA GRADIENT, INIBACK, LEVERAGE /'
 * StForAllSynapsesF77 = 'DO 140 I = 1, %sSYNAPSES'             # <<<<<<<<<<<<<<
 * StForAllSynapsesjF77 = 'DO 150 J = 1, %sSYNAPSES'
 * StForAllSynapsesF90 = 'DO I = 1, %sSYNAPSES'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllSynapsesF77, __pyx_kp_u_DO_140_I_1_sSYNAPSES) < 0) __PYX_ERR(0, 488, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":489
 * StDataGrIniLevF77 = 'DATA GRADIENT, INIBACK, LEVERAGE /'
 * StForAllSynapsesF77 = 'DO 140 I = 1, %sSYNAPSES'
 * StForAllSynapsesjF77 = 'DO 150 J = 1, %sSYNAPSES'             # <<<<<<<<<<<<<<
 * StForAllSynapsesF90 = 'DO I = 1, %sSYNAPSES'
 * StForAllSynapsesjF90 = 'DO J = 1, %sSYNAPSES'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllSynapsesjF77, __pyx_kp_u_DO_150_J_1_sSYNAPSES) < 0) __PYX_ERR(0, 489, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":490
 * StForAllSynapsesF77 = 'DO 140 I = 1, %sSYNAPSES'
 * StForAllSynapsesjF77 = 'DO 150 J = 1, %sSYNAPSES'
 * StForAllSynapsesF90 = 'DO I = 1, %sSYNAPSES'             # <<<<<<<<<<<<<<
 * StForAllSynapsesjF90 = 'DO J = 1, %sSYNAPSES'
 * StMatrixComputeF = '%sMATRIX(K) * GRADIENT(J)'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllSynapsesF90, __pyx_kp_u_DO_I_1_sSYNAPSES) < 0) __PYX_ERR(0, 490, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":491
 * StForAllSynapsesjF77 = 'DO 150 J = 1, %sSYNAPSES'
 * StForAllSynapsesF90 = 'DO I = 1, %sSYNAPSES'
 * StForAllSynapsesjF90 = 'DO J = 1, %sSYNAPSES'             # <<<<<<<<<<<<<<
 * StMatrixComputeF = '%sMATRIX(K) * GRADIENT(J)'
 * StForAllOutputsF77 = 'DO 140 I = 2, %sOUTPUTS'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllSynapsesjF90, __pyx_kp_u_DO_J_1_sSYNAPSES) < 0) __PYX_ERR(0, 491, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":492
 * StForAllSynapsesF90 = 'DO I = 1, %sSYNAPSES'
 * StForAllSynapsesjF90 = 'DO J = 1, %sSYNAPSES'
 * StMatrixComputeF = '%sMATRIX(K) * GRADIENT(J)'             # <<<<<<<<<<<<<<
 * StForAllOutputsF77 = 'DO 140 I = 2, %sOUTPUTS'
 * StForAllOutputsF90 = 'DO I = 2, %sOUTPUTS'
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StMatrixComputeF, __pyx_kp_u_sMATRIX_K_GRADIENT_J) < 0) __PYX_ERR(0, 492, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":493
 * StForAllSynapsesjF90 = 'DO J = 1, %sSYNAPSES'
 * StMatrixComputeF = '%sMATRIX(K) * GRADIENT(J)'
 * StForAllOutputsF77 = 'DO 140 I = 2, %sOUTPUTS'             # <<<<<<<<<<<<<<
 * StForAllOutputsF90 = 'DO I = 2, %sOUTPUTS'
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllOutputsF77, __pyx_kp_u_DO_140_I_2_sOUTPUTS) < 0) __PYX_ERR(0, 493, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":494
 * StMatrixComputeF = '%sMATRIX(K) * GRADIENT(J)'
 * StForAllOutputsF77 = 'DO 140 I = 2, %sOUTPUTS'
 * StForAllOutputsF90 = 'DO I = 2, %sOUTPUTS'             # <<<<<<<<<<<<<<
 * 
 * std_re = "(?P<%s>%s)"% ("std", '%.+%')
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_StForAllOutputsF90, __pyx_kp_u_DO_I_2_sOUTPUTS) < 0) __PYX_ERR(0, 494, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":496
 * StForAllOutputsF90 = 'DO I = 2, %sOUTPUTS'
 * 
 * std_re = "(?P<%s>%s)"% ("std", '%.+%')             # <<<<<<<<<<<<<<
 * lower_re = "(?P<%s>%s)"% ("lower", '%l.+l%')
 * upper_re = "(?P<%s>%s)"% ("upper", '%U.+U%')
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_std_re, __pyx_kp_u_P_std) < 0) __PYX_ERR(0, 496, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":497
 * 
 * std_re = "(?P<%s>%s)"% ("std", '%.+%')
 * lower_re = "(?P<%s>%s)"% ("lower", '%l.+l%')             # <<<<<<<<<<<<<<
 * upper_re = "(?P<%s>%s)"% ("upper", '%U.+U%')
 * subpattern = '|'.join([lower_re, upper_re, std_re])
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_lower_re, __pyx_kp_u_P_lower_l_l) < 0) __PYX_ERR(0, 497, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":498
 * std_re = "(?P<%s>%s)"% ("std", '%.+%')
 * lower_re = "(?P<%s>%s)"% ("lower", '%l.+l%')
 * upper_re = "(?P<%s>%s)"% ("upper", '%U.+U%')             # <<<<<<<<<<<<<<
 * subpattern = '|'.join([lower_re, upper_re, std_re])
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_upper_re, __pyx_kp_u_P_upper_U_U) < 0) __PYX_ERR(0, 498, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":499
 * lower_re = "(?P<%s>%s)"% ("lower", '%l.+l%')
 * upper_re = "(?P<%s>%s)"% ("upper", '%U.+U%')
 * subpattern = '|'.join([lower_re, upper_re, std_re])             # <<<<<<<<<<<<<<
 * 
 * proprietaryname = 'LartenNetral_JLP_04_05_47'
 */
  __pyx_t_22 = __Pyx_GetModuleGlobalName(__pyx_n_s_lower_re); if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_22);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_upper_re); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_std_re); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_19 = PyList_New(3); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __Pyx_GIVEREF(__pyx_t_22);
  PyList_SET_ITEM(__pyx_t_19, 0, __pyx_t_22);
  __Pyx_GIVEREF(__pyx_t_7);
  PyList_SET_ITEM(__pyx_t_19, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_3);
  PyList_SET_ITEM(__pyx_t_19, 2, __pyx_t_3);
  __pyx_t_22 = 0;
  __pyx_t_7 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyUnicode_Join(__pyx_kp_u__18, __pyx_t_19); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_subpattern, __pyx_t_3) < 0) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "metaphor/monal/lcode/codeconst.py":501
 * subpattern = '|'.join([lower_re, upper_re, std_re])
 * 
 * proprietaryname = 'LartenNetral_JLP_04_05_47'             # <<<<<<<<<<<<<<
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_proprietaryname, __pyx_n_u_LartenNetral_JLP_04_05_47) < 0) __PYX_ERR(0, 501, __pyx_L1_error)

  /* "metaphor/monal/lcode/codeconst.py":2
 *  #-*- coding: ISO-8859-15 -*-
 * #===============================================================================             # <<<<<<<<<<<<<<
 * # $Id: codeconst.py 4732 2018-03-20 20:00:04Z jeanluc $
 * #
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_3) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_XDECREF(__pyx_t_21);
  __Pyx_XDECREF(__pyx_t_22);
  __Pyx_XDECREF(__pyx_t_23);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init metaphor.monal.lcode.codeconst", 0, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init metaphor.monal.lcode.codeconst");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* GetModuleGlobalName */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    if (likely(result)) {
        Py_INCREF(result);
    } else if (unlikely(PyErr_Occurred())) {
        result = NULL;
    } else {
#else
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* PyErrFetchRestore */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* CLineInTraceback */
    #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
      use_cline = __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback);
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (PyObject_Not(use_cline) != 0) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
    static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
    #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
    static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
    #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
    static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
    static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
    #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
    static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
    static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
